/*
 * amd16.c: AMD 16 bit (1x 16 bit) flash driver
 *
 * Copyright (C) 2001  Erik Mouw (J.A.K.Mouw@its.tudelft.nl)
 * Copyright (C) 2003  Russ Dill (Russ.Dill@asu.edu)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * http://www.amd.com/us-en/assets/content_type/white_papers_and_tech_docs/21358.pdf
 *
 */

#ident "$Id: amd16.c,v 1.1 2003/09/03 01:44:28 russd Exp $"

#ifdef HAVE_CONFIG_H
# include <blob/config.h>
#endif

#include <blob/errno.h>
#include <blob/flash.h>
#include <blob/util.h>
#include <blob/serial.h>


/* flash commands for a 16 bit AMD flash chips */
#define READ_ARRAY		0x00F0
#define UNLOCK1			0x00AA
#define UNLOCK2			0x0055
#define CONFIG_QUERY		0x0090
#define EXIT_UNLOCK1		0x0090
#define EXIT_UNLOCK2		0x0000
#define ERASE_SETUP		0x0080
#define ERASE_CONFIRM		0x0030
#define PGM_SETUP		0x00A0
#define UNLOCK_BYPASS		0x0020
#define FLASH_ADDR1		(0x00005555 << 1)
#define FLASH_ADDR2		(0x00002AAA << 1)
#define ERASE_DONE		(1 << 7)
#define RDY_MASK		(1 << 7)
#define TOGGLE_BIT		(1 << 6)
#define STATUS_PGM_ERR		(1 << 5)
#define STATUS_ERASE_ERR	(1 << 0)


static void flash_read_array_amd16(void)
{
	*(u16 *)FLASH_ADDR1 = READ_ARRAY;
	barrier();
}


static int get_status_amd16(u16 *addr)
{
	u16 result, last;
	int last_chance = 0;

	last = *addr;
	barrier();

	for (;;) {
		result = *addr;
		barrier();

		/* didn't toggle */
		if (!((result ^ last) & TOGGLE_BIT)) return 0;

		/* toggled on its last chance */
		if (last_chance) return -1;
		
		if (result & STATUS_PGM_ERR) {
			last_chance = 1;
			result = *addr; /* dummy read */
			barrier();
		}

		last = result;
	}

	return 0;
}


static int flash_erase_amd16(u32 *addr)
{
	/* prepare for erase */
	*(u16 *)FLASH_ADDR1 = UNLOCK1;
	barrier();
	*(u16 *)FLASH_ADDR2 = UNLOCK2;
	barrier();
	*(u16 *)FLASH_ADDR1 = ERASE_SETUP;
	barrier();

	/* erase command */
	*(u16 *)FLASH_ADDR1 = UNLOCK1;
	barrier();
	*(u16 *)FLASH_ADDR2 = UNLOCK2;
	barrier();
	*((u16 *) addr) = ERASE_CONFIRM;

	if (get_status_amd16((u16 *) addr) < 0) {
		/* put flash back into Read Array mode */
		flash_read_array_amd16();
		return -EFLASHERASE;
	}

	return 0;
}


static int flash_write_amd(u16 *dst, const u16 *src)
{
	int ret;

	*(u16 *)FLASH_ADDR1 = UNLOCK1;
	barrier();
	*(u16 *)FLASH_ADDR2 = UNLOCK2;
	barrier();
	*(u16 *)FLASH_ADDR1 = PGM_SETUP;
	barrier();

	*dst = *src;
	barrier();

	ret = get_status_amd16(dst);
	
	barrier();
	if (ret < 0 || *dst != *src) {
		barrier();
		flash_read_array_amd16();	
		return -EFLASHPGM;
	}

	return 0;
}
	

static int flash_write_amd16(u32 *dst, const u32 *src)
{
	if (flash_write_amd((u16 *)dst, (const u16 *)src) < 0)
		return -EFLASHPGM;
	if (flash_write_amd(((u16 *)dst) + 1, ((const u16 *) src) + 1) < 0)
		return -EFLASHPGM;
	return 0;
}


static int flash_lock_block_amd16(u32 *blockStart)
{
	/* FIXME: this function should be fleshed out -- Erik */
	return 0;
}


static int flash_unlock_block_amd16(u32 *blockStart)
{
	/* FIXME: this function should be fleshed out -- Erik */
	return 0;
}


static int flash_query_block_lock_amd16(u32 *blockStart)
{
	int result;
	*(u16 *)FLASH_ADDR1 = UNLOCK1;
	barrier();
	*(u16 *)FLASH_ADDR2 = UNLOCK2;
	barrier();
	*(u16 *)FLASH_ADDR1 = CONFIG_QUERY;
	barrier();

	result = *(u16 *)(((u16 *)blockStart) + 2);
	flash_read_array_amd16();
	return !!result;

}



/* flash driver structure */
flash_driver_t amd16_flash_driver = {
	erase:			flash_erase_amd16,
	write:			flash_write_amd16,
	lock_block:		flash_lock_block_amd16,
	unlock_block:		flash_unlock_block_amd16,
	query_block_lock:	flash_query_block_lock_amd16
};
