/*
 * amd32.c: AMD 32 bit (2x 16 bit) flash driver
 *
 * Copyright (C) 2001  Erik Mouw (J.A.K.Mouw@its.tudelft.nl)
 * Copyright (C) 2001  Russ Dill (Russ.Dill@asu.edu)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * http://www.amd.com/us-en/assets/content_type/white_papers_and_tech_docs/21358.pdf
 *
 */

#ident "$Id: amd32.c,v 1.6 2002/06/19 04:41:49 russd Exp $"

#ifdef HAVE_CONFIG_H
# include <blob/config.h>
#endif

#include <blob/errno.h>
#include <blob/flash.h>
#include <blob/util.h>
#include <blob/serial.h>


/* flash commands for two 16 bit AMD flash chips */
#define READ_ARRAY		0x00F000F0
#define UNLOCK1			0x00AA00AA
#define UNLOCK2			0x00550055
#define CONFIG_QUERY		0x00900090
#define EXIT_UNLOCK1		0x00900090
#define EXIT_UNLOCK2		0x00000000
#define ERASE_SETUP		0x00800080
#define ERASE_CONFIRM		0x00300030
#define PGM_SETUP		0x00A000A0
#define UNLOCK_BYPASS		0x00200020
#define FLASH_ADDR1		(0x00000555 << 2)
#define FLASH_ADDR2		(0x000002AA << 2)
#define ERASE_DONE		0x00800080
#define TOGGLE_BIT		0x00400040
#define RDY_MASK		0x00800080
#define STATUS_PGM_ERR		0x00200020
#define STATUS_ERASE_ERR	0x00000001

#define READY 		1
#define LAST_CHANCE 	2
#define LAST_CHANCE2	3
#define ERR		4


static void flash_read_array_amd32(void)
{
	*(u32 *)FLASH_ADDR1 = data_to_flash(READ_ARRAY);
	barrier();
}

/* NOT! for amd16.c, this examines one chip of the two chip interleave */
static int get_status_amd16(u32 result, u32 last, int chip)
{
	if (chip == LAST_CHANCE)
		chip = LAST_CHANCE2;
	else if (!(result ^ last)) /* didn't toggle */
		chip = READY;
	else if (chip == LAST_CHANCE2)
		chip = ERR;
	else if (result & STATUS_PGM_ERR)
		chip = LAST_CHANCE;
	return chip;
}

static int get_status_amd32(u32 *addr)
{
	u32 result, last;
	int chip1, chip2;

	chip1 = chip2 = 0;

	last = data_from_flash(*addr);
	barrier();

	do {
		result = data_from_flash(*addr);
		barrier();

		if ((chip1 != ERR && chip1 != READY))
			chip1 = get_status_amd16(result & 0xFFFF, last & 0xFFFF, chip1);

		if ((chip2 != ERR && chip2 != READY))
			chip2 = get_status_amd16(result >> 16, last >> 16, chip2);
	
		last = result;

	} while ((chip1 != ERR && chip1 != READY) || (chip2 != ERR && chip2 != READY));

	if (chip1 == ERR || chip2 == ERR) return -1;
	
	return 0;

}


static int flash_erase_amd32(u32 *addr)
{
	/* prepare for erase */
	*(u32 *)FLASH_ADDR1 = data_to_flash(UNLOCK1);
	barrier();
	*(u32 *)FLASH_ADDR2 = data_to_flash(UNLOCK2);
	barrier();
	*(u32 *)FLASH_ADDR1 = data_to_flash(ERASE_SETUP);
	barrier();

	/* erase command */
	*(u32 *)FLASH_ADDR1 = data_to_flash(UNLOCK1);
	barrier();
	*(u32 *)FLASH_ADDR2 = data_to_flash(UNLOCK2);
	barrier();
	*addr = data_to_flash(ERASE_CONFIRM);

	/* I just can't find clean ways of dealing with this flash...
	 * The error bit is a *set* bit, so if its read, and bit 7 is 0,
	 * but bit 5 is 1, its an error, however, after these status reads
	 * are done, erased flash goes to 0xff...sooo...each chip has to
	 * be caught where the bits are the status bits -- Russ */

	/* Russ, why don't you do this like the LART does? Just check
         * the status of chips with a single compare. -- Erik */

	/* Maybe this restructuring will amke it more clear, when the flash
	 * is done erasing, it will return 0xFFFF on read, which contains the
	 * status pgm err bit. So an error condition occurs if the status pgm
	 * bit is set, but the erase done bit is not. OK, switch this over
	 * to detecting bit toggles. --Russ */

	if (get_status_amd32(addr) < 0) {
		/* put flash back into Read Array mode */
		flash_read_array_amd32();
		return -EFLASHERASE;
	}

	return 0;
}




static int flash_write_amd32(u32 *dst, const u32* src)
{

	*(u32 *)FLASH_ADDR1 = data_to_flash(UNLOCK1);
	barrier();
	*(u32 *)FLASH_ADDR2 = data_to_flash(UNLOCK2);
	barrier();
	*(u32 *)FLASH_ADDR1 = data_to_flash(PGM_SETUP);
	barrier();

	*dst = *src;
	barrier();
	
	if (get_status_amd32(dst) < 0 || *dst != *src) {
		barrier();
		flash_read_array_amd32();	
		return -EFLASHPGM;
	}

	return 0;
}




static int flash_lock_block_amd32(u32 *blockStart)
{
	/* FIXME: this function should be fleshed out -- Erik */
	return 0;
}




static int flash_unlock_block_amd32(u32 *blockStart)
{
	/* FIXME: this function should be fleshed out -- Erik */
	return 0;
}




static int flash_query_block_lock_amd32(u32 *blockStart)
{
	int result;
	*(u32 *)FLASH_ADDR1 = data_to_flash(UNLOCK1);
	barrier();
	*(u32 *)FLASH_ADDR2 = data_to_flash(UNLOCK2);
	barrier();
	*(u32 *)FLASH_ADDR1 = data_to_flash(CONFIG_QUERY);
	barrier();

	result = *(u32 *)(blockStart + 2);
	flash_read_array_amd32();
	return !!result;

}




/* flash driver structure */
flash_driver_t amd32_flash_driver = {
	erase:			flash_erase_amd32,
	write:			flash_write_amd32,
	lock_block:		flash_lock_block_amd32,
	unlock_block:		flash_unlock_block_amd32,
	query_block_lock:	flash_query_block_lock_amd32
};
