/*-------------------------------------------------------------------------
 * Filename:      load_kernel.c
 * Version:       $Id: load_kernel.c,v 1.16 2003/08/06 22:55:41 abz Exp $
 * Copyright:     Copyright (C) 2001, Russ Dill
 * Author:        Russ Dill <Russ.Dill@asu.edu>
 * Description:   interface between kernel loaders and boot loader.
 		  if everything was done right, this *should be the only
 		  file that needs changing for porting to different boot
 		  loaders.
 *-----------------------------------------------------------------------*/
/*
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ident "$Id: load_kernel.c,v 1.16 2003/08/06 22:55:41 abz Exp $"

#ifdef HAVE_CONFIG_H
# include <blob/config.h>
#endif

#include <blob/arch.h>
#include <blob/command.h>
#include <blob/errno.h>
#include <blob/error.h>
#include <blob/load_kernel.h>
#include <blob/util.h>
#include <blob/flash.h>
#include <blob/partition.h>

#ifdef CONFIG_CRAMFS_SUPPORT
extern struct kernel_loader cramfs_load;
#endif
#ifdef CONFIG_ZIMAGE_SUPPORT
extern struct kernel_loader zImage_load;
#endif
#ifdef CONFIG_JFFS2_SUPPORT
extern struct kernel_loader jffs2_load;
#endif

struct kernel_loader *known_loaders[] = {
#ifdef CONFIG_CRAMFS_SUPPORT
	&cramfs_load, 
#endif
#ifdef CONFIG_ZIMAGE_SUPPORT
	&zImage_load,
#endif
#ifdef CONFIG_JFFS2_SUPPORT
	&jffs2_load, 
#endif
	0
};

#define FODDER_RAM_BASE (0xC0F00000)

const void *fodder_ram_base = (void *) FODDER_RAM_BASE;

	

/* function calls for blob */
void ldr_update_progress(void)
{
	printf(".");
}

int load_kernel(blob_status_t *blob_status)
{
	u32 size;
	const struct kernel_loader *loader = 0;
	const blob_partition_t *p;
	struct part_info part;

	for (p = pt_first(); p != 0; p = pt_next(p)) {
#ifdef CONFIG_JFFS2_SUPPORT
		if (p->flags & BLOB_PART_FLAG_JFFS2)
			loader = &jffs2_load;
		else 
#endif
#ifdef CONFIG_CRAMFS_SUPPORT
		if (p->flags & BLOB_PART_FLAG_CRAMFS)
			loader = &cramfs_load;
		else 
#endif
#ifdef CONFIG_ZIMAGE_SUPPORT
		if (p->flags & BLOB_PART_FLAG_EXEC) /* ?? */
			loader = &zImage_load;
		else
#endif
			continue;
			
		part.offset = pt_flash_start(p);
		part.size = p->size;
		part.erasesize = flash_get_block_size((u32)part.offset);

		if (loader->check_magic(&part))
			break;
	}

	if (!p) {
		eprintf("Unable to find kernel\n");
		return EINVAL;
	}


	printf("Loading kernel from partition %s (%s) ...",
	       p->name, loader->name);
	if ((size = loader->load_kernel((u32 *) KERNEL_RAM_BASE,
					&part, "/boot/linux")) == 0) {
		eprintf("error loading kernel!\n");
		return(EINVAL);
	}

	blob_status->kernelSize = size;
	blob_status->kernelType = fromFlash;

	printf("loaded 0x%08x (%d) bytes\n", size, size);
	return(ENOERROR);
}
