/*
 * (C) Copyright 2006 Marvell International Ltd.
 * All Rights Reserved
 */

#ifdef HAVE_CONFIG_H
#include <blob/config.h>
#endif

#include <blob/arch.h>
#include <blob/init.h>
#include <blob/errno.h>
#include <blob/reboot.h>
#include <blob/serial.h>
#include <blob/time.h>
#include <blob/command.h>
#include <blob/util.h>
#include <net/ether.h>
#include <net/net.h>
#include <blob/strata.h>
#include "xllp_dfc_proc.h"
#include "xllp_clkmgr.h"
#include "onenand.h"
#include "main.h"

#undef DEBUG_NAND

#ifdef DEBUG_NAND

#define DPRINTF(fmt,args...) printf( fmt, ##args )

#define PRINT_BUF(buf, num)  print_buf((buf), (num))  

#else

#define DPRINTF(fmt,args...)
#define PRINT_BUF(buf, num)

#endif


#define fls(x)	({ int __r;	\
		asm("clz\t%0, %1":"=r"(__r):"r"(x):"cc"); \
		31 - __r;})

/* The maxium block number is 8096. So the maxium bbt length is 2048 bytes */
#define BBT_LEN		2048

#define NAND_RELOC_MAX	127

#define NAND_RELOC_HEADER	0x0000524e //= 'N' + ('R' * 256)

/* The bad block table is in the last good block of the device */
#define	NAND_BBT_LASTBLOCK	0x00000010
/* bbt has a version counter at offset veroffs */
#define NAND_BBT_VERSION	0x00000100

#define MAX_TIM_SIZE	0x1000
#define MAX_BBT_SLOTS	40

#define BBT_BLOCK_BAD	0x03
#define BBT_BLOCK_GOOD	0x00
#define BBT_BLOCK_REV1	0x01
#define BBT_BLOCK_REV2	0x02

#define ERR		0x100
#define EBADBLOCK	(ERR + 1)
#define ECMDFAIL	(ERR + 2)
#define ENOSPACE	(ERR + 3)
#define ERELOCATED		(ERR + 4)

#define BUFLEN		(2048 + 64)

#define READ_NORM       0
#define READ_OOB        1
#define READ_RAW        2

typedef enum {
	PU_REASON_HARDWARE_RESET,
	PU_REASON_WATCHDOG,
	PU_REASON_RTC_ALARM,
	PU_REASON_PWR_KEY_PRESS,
	PU_REASON_USB_CABLE,
	PU_REASON_USB_FACTORY,
	PU_REASON_USB_CHARGER,
	PU_REASON_REBOOT_NORMAL,
	PU_REASON_REBOOT_FACTORY,
	PU_REASON_REBOOT_RECOVERY,
	PU_REASON_REBOOT_FOTA,
	PU_REASON_REBOOT_ASSERT,
	PU_REASON_LOW_PWR_RESET,
	PU_REASON_SDL_FLASHMODE,
	PU_REASON_MAX
} powerup_reason_t;

extern int get_ramdisk_len( void );
extern void set_ramdisk_len( int );
extern int get_tftp_received_len( void );
extern unsigned int get_powerup_reason( void );

enum {
	MHN_OBM_NULL = 0,
	MHN_OBM_V1,
	MHN_OBM_V2,
	MHN_OBM_V3,
	MHN_OBM_INVAL
};

struct nand_rel_table{
	u16 header;
	u16 total;
	struct {
		u16 from;
		u16 to;
	} rel[NAND_RELOC_MAX];
};

struct nand_bbt_descr {
	int	options;
	int	pages;
	int	offs;
	int	veroffs;
	u8	version;
	int	len;
	int 	maxblocks;
	int	reserved_block_code;
	u8	*pattern;
};

struct yaffs_oob_desc {
	int	offset;
	int	size;
};

struct nand_info {
	u8 maf_id;
	u8 dev_id;
	u32 chipsize;
	int page_shift;
	int phys_erase_shift;
	int eccbytes;
	u32 oobsize;
	u32 hwecc;
	u8 bbt[BBT_LEN];
	u32 badblockpos;
	u32 badblockbytes;
	struct nand_bbt_descr *td;
	struct nand_bbt_descr *md;
	/* current slot in relocation table */
	s32 current_slot;


	XLLP_DFC_CONTEXT_T *pcontext;
	struct onenand_info *onenand_info;
	
	struct nand_rel_table table;
	unsigned int table_init;
}; 

typedef unsigned char u_char;

typedef enum {
	NULLFUNC = 0,
	ZIMAGE = 1,
	JFFS2 = 2,
	YAFFS = 3
}WRITE_FUNC;

static int nand_reloc_max;

static struct nand_info monahans_nand;
static u8 jffs2_sb_oob[16] = {
	0xFF, 0xFF, 0x85, 0x19, 0x03, 0x20, 0x08, 0x00,
	0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF
};

/* Need auto conversion according Onenand ecc layout -- yfw */
static u8 jffs2_lb_oob[64] = {
	0xFF, 0xFF, 0x85, 0x19, 0x03, 0x20, 0x08, 0x00,
	0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF
};

static struct yaffs_oob_desc yaffs_oob = {2, 38};

/* This pattern is used to check bad block */
static u8 scan_ff_pattern[] = { 0xff, 0xff };
static struct nand_bbt_descr __attribute__((unused))
monahans_bbt_default = {
	.options = NAND_BBT_LASTBLOCK | NAND_BBT_VERSION,
	.maxblocks = 2,
	.len = 2,
	.offs = 0,
	.pattern = scan_ff_pattern
};

/* These two patterns are used to check BBT */
static u8 scan_main_bbt_pattern[] = { 'p', 'x', 'a', '1'};
static struct nand_bbt_descr monahans_bbt_main = {
	.options = NAND_BBT_LASTBLOCK | NAND_BBT_VERSION,
	.veroffs = 6,
	.maxblocks = 2,
	.len = 4,
	.offs = 2,
	.pattern = scan_main_bbt_pattern,
};

static u8 scan_mirror_bbt_pattern[] = { '0', 'a', 'x', 'p'};
static struct nand_bbt_descr monahans_bbt_mirror = {
	.options = NAND_BBT_LASTBLOCK | NAND_BBT_VERSION,
	.veroffs = 6,
	.maxblocks = 2,
	.len = 4,
	.offs = 2,
	.pattern = scan_mirror_bbt_pattern,
};

extern blob_status_t blob_status;
#if defined(CHERRY) || defined(BSTD)
extern inline unsigned get_boot_info( void );
#endif
static unsigned char buf[BUFLEN];
static int write_bbt (struct nand_info *info, u8 *buf, struct nand_bbt_descr *td, struct nand_bbt_descr *md);
static inline void set_bbt(u32 block, u8 val, struct nand_info *info);
static int is_bad_block(u32 block, struct nand_info *info);
static int mark_badblock(u32 block, struct nand_info *info);
static int sync_rel_table(struct nand_info *info, int *idx);
static int search_rel_block(int block, struct nand_info *info);
static int update_rel_table(struct nand_info *info, int block);
static void scan_nand(struct nand_info *info);

static void print_buf(char *buf, int num)
{
	int i = 0; 
	while (i < num) {
		printf("%07x %04x %04x %04x %04x %04x %04x %04x %04x\n", 
				(unsigned int) (i),  *((unsigned short int *)(buf + i)), 
				*((unsigned short int *)(buf + i + 2)), 
				*((unsigned short int *)(buf + i + 4)), 
				*((unsigned short int *)(buf + i + 6)),
				*((unsigned short int *)(buf + i + 8)),
				*((unsigned short int *)(buf + i + 10)),
				*((unsigned short int *)(buf + i + 12)),
				*((unsigned short int *)(buf + i + 14)));
		i += 16;
	}
}


int is_digital( char *str )
{
    char *p;
    int  i;
    
    p = str;
    
    if ( (('0' == p[0]) && ('X' == p[1] || 'x' == p[1] )) ) {
        for ( i = 2; 0 != p[i]; i++) {
            if ( (('0' <= p[i]) && ('9' >= p[i])) || (('a' <= p[i] ) && ('f' >= p[i])) || 
            (('A' <= p[i] ) && ('F' >= p[i])) ) {
				continue;
            }else {
                return 0;
            }
        }
    }else {
        for ( i = 0; 0 != p[i]; i++ ) {
            if ( ('0' > p[i]) || ('9' < p[i]) ) {
                return 0;
            }
        }
    }
    
    return 1;
}


#ifdef CONFIG_ONENAND_SUPPORT	
#include "onenand.c"
#else
#include "pxa3xx_nand.c"
#endif

/*
 * MOBM V2 is used on MH-P B0/B1/B2, MH-PL B1, MH-L A0.
 * MOBM V3 is used on MH-LV A0, ...
 */
static int calc_obm_ver(void)
{
	unsigned int	cpuid;
	/* read CPU ID */
	__asm__ (
		"mrc p15, 0, %0, c0, c0, 0\n"
		: "=r" (cpuid)
	);
	if ((cpuid & 0xFFFFFFF0) == 0x56056930) {
		/* It's TavorPV series */
		return MHN_OBM_V3;
	}

	if ((cpuid & 0xFFFF0000) != 0x69050000) {
		/* It's not xscale chip. */
		return MHN_OBM_INVAL;
	}
	if ((cpuid & 0x0000FFF0) == 0x00006420) {
		/* It's MhnP Ax */
		return MHN_OBM_V2;
	}
	if ((cpuid & 0x0000FFF0) == 0x00006820) {
		/* It's MhnP Bx */
		if ((cpuid & 0x0F) <= 6)
			return MHN_OBM_V2;
		else
			return MHN_OBM_V3;
	}
	if ((cpuid & 0x0000FFF0) == 0x00006880) {
		/* It's MhnL Ax */
		if ((cpuid & 0x0F) == 0)
			return MHN_OBM_V2;
		else
			return MHN_OBM_V3;
	}
	if ((cpuid & 0x0000FFF0) == 0x00006890) {
		/* It's MhnLV Ax */
		return MHN_OBM_V3;
	}
	if ((cpuid & 0x0000FFFF) == 0x00006834) {
		/* It's Tavor A0 */
		return MHN_OBM_V2;
	}
	if ((cpuid & 0x0000FFFF) == 0x00006835) {
		/* It's Tavor B0 */
		return MHN_OBM_V3;
	}

	if ((cpuid & 0x0000FFFF) == 0x00006837) {
		/* It's Tavor B1 */
		return MHN_OBM_V3;
	}

	if ((cpuid & 0x0000FFFF) == 0x00006838) {
		/* It's Tavor B2 */
		return MHN_OBM_V3;
	}

	return MHN_OBM_INVAL;
}

static void dump_rel_table(struct nand_rel_table *table)
{
	int i, max;

	if (table->total > nand_reloc_max)
		max = nand_reloc_max;
	else
		max = table->total;
	DPRINTF("Header: 0x%x, Total: %d\n", table->header, table->total);
	for (i = 0; i < max; i++) {
		printf("Block #%d is relocated to #%d\n", table->rel[i].from,
			table->rel[i].to);
	}
}

static void dump_bbt(struct nand_info *info)
{
	unsigned int i;

	for (i = 0;i < (info->chipsize >> info->phys_erase_shift);i++){
		if (is_bad_block(i, info))
			printf("Block #%d is bad\n", i);
	}
}

static int update_rel_table(struct nand_info *info, int block);

static void set_badblock(int block, struct nand_info *info)
{
	//unsigned char buf[BUFLEN];
	int pagesize, page;
	u32 i;

	pagesize = 1 << info->page_shift;
	memset(buf, 0xFF, BUFLEN);
	for (i = 0; i < info->badblockbytes; i++)
		buf[pagesize + info->badblockpos + i] = 0;

	mhn_erase_one_block(info, block);
	page = block << (info->phys_erase_shift - info->page_shift);
	DPRINTF("set offset:0x%x as bad block\n",
		block << info->phys_erase_shift);

	/* doesn't modify the ECC
	 * Set the first page of the block as bad.
	 */
	mhn_write_one_page(info, buf, page, 0);
	mhn_write_one_page(info, buf, page + 1, 0);
	//update_rel_table(info, block);
}

static void clear_badblock(int block, struct nand_info *info)
{
	unsigned int max, i, j;
	struct nand_rel_table table;
	unsigned char temp_buf[BUFLEN];

	/* cancel the relocation for this block */
	if (info->table.total > nand_reloc_max)
		max = nand_reloc_max;
	else
		max = info->table.total;

	memcpy(&table, &info->table, sizeof(struct nand_rel_table));
	for (i = 0, j = 0; i < max; i++) {
		if (block == info->table.rel[i].from)
			continue;

		table.rel[j].from = info->table.rel[i].from;
		table.rel[j++].to = info->table.rel[i].to;
	}
	table.total = j;

	/* mark it as good if it is marked as bad block previously */
	if (is_bad_block(block, info)){
		set_bbt(block, BBT_BLOCK_GOOD, info);
		write_bbt(info, (u8 *)temp_buf, info->td, info->md);
		write_bbt(info, (u8 *)temp_buf, info->md, info->td);
	}

	/* mark it as good in its spare area */
	mhn_erase_one_block(info, block);

	/* write relocation table and BBT to flash if needed */
	if (memcmp(&table, &info->table, sizeof(struct nand_rel_table))){
		memcpy(&info->table, &table, sizeof(struct nand_rel_table));
		sync_rel_table(info, &(info->current_slot));
	}
}

/*
 * MOBM maintains a relocation table. It's used to replace bad blocks.
 * If block A is bad, it will use block B instead.
 * There're 127 relocated blocks. All of them reside in the bottom of NAND
 * flash. So they're reserved and can't be calculated in mtd size and chip
 * size.
 */
static int read_rel_table(struct nand_info *info)
{
	int page, pagesize, maxslot;
	int obm, valid, ret;

	obm = calc_obm_ver();
	memset(buf, 0, BUFLEN);

	page = 1 << (info->phys_erase_shift - info->page_shift);
	pagesize = 1 << info->page_shift;

	valid = 0;
	if (obm == MHN_OBM_V2) {
		/* On MOBM V2, the relocation table resides in the last page
		 * of the first block.
		 * Hardware ECC isn't necessary at this time.
		 */
		page = (1 << (info->phys_erase_shift - info->page_shift)) - 1;
		ret = mhn_read_one_page(info, buf, page, READ_RAW);
		memcpy(((unsigned char *)&(info->table)), buf,
			sizeof(struct nand_rel_table));
		if (ret == 0 && info->table.header == NAND_RELOC_HEADER)
			valid = 1;
	} else if (obm == MHN_OBM_V3) {
		/* On MOBM V3, there're several relocation tables in the first
		 * block.
		 * When new bad blocks are found, a new relocation table will
		 * be generated and written back to the first block. But the
		 * original relocation table won't be erased. Even if the new
		 * relocation table is written wrong, system can still find an
		 * old one.
		 * One page contains one slot.
		 */
		maxslot = 1 << (info->phys_erase_shift - info->page_shift);
		page = maxslot - MAX_BBT_SLOTS;
		for (; page < maxslot; page++) {
			ret = mhn_read_one_page(info, buf, page, READ_RAW);
			if (ret == 0) {
				if (((struct nand_rel_table *)buf)->header != NAND_RELOC_HEADER) {
					continue;
				} else {
					info->current_slot = maxslot - page - 1;
					printf("Slot %d Found\n", info->current_slot);
					memcpy(((unsigned char *)&(info->table)),
						buf, sizeof(struct nand_rel_table));
					valid = 1;
					break;
				}
			}
		}
	} else {
		printf("The version of MOBM isn't supported\n");
	}
	if (valid) {
		memcpy(((unsigned char *)&(info->table)), buf,
			sizeof(struct nand_rel_table));
		if (info->table.total > nand_reloc_max)
			info->table.total = nand_reloc_max;
		info->table_init = 1;
		dump_rel_table(&(info->table));
	} else {
		/* There should be a valid relocation table slot at least. */
		printf("NO VALID relocation table can be recognized\n");
		printf("CAUTION: It may cause unpredicated error\n");
		printf("Please re-initialize the NAND flash.\n");
		memset((unsigned char *)&(info->table), 0, 
				sizeof(struct nand_rel_table));
		info->table_init = 0;
		return -EINVAL;
	}
	return 0;
}

/* add the relocation entry into the relocation table
 * It's valid on MOBM V3.
 */
static int update_rel_table(struct nand_info *info, int block)
{
	u32 bad_block = 0xFFFFFFFF;
	struct nand_rel_table *table = NULL;
	int status;
	unsigned int reloc_block;
	int obm, i, entry_num = -1;
	char rel_dist[NAND_RELOC_MAX];

	if ((unsigned int)block >= (info->chipsize >> info->phys_erase_shift))
		return -EBADBLOCK;

	obm = calc_obm_ver();
	if ((obm == MHN_OBM_V3) || (obm == MHN_OBM_V2 && !info->table_init)) {
		table = &info->table;
		if (info->table_init == 0) {
			printf("Error: the initial relocation table can't be read\n");
			memset(table, 0, sizeof(struct nand_rel_table));
			table->header = NAND_RELOC_HEADER;
			info->table_init = 1;
		}

		if (table->total >= nand_reloc_max) {
			printf("Relocation table exceed max num,");
			printf("cannot relocate block 0x%x\n", block);
			return -ENOSPACE;
		}

		//identify whether the block has been relocated
		for(i = table->total - 1; i >= 0; i --) {
			if(block == table->rel[i].from)
				entry_num = i;
		}

		memset(rel_dist, 0, NAND_RELOC_MAX);
		//find the available block with the largest number in reservered area
		for (i = 0; i < table->total; i ++) {
			int _rel = (table->rel[i].to != 65535) ? table->rel[i].to : table->rel[i].from;

			rel_dist[(_rel - (info->chipsize >> info->phys_erase_shift))] = 1;
		}

		while (1) {
			/* Make sure that reloc_block is pointing to a valid block */
			for (reloc_block = nand_reloc_max - 1; (int)reloc_block >= 0; reloc_block--) {
				if (rel_dist[reloc_block] == 0) {
					break;
				}
			}

			reloc_block = reloc_block + (info->chipsize >> info->phys_erase_shift);

			if ((unsigned int)reloc_block < (info->chipsize >> info->phys_erase_shift)) {
				if (entry_num >= 0) {
					table->rel[entry_num].from = table->rel[entry_num].to;
					table->rel[entry_num].to = 65535;
				}
				return -ENOSPACE;
			}

			/* We can not use mhn_erase_one_block to replace this
			 * function directly because it's for relocation table
			 * operation. Need abstract one function to call native
			 * nand/Onenand operation directly.
			 */
			status = mhn_erase_direct(info, reloc_block, 1, (int *)&bad_block);
			if (status == 0)
				break;
			else {
				DPRINTF("status:0x%x found at erasing reloc block %d\n", reloc_block);
				/* skip it if the reloc_block is also a 
				 * bad block
				 */
				if (bad_block == (unsigned int)reloc_block) {
					table->rel[table->total].from = reloc_block;

					table->rel[table->total].to = 65535;
					table->total++;
					rel_dist[(reloc_block - (info->chipsize >> info->phys_erase_shift))] = 1;;
					continue;
				} else
					return -ECMDFAIL;
			}
		}

		/* Create the relocated block information in the table */
		//when the block is relocated before, blob should modify the original entry to new
		//relocated block and the old relocated block point to 65535. If not the situation, 
		//create a new entry
		if (entry_num != -1) {
			table->rel[table->total].from = table->rel[entry_num].to;
			table->rel[table->total].to = 65535;
			table->total++;
			table->rel[entry_num].to = reloc_block;
		} else {
			table->rel[table->total].from = block;
			table->rel[table->total].to = reloc_block;
			table->total++;
		}
	}else{
		return -ENOSPACE;
	}
	return 0;
}

/* Write the relocation table back to device, if there's room. */
static int sync_rel_table(struct nand_info *info, int *idx)
{
	int obm, start_page, len;

	if (*idx < -1) {
		DPRINTF("Wrong Slot is specified.\n");
		return -1;
	}
	obm = calc_obm_ver();
	if ((obm == MHN_OBM_V3 && (*idx + 1) >= MAX_BBT_SLOTS) ||
	    (obm == MHN_OBM_V2 && (*idx + 1) >= 1)){
		printf("Can't write relocation table to device any more.\n");
		return -1;
	}

	if (info->table_init == 0) {
		memset(&info->table, 0, sizeof(struct nand_rel_table));
		info->table.header = NAND_RELOC_HEADER;
		info->table_init = 1;
	}

	*idx += 1;
	len = 4;
	len += info->table.total << 2;
	/* write to device */
	start_page = 1 << (info->phys_erase_shift - info->page_shift);
	start_page = start_page - 1 - *idx;
	memset(buf, 0xFF, BUFLEN);
	memcpy(buf, &(info->table), len);

	DPRINTF("Write relocation table to slot %d\n", *idx);
	DPRINTF("write relocation table to page:0x%x\n", start_page);
	mhn_write_one_page(info, buf, start_page, 1);
#if 0
	/* dump it */
	memset(buf, 0, BUFLEN);
	mhn_read_one_page(info, buf, start_page, READ_OOB);
	PRINT_BUF(buf, len);
#endif

	return 0;
}

/* Find the relocated block of the bad one.
 * If it's a good block, return 0. Otherwise, return a relocated one.
 * idx points to the next relocation entry
 */
static int search_rel_block(int block, struct nand_info *info)
{
	struct nand_rel_table *table = NULL;
	int max, i, reloc_block = block;

	if (info->table_init == 0)
		return reloc_block;

	table = &info->table;
	if ((block <= 0) || (block >= (int)(info->chipsize >> info->phys_erase_shift))	
			|| (table->total == 0))
		return reloc_block;
	if (table->total > nand_reloc_max)
		max = nand_reloc_max;
	else
		max = table->total;
	for (i = max - 1; i >= 0; i --) {
		if (block == table->rel[i].from) {
			reloc_block = table->rel[i].to;
			break;
		}
	}
	return reloc_block;
}

static int is_marked_bad(struct nand_info *info, int block)
{
	int pages_shift, pagesize, ret = 0;
	unsigned int bytes;

	pages_shift = info->phys_erase_shift - info->page_shift;
	pagesize = 1 << info->page_shift;
	
	/* check the first page of each block */
	ret = mhn_read_one_page(info, buf, block << pages_shift, READ_OOB);
	if (ret) {
		DPRINTF("read page error at page 0x%x", block << pages_shift);
		DPRINTF(" when bbt is created\n");
		return -ERR;
	}
	for (bytes = 0; bytes < info->badblockbytes; bytes++) {
		if (buf[pagesize + info->badblockpos + bytes] != 0xFF) {
			return 1;
		}
	}
	/* check the second page of each block */
	ret = mhn_read_one_page(info, buf, (block << pages_shift) + 1, READ_OOB);
	if (ret) {
		DPRINTF("read page error at page 0x%x", block << pages_shift);
		DPRINTF(" when bbt is created\n");
		return -ERR;
	}
	for (bytes = 0; bytes < info->badblockbytes; bytes++) {
		if (buf[pagesize + info->badblockpos + bytes] != 0xFF) {
			return 1;
		}
	}
	return 0;
}

/* Produce relocation table on some pages of the first block. */
static int set_rel_table(struct nand_info *info)
{
	int numblocks, pages_shift, badblocks;
	int start_page, pagesize;
	int i, bytes, ret;
	struct nand_rel_table table;
	char *src = NULL, *dst = NULL;

	numblocks = info->chipsize >> info->phys_erase_shift;
	if (info->td)
		numblocks -= info->td->maxblocks;
	if (info->md)
		numblocks -= info->md->maxblocks;

	pages_shift = info->phys_erase_shift - info->page_shift;
	start_page = (int)((1 << pages_shift) - 1);
	pagesize = 1 << info->page_shift;
	badblocks = 0;

	/* copy an existed relocation for backup */
	memcpy(&table, &info->table, sizeof(struct nand_rel_table));
	/* produce a relocation table */
	for (i = 0; i < numblocks; i++) {
		ret = 0;
		/* If it's a bad block registered in relocation table, 
		 * skip it.
		 * Try to find a new bad block in flash.
		 */
		ret = is_marked_bad(info, i);
		if (ret != 0) {
			if (is_bad_block(i, info))
				continue;	

			/* it's a bad one */
			//set_bbt(i, BBT_BLOCK_BAD, info);
			mark_badblock(i, info);
		}
	}

	/* compare it with the relocation table of block 0 */
	ret = 0;
	dst = (char *)&info->table;
	src = (char *)&table;
	bytes = sizeof(struct nand_rel_table);
	for (i = 0; i < bytes; i++) {
		if (*src != *dst)
			break;
		src++;
		dst++;
	}

	if (i < bytes || info->table_init == 0) {
		/* These two table aren't same. The relocation table should
		 * be synchronized.
		 */
		ret =  sync_rel_table(info, &(info->current_slot));
	}

	return ret;
}

/* check whether a block is a bad block. */
static int is_bad_block(u32 block, struct nand_info *info)
{
	struct nand_bbt_descr *td = info->td;
	struct nand_bbt_descr *md = info->md;
	int numblocks;
	int bbt_start, bbt_end;

	numblocks = info->chipsize >> info->phys_erase_shift;
	if (td->options & NAND_BBT_LASTBLOCK) {
		bbt_start = numblocks - td->maxblocks - md->maxblocks;
		bbt_end = numblocks;
	} else {
		bbt_start = 0;
		bbt_end = td->maxblocks + md->maxblocks;
	}	

	/* If this block isn't relocated one, check it directly */
	if ((bbt_start <= block && block < bbt_end) ||
		(block >= numblocks))
		return 0;

	if (((info->bbt[block >> 2] >> ((block & 0x03) << 1)) & 0x03) != 0x00) {
		printf("block 0x%x is bad block.\n", block);
		printf("0x%x\n", info->bbt[block >> 2]);
		return 1;
	}
	return 0;
}

/* set bad block flag in memory based BBT */
static inline void set_bbt(u32 block, u8 val, struct nand_info *info)
{
	val &= 0x3;
	info->bbt[block >> 0x2] &= ~(0x3 << ((block & 0x3) << 1));
	info->bbt[block >> 0x2] |= val << ((block & 0x3) << 1);
}

/* get bad block flag from memory based BBT */
static inline void get_bbt(u32 block, u8 *pval, struct nand_info *info)
{
	*pval = (info->bbt[block >> 0x2] >> ((block & 0x3) << 1)) & 0x3;
}

/* Mark a bad block in memory based BBT and write it to flashed based one.
 * On MOBM V3, relocation table should also be updated.
 * return: 
 * 	-ERELOCATED: relocate successfully
 * 	-EBADBLOCK: marked as bad block
 * 	-ERR:  fail to mark bad block
 */
static int mark_badblock(u32 block, struct nand_info *info)
{
	int idx, pagesize, page, i;
	u8 buf[BUFLEN];

	if (block >= (info->chipsize >> info->phys_erase_shift))
		return -ERR;

	if (update_rel_table(info, block)){
		/* mark bad block in spare area */
		pagesize = 1 << info->page_shift;
		memset(buf, 0xFF, BUFLEN);
		for (i = 0; i < info->badblockbytes; i++)
			buf[pagesize + info->badblockpos + i] = 0;

		page = block << (info->phys_erase_shift - info->page_shift);
		/* doesn't modify the ECC
		 * Set the first and second page of the block as bad.
		 * Avoid to erase it before write the bad block information.
		 */
		mhn_write_one_page(info, buf, page, 0);
		mhn_write_one_page(info, buf, page + 1, 0);

		set_bbt(block, BBT_BLOCK_REV1, info);
#ifdef CONFIG_ONENAND_SUPPORT
	/* We don't create FLASH based BBT for OneNAND because
	 * onenand driver in kernel doesn't support Flash based
	 * BBT
	 */
	return -EBADBLOCK;
#endif

		if (info->td)
			write_bbt(info, buf, info->td, info->md);
		if (info->md) {
			if (info->md->options & NAND_BBT_LASTBLOCK)
				info->md->pages = info->td->pages - (info->td->maxblocks
						<< (info->phys_erase_shift - info->page_shift));
			else
				info->md->pages = info->td->pages + (info->td->maxblocks
						<< (info->phys_erase_shift - info->page_shift));
			write_bbt(info, buf, info->md, info->td);
		}

		return -EBADBLOCK;
	}else{
		return -ERELOCATED;
	}
}

static int write_bbt(struct nand_info *info, u8 *buf,
			struct nand_bbt_descr *td, struct nand_bbt_descr *md)
{
	int i, j, ret, time = 0;
	int startblock = 0, dir = 0, page, numblocks, block = 0, blocks_in_page;
	int pagesize;
	u8 msk[4];
	u8 rcode = td->reserved_block_code;

	if (!rcode)
		rcode = 0xff;
	numblocks = info->chipsize >> info->phys_erase_shift;
	
	/* There was already a version of the table, reuse the page 
	 * This applies for absolute placement too, as we have the 
	 * page nr. in td->pages.
	 */
	if (td->pages != -1) {
		page = td->pages;
		goto write;
	}	

	/* Automatic placement of the bad block table */
	/* Search direction top -> down ? */
	if (td->options & NAND_BBT_LASTBLOCK) {
		startblock = numblocks - 1;
		dir = -1;
	} else {
		startblock = 0;
		dir = 1;
	}	

find_block:
	for (i = 0; i < td->maxblocks; i++) {
		u8 val;

		block = startblock + dir * i;
		get_bbt(block, &val, info);
		/* Check, if the block is bad */
		switch (val) {
			case BBT_BLOCK_BAD:
			case BBT_BLOCK_REV1:
				continue;
		}
		page = block << (info->phys_erase_shift - info->page_shift);
		/* Check, if the block is used by the mirror table */
		//printf("md->pages 0x%x\n", md->pages);
		if (!md || md->pages != page) {
			ret = mhn_erase_one_block(info, block);	
			if (ret) {
				if (ret == -EBADBLOCK) {
					continue;
				} else if (ret == -ERELOCATED){
					goto write;
				}
				return ret;
			}
			goto write;
		}
	}
	printf("No space left to write bad block table\n");
	return -1;
write:

	/* Set up shift count and masks for the flash table */
	msk[0] = 0x00; 
	msk[1] = 0x01; 
	msk[2] = ~rcode; 
	msk[3] = 0x03; 
		
	/* Calc length */
	pagesize = 1 << info->page_shift;
	blocks_in_page = pagesize << 2;
	
	ret = mhn_erase_one_block(info, page >> (info->phys_erase_shift - info->page_shift));	
	while (numblocks) {
		u8 *bbt = &(info->bbt[pagesize * time]);
	
		/* Preset the buffer with 0xff */
		memset(buf, 0xff, pagesize + info->oobsize);
		if (time == 0) {
			/* Pattern is located in oob area of first page */
			for (i = 0; i < 4; i++) {
				DPRINTF("%x ", td->pattern[i]);
			}
			DPRINTF("\n");
			memcpy(&buf[pagesize + td->offs], td->pattern, td->len);
			buf[pagesize + td->veroffs] = td->version;
		}
		//DPRINTF("write bbt :%x\n", time);
		/* walk through the memory table */
		for (i = 0; (i < numblocks) && (i < blocks_in_page); ) {
			u8 dat = bbt[i >> 2];
			
			for (j = 0; j < 4; j++, i++) {
				buf[i >> 2] &= ~(msk[dat & 0x03] << ((i & 0x3) << 1));
				dat >>= 2;
			}
		}
		//PRINT_BUF(buf, BUFLEN);
		ret = mhn_write_one_page(info, buf, page + time, 1);
		if (ret) {
			DPRINTF("write bbt error: ret:%d\n", ret);
			if (ret == -EBADBLOCK) {
				set_bbt(block, BBT_BLOCK_REV1, info);
				goto find_block;
			}
			return ret;
		}
		time++;	
		numblocks -= blocks_in_page;
		if (numblocks < 0)
			numblocks = 0;
	}

	td->pages = page;
	DPRINTF("Bad block table written to page 0x%x, version %d\n", td->pages, td->version);
	return 0;
}

static int check_pattern(u8 *buf, struct nand_bbt_descr *td)
{
	int i;
	u8 *p = buf;
	for (i = 0; i < td->len; i++) {
		//printf("buf:%x pattern:%x\n", p[td->offs + i], td->pattern[i]);
		if (p[td->offs + i] != td->pattern[i])
			return -1;
	}
	return 0;
}

static int read_bbt(struct nand_info *info, u8 *buf, struct nand_bbt_descr *td)
{
	u8 dat, tmp;
	int i = 0, j, block = 0, from = 0, ret;
	int pagesize, len;
	int totlen;	/* BBT length */

	/* Read the primary version, if available */	
	ret = mhn_read_one_page(info, buf, td->pages, READ_OOB);
	if (ret) {
		DPRINTF("read bbt error\n");
		return ret;
	}
	pagesize = 1 << info->page_shift;
#if 0
	PRINT_BUF(buf + pagesize + td->offs, td->len);
	PRINT_BUF(td->pattern, td->len);
#endif

	for (i = 0; i < info->badblockbytes; i++) {
		if (buf[pagesize + info->badblockpos + i] != 0xFF) {
			DPRINTF("skip this bad block\n");
			return -1;
		}
	}
	if (!check_pattern(&(buf[pagesize]), td)) {
		if (td->options & NAND_BBT_VERSION) 
			td->version = buf[pagesize + td->veroffs];
	} else {
		DPRINTF("pattern isn't correct\n");
		return -1;
	}
	
	DPRINTF("BBT found at page 0x%x, version %d\n", td->pages, td->version);

       	totlen = (info->chipsize >> info->phys_erase_shift) >> 2;
	
	while (totlen) {
		len = (totlen < pagesize) ? totlen : pagesize;
		
		from++;
		
		for (i = 0; i < len; i++) {
			dat = buf[i];
			for (j = 0; j < 8; j += 2, block++) {
				tmp = (dat >> j) & 0x3;
				if (tmp == 0x3)
					continue;
				if (td->reserved_block_code && (tmp 
						== td->reserved_block_code)) {
					set_bbt(block, BBT_BLOCK_REV2, info);
					continue;
				}
				/* Factory marked bad or worn out ? */	
				if (tmp == 0)
					set_bbt(block, BBT_BLOCK_BAD, info);
				else
					set_bbt(block, BBT_BLOCK_REV1, info);
				if (block > (info->chipsize >> info->phys_erase_shift))
					return 0;
			}	
		}
		totlen -= len;
		if (totlen) {
			memset(buf, 0, BUFLEN);
			ret = mhn_read_one_page(info, buf, td->pages + from, READ_OOB);
			if (ret) {	
				memset(info->bbt, 0x0, BBT_LEN);
				return ret;
			}
		}
	}
	return 0;
}

/* create a memory based BBT */
static int create_bbt(struct nand_info *info, u8 * buf)
{
	int i, ret;
	int pages_shift, numblocks, blocks_in_page, pagesize;
	unsigned int bytes = 0;

	memset(info->bbt, 0, BBT_LEN);
	pages_shift = info->phys_erase_shift - info->page_shift;
	numblocks = info->chipsize >> info->phys_erase_shift;
	blocks_in_page = numblocks >> 2;
	pagesize = 1 << info->page_shift;

	if (info->td)
		numblocks -= info->td->maxblocks;
	if (info->md)
		numblocks -= info->md->maxblocks;

	for (i = 0; i < numblocks; i++) {
		ret = 0;
		/* check the first and second page */
		ret = mhn_read_one_page(info, buf, i << pages_shift, READ_OOB);
		if (ret) {
			DPRINTF("read page error at page %d when bbt is created\n",
					i << pages_shift);
		}
		for (bytes = 0; bytes < info->badblockbytes; bytes++) {
			if (buf[pagesize + info->badblockpos + bytes]
					!= 0xFF) {
				set_bbt(i, BBT_BLOCK_REV1, info);
				DPRINTF("bad block %d\n", i);
				ret = 1;
				break;
			}
		}
		if (ret)
			continue;
		ret = mhn_read_one_page(info, buf, (i << pages_shift) + 1, READ_OOB);
		if (ret) {
			DPRINTF("read page error at page %d when bbt is created\n",
					i << pages_shift);
		}

		for (bytes = 0; bytes < info->badblockbytes; bytes++) {
			if (buf[(1 << info->page_shift) + info->badblockpos
					+ bytes] != 0xFF) {
				set_bbt(i, BBT_BLOCK_REV1, info);
				DPRINTF("bad block %d, page:0x%x\n", i, i << pages_shift);
				break;
			}
		}
	}

	//PRINT_BUF(info->bbt, BBT_LEN);

	return 0;
}

/* search the BBT on the device */
static int search_bbt(struct nand_info *info, u8 *buf,
			struct nand_bbt_descr *td, int bbt_blocks)
{
	int i, totlen, pagesize;
	int startblock, block, dir;
	u8 *p = NULL;

	/* Search direction top -> down ? */
	if (td->options & NAND_BBT_LASTBLOCK) {
		startblock = (info->chipsize >> info->phys_erase_shift) - 1
				- bbt_blocks;
		dir = -1;
	} else {
		startblock = bbt_blocks;	
		dir = 1;
	}	
	
	p = info->bbt;
       	totlen = (info->chipsize >> info->phys_erase_shift) >> 2;
	pagesize = 1 << info->page_shift;
	/* Scan the maximum number of blocks */
	for (i = 0; i < td->maxblocks; i++) {
		block = startblock + dir * i;
		td->version = 0;	
		td->pages = block << (info->phys_erase_shift
				- info->page_shift);

		if (is_bad_block(block, info)) {
			DPRINTF("BBT is on a bad block\n");
			return -1;
		}
		if (read_bbt(info, buf, td)) {
			memset(buf, 0, BBT_LEN);
			td->pages = -1;
			return -1;
		} else {
			totlen = totlen - pagesize;
			/* BBT is loaded fully */
			if (totlen <= 0)
				break;
			/* copy BBT from buf to bbt[] */
			DPRINTF("copy time: %d\n", i);
			PRINT_BUF(buf, 1 << info->page_shift);
			memcpy(p, buf, 1 << info->page_shift);
			p += 1 << info->page_shift;
		}
	}
	//PRINT_BUF(info->bbt, (1 << info->page_shift) << 1);
	return 0;
}

static int check_create(struct nand_info *info, u8 *buf, struct nand_bbt_descr *td,
		struct nand_bbt_descr *md)
{
	int ret;
	int writeops = 0;
	struct nand_bbt_descr *rd, *rd2;

	
	rd = NULL;
	rd2 = NULL;
	/* Mirrored table avilable ? */
	if (md) {
		if (td->pages == -1 && md->pages == -1) {
			writeops = 0x03;
			goto create;
		}

		if (td->pages == -1) {
			rd = md;				
			td->version = md->version;
			writeops = 1;
			goto writecheck;
		}

		if (md->pages == -1) {
			rd = td;
			md->version = td->version;
			writeops = 2;
			goto writecheck;
		}

		if (td->version == md->version) {
			rd = td;
			if (!(td->options & NAND_BBT_VERSION))
				rd2 = md;
			goto writecheck;
		}	

		if (((td->version - md->version)) > 0) {
			rd = td;
			md->version = td->version;
			writeops = 2;
		}
		else {
			rd = md;
			td->version = md->version;
			writeops = 1;
		}

		goto writecheck;

	} 
	else {
		if (td->pages == -1) {
			writeops = 0x01;
			goto create;
		}
		rd = td;
		goto writecheck;
	}
create:
	/* Create the table in memory by scanning the chip(s) */
	DPRINTF("writeops=%d\n", writeops);
	create_bbt(info, buf);
#ifdef	CONFIG_ONENAND_SUPPORT
	/* We don't create FLASH based BBT for Onenand because
	 * Onenand driver in kernel doesn't support Flash based
	 * BBT
	 */
//	return 0;
#endif
		
	td->version = 1;
	if (md)
		md->version = 1;	
writecheck:	
	DPRINTF("writeops=%d\n", writeops);
	/* Write the bad block table to the device ? */
	if ((writeops & 0x01)) {
		ret = write_bbt(info, buf, td, md);
		if (ret < 0)
			return ret;
	}

	/* Write the mirror bad block table to the device ? */
	if ((writeops & 0x02) && md) {
		if (md->options & NAND_BBT_LASTBLOCK)
			md->pages = td->pages - (td->maxblocks <<
				(info->phys_erase_shift - info->page_shift));
		else
			md->pages = td->pages + (td->maxblocks <<
				(info->phys_erase_shift - info->page_shift));
		ret = write_bbt(info, buf, md, td);
		if (ret < 0)
			return ret;
	}

	return 0;	
}

static void scan_nand(struct nand_info *info)
{
	int len, bbt_blocks = 0;
	unsigned int i = 0;
	int ret;
	int pagesize, start_page;
	//u8 buf[BUFLEN];		/* contains pagesize + oobsize */

	ret = read_rel_table(info);
	if (ret) {
		printf("can't find relocation table!!!! SHOULD use the flash tools recreate one!!!\n\n");
		while(1);
	} else
		printf("get relocation table\n");

	memset(buf, 0, BUFLEN);
	pagesize = 1 << info->page_shift;
	/* the last page of the first block */
	start_page = (1 << (info->phys_erase_shift - info->page_shift)) - 1;
	
	/* for bbt, we have 2 bits for one block */
	len = (info->chipsize >> info->phys_erase_shift) >> 1;
	len = (len > BBT_LEN) ? BBT_LEN : len;

	memset(&(info->bbt), 0x00, len);
	bbt_blocks = 0;
	if (info->td) {
		ret = search_bbt(info, buf, info->td, bbt_blocks);
		if (ret == 0)
			printf("Found Main Bad block table at address 0x%08x, version 0x%02X\n", 
				info->td->pages * pagesize, 
				info->td->version);
		else
			printf("Main Bad block table is missed\n");
		bbt_blocks += info->td->maxblocks;
	}	
	if (info->md) {
		ret = search_bbt(info, buf, info->md, bbt_blocks);
		if (ret == 0)
			printf("Found Mirror Bad block table at address 0x%08x, version 0x%02X\n", 
				info->md->pages * pagesize, 
				info->md->version);
		else
			printf("Mirror Bad block table is missed\n");
		bbt_blocks += info->md->maxblocks;
	}

	DPRINTF("td pages:0x%x md pages:0x%x\n", info->td->pages, info->md->pages);
	/* If necessary, create BBT and write it back to device */
	check_create(info, buf, info->td, info->md);

	for (i = 0; i < (info->chipsize >> info->phys_erase_shift); i++) {
		u8 val;
		get_bbt(i, &val, info);
		switch(val) {
		case 0x1:
			printf("Block #%d is bad due to worn out\n", i);
			break;
		case 0x3:
			printf("Block #%d is bad due to factory  mark\n", i);
			break;
		}
	}

}

/* 
   read from addr at flash, the read length is len. Because there may be 
   bad blocks in the flash, the actual read length may less than len.
   We only read the page.  
*/
static int monahans_nand_read(u8 *buf, u32 addr, u32 len, int option, struct nand_info *info)
{
	u32 block = -1;
	/* start is the beginning page */
	u32 start, pages;
	int pagesize, obm, i = 0;
	u8 * buffer = buf;
	int ret;

	if ((option != READ_RAW) && (addr > info->chipsize || (addr + len) > info->chipsize))
		return -1;

	obm = calc_obm_ver();
	pagesize = 1 << info->page_shift;
	if (addr & (pagesize - 1)) {
		addr &= ~(pagesize - 1);
		printf("address not align with page size, change to 0x%x\n", addr);
	}
	if (len & (pagesize - 1)) {
		len &= ~(pagesize - 1);
		printf("length not align with page size, change to 0x%x\n", len);
	}	
	printf("Read flash from 0x%x, length 0x%x\n", addr, len);
	start = addr >> info->page_shift;
	pages = len >> info->page_shift;
	while ((u32)i < pages) {
		printf(".");
		ret = mhn_read_one_page(info, buffer, start, option);
		if (ret){
			if (ret == -EBADBLOCK){
				block = start >> (info->phys_erase_shift
						- info->page_shift);
				start = (block + 1) << (info->phys_erase_shift
						- info->page_shift);
				continue;
			}else{
				return (buffer - buf);
			}
		}
		buffer += pagesize;
		if (option == READ_OOB)
			buffer += info->oobsize;
		i++;
		start++;
	}
	printf("Done\n");
	return (buffer - buf);
}

int monahans_nand_erase(u32 addr, u32 len, WRITE_FUNC func,
			struct nand_info *info)
{
//	u32 block = -1;
	/* start is the beginning block */
	int start, blocks, blocksize;
	int i = 0, ret, pagesize;
	int pages_per_block = 1 << (info->phys_erase_shift - info->page_shift);
	/* FIXME: the buf size should be info->pagesize + info->oobsize */
	u8 temp_buf[BUFLEN];
	int new_rel_num = 0;
	int max_block_num;

	blocksize = 1 << info->phys_erase_shift;
	pagesize = 1 << info->page_shift;

	printf("Erase flash from 0x%x, length 0x%x\n", addr, len);
	DPRINTF("original len is 0x%x, blocksize is %x ,~block is %x\n", len, blocksize, ~(blocksize- 1));
	len = (addr - (addr & ~(blocksize - 1)) + len + blocksize - 1) & ~(blocksize - 1);
	addr = (addr) & ~(blocksize - 1);
	DPRINTF("Erase flash address is aligned to 0x%x, ", addr);
	DPRINTF("length is aligned to 0x%x\n", len);
	start = addr >> info->phys_erase_shift;
	blocks = len >> info->phys_erase_shift;
	max_block_num = info->chipsize >> info->phys_erase_shift;
	
	if (addr == 0)
		printf("Warnning!!!the blob attempt to earse the BLOCK 0!!!factory marked BBT would lost!!!\n\n\n");

	while (i < blocks && i < max_block_num) {
		if (is_bad_block(start, info)) {
			//update_rel_table(start, info);
			printf("\nSkip for bad block #%d\n", start);
			goto next;
		}
		printf(".");

		ret = mhn_erase_one_block(info, start);
		if (ret) {
			if (ret == -EBADBLOCK) {
				goto next;
			} else if (ret == -ERELOCATED){
				new_rel_num++;
			} else {
				return ret;
			}
		}
		/* we may need write OOB */
		memset(temp_buf, 0xFF, pagesize + info->oobsize);

		switch(func) {
			case ZIMAGE:
			case YAFFS:
				break;
			case JFFS2:
				/* yfw. auto spare area layout */
				if (pagesize == 512)
					memcpy(temp_buf + pagesize,
						jffs2_sb_oob, info->oobsize);
				else
					memcpy(temp_buf + pagesize, 
						jffs2_lb_oob, info->oobsize);

				ret = mhn_write_one_page(info, temp_buf, 
						start * pages_per_block, 0);
				if (ret) {
					if (ret == -EBADBLOCK) {
						goto next;
					} else if (ret == -ERELOCATED){
						new_rel_num++;
					}else{
						return ret;
					}
				}
				break;	
			case NULLFUNC:
				break;
		}
		i++;
next:
		start++;
	}
	
	if (new_rel_num || info->table_init == 0)
		ret =  sync_rel_table(info, &(info->current_slot));
	else
		ret = 0;	

	printf("Done\n");

	return ret;
}

/* 
   write from buf to addr of flash. The length of data for write in buf is len.
   In the blob, func defines what is writen to nand:
	zImage:		The linux kernel. OOB is not cared.
	jffs2 image:	The jffs2 file system image. OOB should be writen.
	yaffs image:	The yaffs file system image. OOB shoudl be writen.
   Because there may be bad block in flash, the actual length for writing in
   flash may be longer than len. 

   NOTES: the size of buf will be larger than len when we try to burn YAFFS image.
*/
static int monahans_nand_write(u8 *buf, u32 addr, u32 len, WRITE_FUNC func,
			struct nand_info *info)
{
	u32 block = -1, rel_block, start;
	int pages, pagesize;
	int i = 0, ret = 0, obm;
	int pages_shift;
	u8 *buffer = buf;
	u8 temp_buf[BUFLEN];
	int maxslot = 0, minslot = 0, rel_table_len, flash_offset, rel_table_buf;
	unsigned int rel_table_off;
	int new_rel_num = 0, blank_size;
	struct nand_bbt_descr *td = info->td;
	struct nand_bbt_descr *md = info->md;
	unsigned int numblocks;
	unsigned int bbt_start, bbt_end;

	if (addr > info->chipsize || (addr + len) > info->chipsize){
		printf("flash address range: 0x%08x~0x%08x\n", 0, info->chipsize);
		return -1;
	}

	pages_shift = info->phys_erase_shift - info->page_shift;
	pagesize = 1 << info->page_shift;
	/* addr aligns to block size for erasing */
	printf("Write flash from 0x%x, length 0x%x\n", addr, len);
	len = (addr - (addr & ~(pagesize - 1)) + len + pagesize - 1) & ~(pagesize - 1);
	addr = (addr) & ~(pagesize - 1);

	/* backup relocation table if block 0 will be erased */
	if (addr >= (u32)(1 << info->phys_erase_shift))
		goto erase_blocks;

	obm = calc_obm_ver();
	if (obm == MHN_OBM_V2) {
		/* On MOBM V2, the relocation table resides in the last page
		 * of the first block.
		 * Hardware ECC isn't necessary at this time.
		 */
		maxslot = 1 << (info->phys_erase_shift - info->page_shift);
		minslot = maxslot - 1;
	} else if (obm == MHN_OBM_V3) {
		/* On MOBM V3, there're several relocation tables in the first
		 * block.
		 * One page contains one slot.
		 */
		maxslot = 1 << (info->phys_erase_shift - info->page_shift);
		minslot = maxslot - MAX_BBT_SLOTS;
	}

	rel_table_off = minslot << info->page_shift;
	rel_table_len = (maxslot - minslot) << info->page_shift;
	rel_table_buf = (u32)buffer - addr + rel_table_off;
	flash_offset = rel_table_off;
	while(rel_table_len > 0){
		memset(temp_buf, 0xff, pagesize);
		ret = monahans_nand_read((void*)temp_buf , flash_offset,
				pagesize, READ_NORM, &monahans_nand);
		if (ret != pagesize){
			printf("relocation table backup failed!\n");
			return -1;
		}
		memcpy((void *)rel_table_buf, (void*)temp_buf, pagesize);
		rel_table_len -= pagesize;
		flash_offset += pagesize;
		rel_table_buf += pagesize;
	}
			
	if (addr > rel_table_off){
		len = len + (addr - rel_table_off);
		addr = rel_table_off;
	}
	if ((addr + len) < (u32)(1 << info->phys_erase_shift))
		len = (1 << info->phys_erase_shift) - addr;
erase_blocks:
	/* first erase the needed block */
	monahans_nand_erase(addr, len, func, info);

	/* write data to pages */
	start = addr >> info->page_shift;
	pages = len >> info->page_shift;
	DPRINTF("Write flash address is aligned to 0x%x,", addr);
        DPRINTF(" length is aligned to 0x%x\n", len);
	//printf("Write flash from 0x%x, length 0x%x\n", addr, len);
	/* first erase it for writing */
	numblocks = info->chipsize >> info->phys_erase_shift;
	if (td->options & NAND_BBT_LASTBLOCK) {
		bbt_start = numblocks - td->maxblocks - md->maxblocks;
		bbt_end = numblocks;
	} else {
		bbt_start = 0;
		bbt_end = td->maxblocks + md->maxblocks;
	}	

	/* If we are writing yaffs2 image, we need
	 * consider oob data into account
	 */
	if (YAFFS == func)
		blank_size = pagesize + info->oobsize;
	else
		blank_size = pagesize;

	i = 0;
	while (i < pages) {
		if (block != start >> pages_shift) {
			block = start >> pages_shift;
			if (is_bad_block(block, info)) {
				start = (block + 1) << pages_shift;
				continue;
			}
			printf(".");
		}

		if ( bbt_start <= block && block < bbt_end){
			start = (block + 1) << pages_shift;
			continue;
		}

		if (block >= numblocks){
			printf("flash is full, block %d numblocks %d!\n", block, numblocks);
			ret = -1;
			goto sync;
		}
#if 0 //if image file was filled with 0xff, we need comment this 
		if (is_buf_blank(buffer, blank_size)) {
			if ( YAFFS != func )	//YAFFS need write spare area whether the buf of page is blank.
				goto next;
		}
#endif		
		rel_block = search_rel_block(block, info);

		memcpy(temp_buf, buffer, pagesize);
		if (YAFFS == func) {
			/* yfw spare area auto layout */
			memset(temp_buf + pagesize, 0xFF, info->oobsize);
			memcpy(temp_buf + pagesize + yaffs_oob.offset, buffer + pagesize, yaffs_oob.size);
		} else {
			memset(temp_buf + pagesize, 0xFF, info->oobsize);
		}
		ret = mhn_write_one_page(info, temp_buf, start, 1);
		if (ret) {
			if (block == 0){
				printf("WARNING: failed to write block 0,"
						"flash might be broken\n");
				return -1;
			}
			if (ret == -EBADBLOCK || ret == -ERELOCATED) {
				if (ret == -EBADBLOCK)
					start = (block + 1) << pages_shift;
				else {
					start = block << pages_shift;
				}
				i -= ((i & (1 << pages_shift)) - 1);
				buffer -= pagesize*((i & (1 << pages_shift)) - 1);
				if (buffer < buf){
					start = addr >> info->page_shift;
					buffer = buf;
					i = 0;
				}
				continue;
			}else{
				ret = 0;
				goto sync;
			}
		}

		ret = mhn_read_one_page(info, temp_buf, start, READ_OOB);
		if (ret || memcmp(buffer, temp_buf, pagesize)) {
			ret = mark_badblock(block, info);
			if(ret == -EBADBLOCK){
				start = (block + 1) << pages_shift;
				DPRINTF("block %d is bad,use next block!\n", block);
			} else if (ret == -ERELOCATED){
				start = block << pages_shift;
				new_rel_num++;
				if (rel_block != block)
					DPRINTF("block %d which is already relocated to %d is bad, use another relocated block!\n", block, rel_block);
				else {
					DPRINTF("block %d is bad,use relocated block!\n", block);
				}

			}else{
				ret = 0;
				goto sync;
			}
			i -= ((i & (1 << pages_shift)) - 1);
			buffer -= pagesize*((i & (1 << pages_shift)) - 1);
			continue;
		}

next:
		if (YAFFS == func) {
			/* The step for yaffs image is pagesize + oobsize */
			buffer += pagesize + info->oobsize;
		} else {
			/* The step for kernel/jffs2 image is pagesize */
			buffer += pagesize;
		}
		start++;
		i++;
	}

sync:
	if (new_rel_num || info->table_init == 0){
		int _tmp;

		printf("\nwrite the rel table\n");
		_tmp = sync_rel_table(info, &(info->current_slot));
		if (ret == 0)
			ret = _tmp;
	}

	if (i >= pages)
		printf("Done\n");

	if (ret)
		return -1;
	else
		return (buffer - buf);
}

void init_monahans_nand_driver(void)
{
	nand_driver_init();
}

__initlist(init_monahans_nand_driver, INIT_LEVEL_OTHER_HARDWARE);


static int cmd_nand_read(int argc, char *argv[] )
{
	int ret;
	int option;
	u32 ram_addr, flash_ofs, len;

	if (argc < 4) {
		ret = -1;
		goto done;
	}
	if (!strncmp("-o", argv[1], 2)) {
		option = READ_OOB;
		strtou32( argv[2], &ram_addr);
        	strtou32( argv[3], &flash_ofs);
		strtou32( argv[4], &len);
	}else if (!strncmp("-r", argv[1], 2)) {
		option = READ_RAW;
		strtou32( argv[2], &ram_addr);
        	strtou32( argv[3], &flash_ofs);
		strtou32( argv[4], &len);
	}else {
		option = READ_NORM;
		strtou32( argv[1], &ram_addr);
        	strtou32( argv[2], &flash_ofs);
		strtou32( argv[3], &len);
	}

	printf("Read 0x%x length data from flash: 0x%x to RAM address: 0x%x\n",
			len, flash_ofs, ram_addr);
	ret = monahans_nand_read((u8 *)ram_addr, flash_ofs, len, option,
			&monahans_nand);
	print_buf(ram_addr, len + monahans_nand.oobsize);

done:
	return ret; 
} 

static int cmd_nkernel(int __attribute__((unused)) argc,
			char __attribute__((unused)) *argv[] )
{
	int __attribute__((unused)) obm, ret;
	int option = READ_NORM;
	u32 ram_addr = KERNEL_RAM_BASE;
	u32 flash_ofs;
	u32 len = KERNEL_FLASH_LEN;

#if defined(CONFIG_CPU_TAVOR_P) || defined(CONFIG_CPU_TAVOR_PV) || defined(CONFIG_CPU_MONAHANS_LV)
#if defined(BENZ3G512)||defined(BEACON)
	if( PU_REASON_REBOOT_RECOVERY == get_powerup_reason() ){
		flash_ofs = RECOVERY_KERNEL_FLASH_BASE;
		len = RECOVERY_KERNEL_FLASH_LEN;
	}else if( PU_REASON_SDL_FLASHMODE == get_powerup_reason() ){
		flash_ofs = RECOVERY_KERNEL_FLASH_BASE;
		len = RECOVERY_KERNEL_FLASH_LEN;
	}else{
		flash_ofs = KERNEL_FLASH_BASE;
	}
#else
	flash_ofs = KERNEL_FLASH_BASE;
#endif
#else
	obm = calc_obm_ver();
	if (obm == MHN_OBM_V3)
		flash_ofs = 0x60000;
	else
		flash_ofs = 0x40000;
#endif

	printf("load kernel\n");
	ret = monahans_nand_read((u8 *)ram_addr, flash_ofs, len, option,
					&monahans_nand);
done:
	return ret; 
} 

static int cmd_nftmkernel(int __attribute__((unused)) argc,
			char __attribute__((unused)) *argv[] )
{
	int obm, ret;
	int option = READ_NORM;
	u32 ram_addr = KERNEL_RAM_BASE;
#if defined(FTM_KERNEL_FLASH_BASE) && defined(FTM_KERNEL_FLASH_LEN)
	u32 flash_ofs = FTM_KERNEL_FLASH_BASE;
	u32 len = FTM_KERNEL_FLASH_LEN;
#else
	u32 flash_ofs = KERNEL_FLASH_BASE;
	u32 len = KERNEL_FLASH_LEN;
#endif
	blob_status.cmdline[0] = '\0';
	printf("load FTM kernel\n");
	if (3 == argc){
		if ( (!is_digital(argv[1]))||(!is_digital(argv[2])) ){
			printf("invalid parameter\n");
			ret = 0;
			goto done;
		}
		strtou32( argv[1], &flash_ofs);
		strtou32( argv[2], &len);
		printf("load base = 0x%x, load len = 0x%x\n", flash_ofs, len);
		ret = monahans_nand_read((u8 *)ram_addr, flash_ofs, len, option,
						&monahans_nand);
	}
	else{
		ret = monahans_nand_read((u8 *)ram_addr, flash_ofs, len, option,
						&monahans_nand);
	}
done:
	return ret; 
} 

#if defined(CHERRY) || defined(BSTD)
static int cmd_nrmkernel(int __attribute__((unused)) argc,
			char __attribute__((unused)) *argv[] )
{
	int __attribute__((unused)) obm, ret;
	int option = READ_NORM;
	u32 ram_addr = KERNEL_RAM_BASE;
	u32 flash_ofs = RECOVERY_KERNEL_FLASH_BASE;
	u32 len = RECOVERY_KERNEL_FLASH_LEN;
	
	blob_status.cmdline[0] = '\0';

	printf("load recovery kernel\n");
	ret = monahans_nand_read((u8 *)ram_addr, flash_ofs, len, option,
					&monahans_nand);

done:
	return ret; 
} 
#endif




static int cmd_nramdisk(int __attribute__((unused)) argc,
			char __attribute__((unused)) *argv[] )
{
	int ret;
	int option = READ_NORM;
	u32 ram_addr = RAMDISK_RAM_BASE;	//0x80600000;
	u32 flash_ofs = RAMDISK_FLASH_BASE;	//0x240000;
	u32 len = RAMDISK_FLASH_LEN;		//0x1000000;

	int ramdisk_len;

	if( PU_REASON_REBOOT_RECOVERY == get_powerup_reason() ){
#if defined(BENZ3G512)||defined(BEACON)
		flash_ofs = RECOVERY_KERNEL_FLASH_BASE + RECOVERY_KERNEL_FLASH_LEN;
#else
		flash_ofs += RAMDISK_FLASH_LEN;
#endif
	}
#if defined(BEACON)
	else if( PU_REASON_SDL_FLASHMODE == get_powerup_reason() ){
		flash_ofs = RECOVERY_KERNEL_FLASH_BASE + 
			RECOVERY_KERNEL_FLASH_LEN + RAMDISK_FLASH_LEN;
	}
#endif

#if defined(CHERRY)||defined(BSTD)
	else if (RECOVER_KERNEL_BOOTUP_REASON & get_boot_info())
		flash_ofs = RECOVERY_KERNEL_FLASH_BASE + RECOVERY_KERNEL_FLASH_LEN;
#endif

	printf("load ramdisk\n");
	ret = monahans_nand_read((u8 *)ram_addr, flash_ofs, len, option,
					&monahans_nand);

#if 1
	ramdisk_len = *((int*)( ram_addr + len - sizeof(ramdisk_len) ) );
	set_ramdisk_len( ramdisk_len );
#endif

done:
	return ret; 
} 

#if 1
static int cmd_nw_ramdisk(int argc, char *argv[] )
{
	int ret;
	u32 ram_addr = TFTP_RAM_START;	//0x80600000;
	u32 flash_ofs = RAMDISK_FLASH_BASE;	//0x240000;
	u32 len = RAMDISK_FLASH_LEN;		//0x1000000;

	int len_addr;

	if( argc >= 2 ){
		if( !strncmp( "-r", argv[1], 2 ) )
			flash_ofs += RAMDISK_FLASH_LEN;
	}

	set_ramdisk_len( get_tftp_received_len() );
	len_addr = ram_addr + len - sizeof(len_addr);

	*( (int*)( len_addr ) ) = get_ramdisk_len();
	printf( "ramdisk_len = %d\n", *( (int*)( len_addr ) ) );

	ret = monahans_nand_write( (u8 *)ram_addr, flash_ofs, len, ZIMAGE, &monahans_nand);

	return ret;
}

static char nw_ramdiskhelp[] = "nw_ramdisk command\n"
"usage: nw_amdisk\n"
"Write ramdisk from DDR to flash\n";
__commandlist( cmd_nw_ramdisk, "nw_ramdisk", nw_ramdiskhelp );
#endif

int blob_nand_erase( unsigned int flash_ofs, 
		unsigned int size, char *type )
{
	int ret;
	WRITE_FUNC func;
	u32 len, blocks;
	cmd_nand_unlock(2,NULL);

	len = size;

	if (!strncmp("-j", type, 2))
		func = JFFS2;
	else if (!strncmp("-z", type, 2))
		func = ZIMAGE;
	else if (!strncmp("-y", type, 2))
		func = YAFFS;
	else {
		ret = -1;
		goto done;
	}	
		if (YAFFS == func) {
		/* If it's yaffs, we need change yaffs image length to nand block
		 * length. Actually, we don't suggest use -y option to erase the
		 * flash because it's difficult to calculate the blocks manually.
		 * by yfw.
		 */
		blocks = len / ((1 << monahans_nand.page_shift) + monahans_nand.oobsize);
		len = blocks << monahans_nand.page_shift;
	}
	printf("Erase 0x%x length data from flash: 0x%x\n", len, flash_ofs);
	ret = monahans_nand_erase(flash_ofs, len, func, &monahans_nand);

done:
	return ret; 
} 

/* Add constrain to deny yaffs on Onenand -- yfw */
static int cmd_nand_erase(int argc, char *argv[] )
{
	int ret;
	WRITE_FUNC func;
	u32 flash_ofs, len, blocks;
	cmd_nand_unlock(2,NULL);

	if (argc < 4) {
		ret = -1;
		goto done;
	}
	if (!strncmp("-j", argv[1], 2))
		func = JFFS2;
	else if (!strncmp("-z", argv[1], 2))
		func = ZIMAGE;
	else if (!strncmp("-y", argv[1], 2))
		func = YAFFS;
	else {
		ret = -1;
		goto done;
	}
        strtou32( argv[2], &flash_ofs);
	strtou32( argv[3], &len);
	if (YAFFS == func) {
		/* If it's yaffs, we need change yaffs image length to nand block
		 * length. Actually, we don't suggest use -y option to erase the
		 * flash because it's difficult to calculate the blocks manually.
		 * by yfw.
		 */
		blocks = len / ((1 << monahans_nand.page_shift) + monahans_nand.oobsize);
		len = blocks << monahans_nand.page_shift;
	}
	printf("Erase 0x%x length data from flash: 0x%x\n", len, flash_ofs);
	ret = monahans_nand_erase(flash_ofs, len, func, &monahans_nand);

done:
	return ret; 
} 

int blob_nand_write(unsigned int ram_addr, unsigned int flash_ofs,
	   unsigned long write_size, char * type )
{
	int ret;
	WRITE_FUNC func;
	u32 blocks, len;

	len =  write_size;

	cmd_nand_unlock(2,NULL);
	
	if (!strncmp("-j", type, 2))
		func = JFFS2;
	else if (!strncmp("-z", type, 2))
		func = ZIMAGE;
	else if (!strncmp("-y", type, 2))
		func = YAFFS;
	else {
		ret = -1;
		goto done;
	}
	/* yfw */
	if (YAFFS == func) {
		/* If it's yaffs, we need change yaffs image length to nand block
		 * length. Actually, we don't suggest use -y option to erase the
		 * flash because it's difficult to calculate the blocks manually.
		 * by yfw.
		 */
		blocks = len / ((1 << monahans_nand.page_shift) + monahans_nand.oobsize);
		len = blocks << monahans_nand.page_shift;
	}
	ret = monahans_nand_write((u8 *)ram_addr, flash_ofs, len, func,
					&monahans_nand);

done:
	return ret; 
}

/* Add constrain to deny yaffs on Onenand -- yfw */
static int cmd_nand_write(int argc, char *argv[] )
{
	int ret;
	WRITE_FUNC func;
	u32 ram_addr, flash_ofs, len, blocks;

	cmd_nand_unlock(2,NULL);
	if (argc < 5) {
		ret = -1;
		goto done;
	}
	if (!strncmp("-j", argv[1], 2))
		func = JFFS2;
	else if (!strncmp("-z", argv[1], 2))
		func = ZIMAGE;
	else if (!strncmp("-y", argv[1], 2))
		func = YAFFS;
	else {
		ret = -1;
		goto done;
	}
	if (!is_digital( argv[2] ) ) {
	    printf("invalid ram address \n");
	    ret = -1;
		goto done;
	}
	if (!is_digital( argv[3] ) ) {
	    printf("invalid flash address \n");
	    ret = -1;
		goto done;
	}
	if (!is_digital( argv[4] ) ) {
    	printf("invalid size \n");
    	ret = -1;
		goto done;
	}
	strtou32( argv[2], &ram_addr);
    strtou32( argv[3], &flash_ofs);
	strtou32( argv[4], &len);
	printf("Write 0x%x length data from RAM: 0x%x to flash: 0x%x\n",
		len, ram_addr, flash_ofs);
	/* yfw */
	if (YAFFS == func) {
		/* If it's yaffs, we need change yaffs image length to nand block
		 * length. Actually, we don't suggest use -y option to erase the
		 * flash because it's difficult to calculate the blocks manually.
		 * by yfw.
		 */
		blocks = len / ((1 << monahans_nand.page_shift) + monahans_nand.oobsize);
		len = blocks << monahans_nand.page_shift;
	}
	ret = monahans_nand_write((u8 *)ram_addr, flash_ofs, len, func,
					&monahans_nand);

done:
	return ret; 
}

int cmd_nand_unlock(int __attribute__((unused)) argc,
		char __attribute__((unused)) *argv[] )
{
	struct nand_info * info = &monahans_nand;
	return unlock_nand(info);
}

static int cmd_set_badblk(int argc, char *argv[])
{
	int ret = 0;
	u32 blk_num = 0;
	struct nand_info *info = &monahans_nand;

	if (argc < 2) {
		ret = -1;
		printf("Please input valid parameters\n");
		printf("usage: set_badblk blk_num\n");
		goto done;
	}

	strtou32(argv[1], &blk_num);
	if ((blk_num <= 0) ||
	       	(blk_num >= (info->chipsize>> info->phys_erase_shift))) {
		ret = -1;
		printf("Input wrong parameter: %d\n", blk_num);
		printf("Please specify an valid block number\n");
		goto done;
	}
	set_badblock(blk_num, &monahans_nand);
done:
	return ret;
}

static int cmd_clear_badblk(int argc, char *argv[])
{
	int ret = 0;
	u32 blk_num;
	struct nand_info *info = &monahans_nand;

	if (argc < 2) {
		ret = -1;
		printf("Please input valid parameters\n");
		printf("usage: clear_badblk blk_num\n");
		goto done;
	}
	strtou32(argv[1], &blk_num);
	if ((blk_num <= 0) ||
		(blk_num >= (info->chipsize>> info->phys_erase_shift))) {
			
		ret = -1;
		printf("Please specify an valid block number\n");
		goto done;
	}

	clear_badblock(blk_num, info);

done:
	return ret;
}

static int cmd_print_bbt(int __attribute__((unused)) argc,
			char __attribute__((unused)) *argv[])
{
	struct nand_info *info = &monahans_nand;

	/* scan nand and print the relocation table and bad block table */
	scan_nand(info);

	return 0;
}

/* It will empty all relocation tables.
 * After that, please erase the whole flash to find out the badblocks
 * and generate the initial relocation table
 */
static int cmd_reset_bbt(int __attribute__((unused)) argc,
			char __attribute__((unused)) *argv[])
{
	char *buffer = (char *)BLOB_RAM_BASE;
	char *buf = (char *)BLOB_RAM_BASE;
	struct nand_info *info = &monahans_nand;
	int ret;
	int obm, maxslot = 0, minslot = 0, rel_table_off, rel_table_len;
	int start, block = 0, pages_shift, pagesize, pages;
	unsigned int i;
	char temp_buf[BUFLEN];

	printf("empty all relocation tables and bad block table,\n"
			"please erase flash to generate relocation table again!\n");

	/* reset Bad Block Table */
	for (i = 0;i < (info->chipsize >> info->phys_erase_shift);i++)
		set_bbt(i, BBT_BLOCK_GOOD, info);
	write_bbt(info, (u8 *)temp_buf, info->td, info->md);
	write_bbt(info, (u8 *)temp_buf, info->md, info->td);

	/* back up the block 0 */
	ret = monahans_nand_read(buffer, 0,
		       	(u32)(1 << info->phys_erase_shift), READ_NORM, info);
	if (ret != (1 << info->phys_erase_shift)){
		printf("block 0 backup failed!\n");
		return -1;
	}

	obm = calc_obm_ver();
	if (obm == MHN_OBM_V2) {
		/* On MOBM V2, the relocation table resides in the last page
		 * of the first block.
		 * Hardware ECC isn't necessary at this time.
		 */
		maxslot = 1 << (info->phys_erase_shift - info->page_shift);
		minslot = maxslot - 1;
	} else if (obm == MHN_OBM_V3) {
		/* On MOBM V3, there're several relocation tables in the first
		 * block.
		 * One page contains one slot.
		 */
		maxslot = 1 << (info->phys_erase_shift - info->page_shift);
		minslot = maxslot - MAX_BBT_SLOTS;
	}

	/* empty the relocation tables in backup buffer */
	rel_table_off = minslot << info->page_shift;
	rel_table_len = (maxslot - minslot) << info->page_shift;
	memset(buffer + rel_table_off, 0xff, rel_table_len);

	/* erase the block 0 */
	monahans_nand_erase(0, 1 << info->phys_erase_shift , NULLFUNC, info);

	/* write back to the block 0 */
	pages_shift = info->phys_erase_shift - info->page_shift;
	pagesize = 1 << info->page_shift;
	start = 0;
	pages = 1 << pages_shift;
	i = 0;
	while (i < pages) {
		if (block != start >> pages_shift) {
			block = start >> pages_shift;
			if (is_bad_block(block, info)) {
				start = (block + 1) << pages_shift;
				continue;
			}
			printf(".");
		}
		
		/* for some images, there are padding bytes. we do not write */
		if (is_buf_blank(buffer, pagesize)) {
			goto next;
		}
		memcpy(temp_buf, buffer, pagesize);
		memset(temp_buf + pagesize, 0xFF, info->oobsize);
		ret = mhn_write_one_page(info, temp_buf, start, 1);
		if (ret) {
			printf("WARNING: failed to write block 0 and"
					"flash might be broken\n");
			return -1;
		}
next:
		buffer += pagesize;
		start++;
		i++;
	}

	/* empty the relocation table in memory */
	memset(&info->table, 0xFF, sizeof(struct nand_rel_table));
	info->current_slot = -1;
	info->table_init = 0;

	printf("Done\n");
	return (buffer - buf);
}

/*  It will scan the flash and find out those blocks marked as bad block
 *  and then update the relocation table.
 */
static int cmd_init_bbt(int __attribute__((unused)) argc,
			char __attribute__((unused)) *argv[])
{
	struct nand_info *info = &monahans_nand;

	set_rel_table(info);
	printf("slot:%d\n", info->current_slot);
	dump_rel_table(&(info->table));
	dump_bbt(info);

	return 0;
}


static int cmd_set_resbadblk(int argc, char *argv[])
{
	u32 arg;
	int numblocks;
	struct nand_info *info = &monahans_nand;
	struct nand_rel_table *table = &info->table;
	
	if (argc < 2) {
		numblocks = info->chipsize >> info->phys_erase_shift;
		printf("total blocks: %d\n", numblocks);
		printf("dump reltable, slot:%d\n", info->current_slot);
		dump_rel_table(table);
		printf("dump end\n");
		printf("Please input valid parameters\n");
		printf("usage: nandtest bad_blk_num\n");
		return 0;
	}

	strtou32(argv[1], &arg);
	
	printf("dump reltable, slot:%d\n", info->current_slot);
	dump_rel_table(table);
	printf("dump end\n");

	printf("Add bad block %d to reltable...\n", arg);
	table->rel[table->total].from = arg;
	table->rel[table->total].to = -1;
	table->total++;

	printf("sysnc reltable...\n");
	sync_rel_table(info, &info->current_slot); 
		
	printf("dump reltable, slot:%d\n", info->current_slot);
	dump_rel_table(table);
	printf("dump end\n");
	
	return 0;
}

static int cmd_writeflash(int __attribute__((unused)) argc,
		char __attribute__((unused)) *argv[])
{
	int ret;
	WRITE_FUNC func;
	u32 ram_addr, flash_ofs, len, blocks;
	
	ram_addr = TFTP_RAM_START;

	if (1>= argc) {
		printf("writeflash -k (write kernel)\nwriteflash -b (write blob)\n");
		return 0;
	}

	if (!strncmp("-k", argv[1], 2)){
		flash_ofs = KERNEL_FLASH_BASE;
		len = KERNEL_FLASH_LEN;
	} else if (!strncmp("-b", argv[1], 2)) {
		flash_ofs = BLOB_FLASH_BASE;
		len = BLOB_FLASH_LEN;
	}else{
		printf("invalid parameter\n");
		return 0;
	}


	ret = monahans_nand_write((u8 *)ram_addr, flash_ofs, len, ZIMAGE,
					&monahans_nand);
	return 0;
}

void cmd_blob_read_ddr(int argc, char *argv[])
{
	u32 ram_addr, size;
	printf("argc=%d\n", argc);
	if (argc < 3){
		printf("parameter error! \n");
		return;
	}

	if (!is_digital( argv[1] ) ) {
	    printf("invalid ram address! \n");
	    return;
	}
	if (!is_digital( argv[2] ) ) {
	    printf("invalid read size! \n");
	    return;
	}

	strtou32( argv[1], &ram_addr);
    strtou32( argv[2], &size);

	print_buf(ram_addr, size);
	return;
}

static char blob_read_ddr_help[] = "blob read ddr data\n"
"read_ddr [ddr address] [read size]\n";
__commandlist(cmd_blob_read_ddr, "read_ddr", blob_read_ddr_help);

static char set_writeflash_help[] = "writeflash -k write kernel to flash\n"
"writeflash -b write blob to flash\n";
__commandlist( cmd_writeflash, "writeflash", set_writeflash_help );

static char nandreadhelp[] = "nandread command\n"
"usage: nandread [-o] ram_addr flash_addr length\n"
"read pages from flash(with oob) to nand flash\n";
__commandlist( cmd_nand_read, "nandread", nandreadhelp );

static char nkernelhelp[] = "nkernel command\n"
"usage: nkernel\n"
"read kernel from flash 0x40000 to DDR 0x80080000\n";
__commandlist( cmd_nkernel, "nkernel", nkernelhelp );

static char nftmkernelhelp[] = "nftmkernel command\n"
"usage: nftmkernel\n"
"read kernel from flash 0xf400000 to DDR 0x80080000\n"
"usage: nftmkernel 0xxxxxxxxx 0xxxxxxx\n"
"read kernel from flash at address 0xxxxxxxxx length is 0xxxxxxx to DDR 0x80080000\n";
__commandlist( cmd_nftmkernel, "nftmkernel", nftmkernelhelp );

#if defined(CHERRY) || defined(BSTD)
static char nrmkernelhelp[] = "nrmkernel command\n"
"usage: nrmkernel\n"
"read recovery kernel from flash 0x280000 to DDR 0x80080000\n";
__commandlist( cmd_nrmkernel, "nrmkernel", nrmkernelhelp );
#endif

static char nramdiskhelp[] = "nramdisk command\n"
"usage: nramdisk\n"
"read ramdisk from flash to DDR\n";
__commandlist( cmd_nramdisk, "nramdisk", nramdiskhelp );
 
static char nanderasehelp[] = "nanderase command\n"
"usage: nanderase [-j|-z|-y] flash_addr length\n"
"erase flash nand flash with zImage(jffs2 fs image, yaffs fs image) format\n";
__commandlist( cmd_nand_erase, "nanderase", nanderasehelp );

static char nandwritehelp[] = "nandwrite command\n"
"usage: nandwrite [-j|-z|-y] ram_addr flash_addr length\n"
"write zImage(jffs2 fs image, yaffs fs image) to nand flash\n";
__commandlist( cmd_nand_write, "nandwrite", nandwritehelp );

static char set_badblk_help[] = "set_badblk command\n"
"usage: set_badblk blk_num\n"
"mark specified block as bad block\n";
__commandlist( cmd_set_badblk, "set_badblk", set_badblk_help );

static char clear_badblk_help[] = "clear_badblk command\n"
"usage: clear_badblk blk_num\n"
"erase the bad block flag from specified block\n";
__commandlist( cmd_clear_badblk, "clear_badblk", clear_badblk_help );

static char reset_bbt_help[] = "reset_bbt command\n"
"usage: reset_bbt\n"
"reset bad block tables of nand flash\n";
__commandlist( cmd_reset_bbt, "reset_bbt", reset_bbt_help );

static char init_bbt_help[] = "init_bbt command\n"
"usage: init_bbt\n"
"initialize bad block tables of nand flash\n";
__commandlist( cmd_init_bbt, "init_bbt", init_bbt_help );

static char print_bbt_help[] = "print_bbt command\n"
"usage: print_bbt\n"
"print out the relocation table\n";
__commandlist( cmd_print_bbt, "print_bbt", print_bbt_help );

static char nandunlockhelp[] = "nandunlock command\n"
"usage: nandunlock\n"
"unlock nand to enable nandwrite and nanderase operation\n";
__commandlist( cmd_nand_unlock, "nandunlock", nandunlockhelp );

static char set_resbad_help[] = "set_resbad command\n"
"usage: set_resbad blk_num\n"
"mard specified block as bad block in resever region\n";
__commandlist( cmd_set_resbadblk, "set_resbad", set_resbad_help );

