/*
 * start-sa11x0.S: StrongARM SA11x0 specific start code
 *
 * Copyright (C) 1999 2000 2001 2002 Erik Mouw (J.A.K.Mouw@its.tudelft.nl) and
 *                     Jan-Derk Bakker (J.D.Bakker@its.tudelft.nl)
 *
 * $Id: start-sa11x0.S,v 1.7 2003/08/07 19:31:01 abz Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/*
 * Documentation:
 * [1] Intel Corporation, "Intel StrongARM SA-1100 Microprocessor
 *     Developer's Manual", April 1999
 * [2] Intel Corporation, "Intel StrongARM SA-1110 Microprocessor
 *     Advanced Developer's manual, December 1999
 */

.ident "$Id: start-sa11x0.S,v 1.7 2003/08/07 19:31:01 abz Exp $"

#ifdef HAVE_CONFIG_H
# include <blob/config.h>
#endif

#include <blob/arch.h>

.text



/* some defines to make life easier */
/* Register addresses can be found in [1] Appendix A */
IC_BASE:        .word   0x90050000
#define ICMR    0x04
	
PWR_BASE:	.word	0x90020000
#define PSSR	0x04
#define PSPR	0x08
#define PPCR	0x14
#define POSR	0x1C
	
RST_BASE:	.word	0x90030000
#define RCSR	0x04



/* The initial CPU speed. Note that the SA11x0 CPUs can be safely overclocked:
 * 190 MHz CPUs are able to run at 221 MHz, 133 MHz CPUs can do 190 Mhz.
 */
cpuspeed:	.word	CPU_SPEED




/* We start by implementing *all* exception vectors */




/* Reset vector: this initialises the machine */
.globl reset
reset:
	/* First, mask **ALL** interrupts */
	ldr	r0, IC_BASE
	mov	r1, #0x00
	str	r1, [r0, #ICMR]

	/* switch CPU to correct speed */
	ldr	r0, PWR_BASE
	ldr	r1, cpuspeed
	str	r1, [r0, #PPCR]

	/* check if this is a wake-up from sleep */
	ldr	r0, RST_BASE
	ldr	r1, [r0, #RCSR]
	and	r1, r1, #0x0f
	tst	r1, #0x08	/* check the Sleep Mode Reset bit */
	beq	real_reset	/* no, continue booting */

	/* Wait for the oscillator to stabilize */
	ldr	r0, PWR_BASE
wait_for_OOK:
	ldr	r1, [r0, #POSR]
	tst	r1, #1		/* test Oscillator OK bit */
	beq	wait_for_OOK

	/* yes, a wake-up. clear RCSR by writing a 1 (see 9.6.2.1 from [1]) */
	ldr	r0, RST_BASE
	mov	r1, #0x08
	str	r1, [r0, #RCSR]	;

	/* setup memory */
	bl	memsetup
	
	/* handle Power Manager Sleep Status Register (PSSR)
	 * see 9.5.7.5 from [1]*/
	ldr	r0, PWR_BASE
	ldr r1, [r0, #PSSR]
	/* clear PH bit, bring periperal pins out from sleep state */
	orr r1, r1, #0x10
	str r1, [r0, #PSSR]
	
	/* get the value from the PSPR and jump to it */
	ldr	r0, PWR_BASE
	ldr	r1, [r0, #PSPR]
	mov	pc, r1

	
real_reset:
	/* init LED */
	bl	ledinit

	/* setup memory */
	bl	memsetup

	/* turn off the LED. if it stays off it is an indication that
	 * we didn't make it into the C code */
	bl led_off

	/* everything is said and done over here, call normal_boot in
	 * the generic startup code to continue the boot procedure */
	bl normal_boot

	/* oops, normal_boot returns, something went wrong. signal an
	 * error to the user */
	mov	r6, #2
	b	endless_blink




/* we could choose to handle all exceptions in a nice way, but the
 * best is to treat them as errors because blob should not contain
 * errors
 */

/* Undefined instruction exception */
.globl undefined_instruction
undefined_instruction:
	mov	r6, #3
	b	endless_blink




/* SWI */
.globl software_interrupt
software_interrupt:
	/* NOTE: This is NOT an error! If you think that blob should return
	 * from software interrupts, you're plain WRONG. The source of the
	 * problem is in the kernel: you should *disable* CONFIG_ANGELBOOT
	 * simply because blob is not angel. -- Erik
	 */
	mov	r6, #4
	b	endless_blink

	
	

/* prefetch exception. shouldn't happen though we usually run with
 * i-cache enabled */
.globl prefetch_abort
prefetch_abort:
	mov	r6, #5
	b	endless_blink




/* data abort */
.globl data_abort
data_abort:
	mov	r6, #6
	b	endless_blink




/* we *should* never reach this */
.globl not_used
not_used:
	mov	r6, #7
	b	endless_blink




/* interrupt. we could handle this differently later if some kind of
 * driver in blob wants to be interrupt driven. for the time being we
 * treat it as an error.
 */
.globl irq
irq:
	mov	r6, #8
	b	endless_blink




/* FIQ. same as IRQ */
.globl fiq
fiq:
	mov	r6, #9
	b	endless_blink




/* endless loop that blinks the LED. r6 contains the number of blinks */
endless_blink:
	bl	wait
	mov	r0, r6
	bl	led_blink
	b	endless_blink
		
wait:
	/* busy wait loop*/
	mov	r5, #0x1000000
wait0:
	subs	r5, r5, #1
	bne	wait0
	mov	pc, lr
