/*
 * system3.c: PT System3 specific stuff
 *
 * Copyright (C) 2001  Erik Mouw (J.A.K.Mouw@its.tudelft.nl)
 * Copyright (C) 2001,2002  Stefan Eletzhofer 
 *                               (stefan.eletzhofer@www.eletztrick.de)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ident "$Id: system3.c,v 1.35 2003/09/04 17:39:00 russd Exp $"

/**********************************************************************
 * includes
 */

#ifdef HAVE_CONFIG_H
# include <blob/config.h>
#endif

#include <blob/main.h>
#include <blob/arch.h>
#include <blob/errno.h>
#include <blob/error.h>
#include <blob/util.h>
#include <blob/reboot.h>
#include <blob/serial.h>
#include <blob/flash.h>
#include <blob/init.h>
#include <blob/command.h>
#include <blob/uucodec.h>
#include <blob/time.h>
#include <blob/partition.h>
#include <blob/led.h>

#include <blob/arch.h>
#include <blob/proc/sa1111.h>

#include <blob/generic_io.h>

#include <blob/pcmcia.h>
#include <blob/cf.h>
#include <blob/ide.h>
#include <blob/tar.h>
#include <net/ether.h>
#include <net/net.h>

/**********************************************************************
 * defines
 */
/* more readable IMHO */
#define MEM( x )        (*((u32 *)(x)))
#define SET(reg,bit) ((reg) |= (1<<(bit)))
#define RST(reg,bit) ((reg) &= ~(1<<(bit)))

#define SYSTEM3_DEBUG 1

#if SYSTEM3_DEBUG
#	define DBG( x, args... )	if ( sys3_dbg>x ) printf( args )
#else
#	define DBG( x, args... )
#endif


/**********************************************************************
 * globals
 */

extern blob_status_t blob_status;

/**********************************************************************
 * module globals
 */

#if SYSTEM3_DEBUG
static int sys3_dbg = 1;
#else
static int sys3_dbg = 0;
#endif

/* flash descriptor for System3 flash. */
/* System 3 uses 2xINTEL e28F640 Chips */
static const flash_descriptor_t system3_flash_descriptors[] =
{
	{
		size: 2 * 128 * 1024,
		num: 64,
		lockable: 1
	},
	{
		/* NULL block */
	},
};

/* default partition table for SYSTEM3

System3 flash layout
~~~~~~~~~~~~~~~~~~~~

16Mb Flash
256k erase size

0x00000000 BLOB		256k
0x00040000 CONFIG	256k
0x00080000 KERNEL	1024k
0x00180000 INITRD	1536k
0x00300000 ROOT		6M
0x00900000 DATA		7M
0x01000000 total	16M

Note:
	- system3 partitions are named like the files they're flashed
	  with. This is to keep updating easy: Just loop through all
	  files that are on our update media and flash them on the partition
	  named like the file.

 */
static const blob_partition_t system3_default_partition_table[] = {
	{
		/* start of table */
		magic:	BLOB_DEFAULT_PART_TABLE_MAGIC,
		next:	sizeof(blob_partition_t),
		offset:	0x00000000,	/* absolute base address */
		size:	0x02000000  /* total size of the flash */
	},
	{
		/* blob itself */
		magic:		BLOB_PART_VALID_MAGIC,
		next:		sizeof(blob_partition_t),
		offset:		BLOB_FLASH_BASE,
		size:		BLOB_FLASH_LEN,
		name:		"blob",
		mem_base:	BLOB_RAM_BASE
	},
	{
		/* system config area */
		magic:	BLOB_PART_VALID_MAGIC,
		next:	sizeof(blob_partition_t),
		offset:	CONFIG_FLASH_BASE,
		size:	CONFIG_FLASH_LEN,
		name:	"config.tar",
		flags:	0
	},
	{
		/* kernel */
		magic:		BLOB_PART_VALID_MAGIC,
		next:		sizeof(blob_partition_t),
		offset: 	KERNEL_FLASH_BASE,
		size:		KERNEL_FLASH_LEN,
		name:		"zImage",
		flags:		BLOB_PART_FLAG_LOAD | BLOB_PART_FLAG_EXEC,
		mem_base:	KERNEL_RAM_BASE,
		entry_point:	KERNEL_RAM_BASE
	},
	{
		/* ramdisk */
		magic:		BLOB_PART_VALID_MAGIC,
		next:		sizeof(blob_partition_t),
		offset:		RAMDISK_FLASH_BASE,
		size:		RAMDISK_FLASH_LEN,
		name:		"initrd.gz",
		flags:		BLOB_PART_FLAG_LOAD,
		mem_base:	RAMDISK_RAM_BASE
	},
	{
		/* root (cramfs) */
		magic:		BLOB_PART_VALID_MAGIC,
		next:		sizeof(blob_partition_t),
		offset:		ROOTFS_FLASH_BASE,
		size:		ROOTFS_FLASH_LEN,
		name:		"cramfs",
		flags:		BLOB_PART_FLAG_CRAMFS,
	},
	{
		/* data (cramfs) */
		magic:		BLOB_PART_VALID_MAGIC,
		next:		sizeof(blob_partition_t),
		offset:		DATAFS_FLASH_BASE,
		size:		DATAFS_FLASH_LEN,
		name:		"data.img",
		flags:		BLOB_PART_FLAG_CRAMFS,
	},
	{
		/* xtended data (cramfs) */
		magic:		BLOB_PART_VALID_MAGIC,
		next:		sizeof(blob_partition_t),
		offset:		XDATAFS_FLASH_BASE,
		size:		XDATAFS_FLASH_LEN,
		name:		"xdata.img",
		flags:		BLOB_PART_FLAG_CRAMFS,
	},
	{
		/* last entry */
		magic:	BLOB_PART_LAST_MAGIC
	}
};

/**********************************************************************
 * static functions
 */
static char module_version[] = "$Id: system3.c,v 1.35 2003/09/04 17:39:00 russd Exp $";
static void sa1111_init();

/*********************************************************************
 * init_system3_flash - inits system 3 flash driver
 */
static void init_system3_flash_driver(void)
{
	flash_descriptors = system3_flash_descriptors;
	flash_driver = &intel32_flash_driver;
}
__initlist(init_system3_flash_driver, INIT_LEVEL_DRIVER_SELECTION);

/*********************************************************************
 * system3_init_hardware - inits system 3 LL hardware stuff
 */
static void system3_init_hardware(void)
{
	/* init on-board CPLD */
	MEM(SYSTEM3_CTRL_0) = 0x03;
	MEM(SYSTEM3_CTRL_1) = 0x00;
	MEM(SYSTEM3_CTRL_2) = 0x00;

	/* initialize GPIO states */
	GPDR = 0x0f35a7fc;
	GPSR |= GPIO_GPIO15 | GPIO_GPIO20; /* 15: core fast, 20: SCL */
	GPCR |= 0x070527fc;

	/* activate SYSCLCK for keyboard controller */
	RST( MEM(TUCR), 31 );
	GAFR |= GPIO_32_768kHz;
	GPDR |= GPIO_32_768kHz;

	/* 3.6864 MHz SYSCLK */
	TUCR = TUCR_3_6864MHz;

	msleep( 1 );

	sa1111_init();
	
	/* set nSRST */
	MEM(SYSTEM3_CTRL_1) = 0x04;

	/* issue a short beep */
	MEM(SYSTEM3_CTRL_1) = 0x14;
	msleep( 50 );
	MEM(SYSTEM3_CTRL_1) = 0x04;

	/* tweak blob config */
	blob_status.boot_delay = 1;

	/* select drivers */
	reboot_driver = &sa11x0_reboot_driver;
	serial_driver = &sa11x0_serial_driver;
	led_driver = &sa11x0_gpio_led_driver;
	ether_driver = &smc9196_ether_driver;
	timer_driver = &intelarm_timer_driver;
}
__initlist(system3_init_hardware, INIT_LEVEL_DRIVER_SELECTION);



/*****************************************************************
 * sysver - print system version
 */
static int sysver_cmd( int argc, char *argv[] )
{
	printf( "version: '%s'\n", module_version );
	switch ( PT_ID_BOARDID ) {
		case PT_BOARD_ALPHA:
			printf( "ALPHA Board" );
			break;
		case PT_BOARD_SYS3:
			printf( "SYSTEM 3 Board" );
			break;
		case PT_BOARD_DAFIT:
			printf( "DAFIT Board" );
			break;
		default:
			printf( "UNKNOWN Board (ID 0x%X)", PT_ID_BOARDID );
			break;

	}
	printf( " revision 0x%02X.\n", PT_ID_REVISION ); 
	return 0;
}
static char sysver_help[] = "print BLOB and system3 version number\n";
__commandlist(sysver_cmd, "sysver", sysver_help);

/**********************************************************************
 * sys3dbg - set system3 debug level
 */
int sys3dbg_cmd( int argc, char *argv[] )
{
	extern int cf_dbg_set(int);
	int ret;
	u32 lvl;

	if ( argc < 2 ) return -EINVAL;

	ret = strtou32(argv[1], &lvl);
	if ( ret < 0 ) return -EINVAL;

	sys3_dbg = lvl;

#if defined(CONFIG_PCMCIA_SUPPORT)
	if ( lvl>5 ) {
		cf_dbg_set( 5 );
		tar_dbg_set( 5 );
	} else {
		cf_dbg_set( 0 );
		tar_dbg_set( 0 );
	}

	if ( lvl>10 ) {
		io_dbg_set( 5 );
		pcmcia_dbg_set( 5 );
	} else {
		io_dbg_set( 0 );
		pcmcia_dbg_set( 0 );
	}
#endif

	printf( "sys3dbg: debug level set to %d\n", sys3_dbg );
	return 0;
}
static char sys3dbg_help[] = "sys3dbg level\n";
__commandlist(sys3dbg_cmd, "sys3dbg", sys3dbg_help);

/*****************************************************************
 * sa1111_init - initialize companion chip
 */
static void sa1111_init()
{

	if ( PT_ID_BOARDID == PT_BOARD_DAFIT )
		return;

	/*
	 * First, set up the 3.6864MHz clock on GPIO 27 for the SA-1111:
	 * (SA-1110 Developer's Manual, section 9.1.2.1)
	 */
	GAFR |= GPIO_32_768kHz;
	GPDR |= GPIO_32_768kHz;
	TUCR = TUCR_3_6864MHz;

	/*
	 * Turn VCO on, and disable PLL Bypass.
	 */
	SBI_SKCR &= ~SKCR_VCO_OFF;
	SBI_SKCR |= SKCR_PLL_BYPASS | SKCR_OE_EN;

	/*
	 * Wait lock time.  SA1111 manual _doesn't_
	 * specify a figure for this!  We choose 100us.
	 */
	msleep(1);

	/*
	 * Enable RCLK.  We also ensure that RDYEN is set.
	 */
	SBI_SKCR |= SKCR_RCLKEN | SKCR_RDYEN;

	/*
	 * Wait 14 RCLK cycles for the chip to finish coming out
	 * of reset. (RCLK=24MHz).  This is 590ns.
	 */
	msleep(1);

	/*
	 * Ensure all clocks are initially off.
	 */
	SKPCR = 0;
}

#if defined(CONFIG_PCMCIA_SUPPORT)
/*****************************************************************
 * pcmcia_init - initialize PCMCIA
 */
int pcmcia_init()
{
	printf( "PT Digital Board PCMCIA init\n" );
	return 0;
}
#endif





/*****************************************************************
 * misc utility funcs
 */

/*****************************************************************
 * manually reference flash and download commands until they
 * are in libblob.
 */
#if 1
extern int fwrite_cmd(int argc, char *argv[]);
extern char fwrite_help[];
__commandlist(fwrite_cmd, "fwrite", fwrite_help);

extern int ferase_cmd(int argc, char *argv[]);
extern char ferase_help[];
__commandlist(ferase_cmd, "ferase", ferase_help);
#endif

extern int dlfile_cmd(int argc, char *argv[]);
extern char dlfile_help[];
__commandlist(dlfile_cmd, "dlfile", dlfile_help);


/**********************************************************************
 * system3 default partition table
 */
static void system3_set_partition_table(void)
{
	/* the default partition table */
	default_partition_table = system3_default_partition_table;
	
	/* we don't know where the flash partition table will live, so
	 * put the pointer at 0x0000000 and let the partition table
	 * parser figure out.
	 */
	flash_partition_table = (blob_partition_t*)0x00000000;
}

__initlist(system3_set_partition_table, INIT_LEVEL_OTHER_STUFF);
/* FIXME: need to think if this is the correct init level, we might
 * want to have this before the param list stuff starts running */
