/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *(C) Copyright 2006 Marvell International Ltd.  
 * All Rights Reserved 
 */

/*++

Module Name:  $Workfile: xllp_ov7660.c $

Abstract:  
     contains all primitive functions for OV7660

Notes:Only valid for processor code named Monahans.

--*/

#include "xllp_defs.h"
#include "xllp_serialization.h"
#include "xllp_clkmgr.h"
#include "xllp_ost.h"
#include "xllp_gpio_plat.h"
#include "xllp_gpio_proc.h"
#include "xllp_camera.h"
#include "xllp_ci.h"
#include "xllp_ov7660.h"
#include "xllp_ov7660_hw.h"

/***********************************************************************
 *
 * OV7660 Functions
 *
 ***********************************************************************/
XLLP_STATUS_T CameraFuncOV7660Init( P_XLLP_Camera_Context_T camera_context )
{
    XLLP_UINT8_T cm_rev, cm_pid;
    XLLP_UINT32_T timeout;
    int status;

    // provide informat about the capabilities of the sensor
    camera_context->sensor_status.caps |= XLLP_SENSOR_CAP_MANUAL_CONTRAST |
                                          XLLP_SENSOR_CAP_MANUAL_WHITEBALANCE |
                                          XLLP_SENSOR_CAP_MANUAL_EXPOSURE;

    // Configure CI according to OV7660's hardware        
    // master parallel with 8 data pins
    XllpCISetMode(camera_context->ci_reg_base, XLLP_CI_MODE_MP, XLLP_CI_DATA_WIDTH8);
    
    // enable pixel clock(sensor will provide pclock) and master clock = 26MHZ
    XllpCISetClock(camera_context->ci_reg_base, camera_context->clk_reg_base, XLLP_TRUE, XLLP_TRUE, 2600);

    // data sample on rising and h,vsync active high
    XllpCISetPolarity(camera_context->ci_reg_base, XLLP_FALSE, XLLP_FALSE, XLLP_FALSE);
    
    // fifo control
    XllpCISetFIFO(camera_context->ci_reg_base, 0, XLLP_CI_FIFO_THL_32, XLLP_TRUE, XLLP_TRUE);

    // OV7660 Power on sequence
    // Take out of Power down mode (GPIO_103), PWRDWN=1, NORMAL=0
    // Assert Reset
    // Delay
    // Remove reset
    // Delay
    OV7660PowerDown( camera_context->gpio_reg_base, camera_context->ost_reg_base, XLLP_CAMERA_POWER_FULL );
    OV7660Reset();
    XllpOstDelayMilliSeconds( (P_XLLP_OST_T)camera_context->ost_reg_base, 1 );
    
    // 2 Turn on M_CLK using xx MHz and wait for 150 ms.
    XllpCIEnable(camera_context->ci_reg_base/*, XLLP_TRUE*/);
    XllpOstDelayMilliSeconds((P_XLLP_OST_T)camera_context->ost_reg_base, 150);

    // read out version
    timeout = 50;
    do
    {
    	cm_pid = cm_rev = 0;
    	status = OV7660VersionRevision(&cm_pid, &cm_rev);
    
	    // Check to make sure we are working with an OV7660
    	if( cm_pid != PID_OV76XX || cm_rev != PID_7660 )
    	{
            OV7660PowerDown( camera_context->gpio_reg_base, camera_context->ost_reg_base, XLLP_CAMERA_POWER_OFF );
            OV7660PowerDown( camera_context->gpio_reg_base, camera_context->ost_reg_base, XLLP_CAMERA_POWER_FULL );
            OV7660Reset();
		    XllpOstDelayMilliSeconds( (P_XLLP_OST_T)camera_context->ost_reg_base, 1 );
		}
    	if (--timeout == 0)
    		return XLLP_STATUS_FAILURE;
    }
    while( cm_pid != PID_OV76XX );
    
	// turn sensor output off
    OV7660ViewfinderOff();

    return XLLP_STATUS_SUCCESS;
}

XLLP_STATUS_T CameraFuncOV7660DeInit(  P_XLLP_Camera_Context_T camera_context )
{
    // power off the external module
    OV7660PowerDown( camera_context->gpio_reg_base, camera_context->ost_reg_base, XLLP_CAMERA_POWER_OFF );

    return XLLP_STATUS_SUCCESS;
}

XLLP_STATUS_T CameraFuncOV7660Sleep(  P_XLLP_Camera_Context_T camera_context )
{
	return CameraFuncOV7660DeInit( camera_context );	
}

XLLP_STATUS_T CameraFuncOV7660Wake(  P_XLLP_Camera_Context_T camera_context )
{
	return CameraFuncOV7660Init( camera_context );
}

XLLP_STATUS_T CameraFuncOV7660SetCaptureFormat(  P_XLLP_Camera_Context_T camera_context )
{
    XLLP_CI_MP_TIMING timing;
    int status;
    XLLP_UINT32_T ovSizeFormat, ovFormat;

    // Set the current mode
    switch(camera_context->capture_input_format) {
        case XLLP_CAMERA_IMAGE_FORMAT_YCBCR422_PLANAR:
        case XLLP_CAMERA_IMAGE_FORMAT_YCBCR422_PACKED:
            ovFormat = OV_FORMAT_YUV_422;
            break;
        case XLLP_CAMERA_IMAGE_FORMAT_RGB565:
            ovFormat = OV_FORMAT_RGB_565;
            break;
        case XLLP_CAMERA_IMAGE_FORMAT_RAW8:
            ovFormat = OV_FORMAT_RAW8;
            break;
        default:
            ovFormat = OV_FORMAT_YUV_422;
            break;
    }
    if ( camera_context->capture_input_width == 88 && camera_context->capture_input_height == 72 )
    	ovSizeFormat = OV_SIZE_QQCIF;
    else if ( camera_context->capture_input_width == 176 && camera_context->capture_input_height == 144 )
    	ovSizeFormat = OV_SIZE_QCIF;
    else if ( camera_context->capture_input_width == 352 && camera_context->capture_input_height == 288 )
    	ovSizeFormat = OV_SIZE_CIF;
    else if ( camera_context->capture_input_width == 160 && camera_context->capture_input_height == 120 )
    	ovSizeFormat = OV_SIZE_QQVGA;
    else if ( camera_context->capture_input_width == 320 && camera_context->capture_input_height == 240 )
    	ovSizeFormat = OV_SIZE_QVGA;
    else if ( camera_context->capture_input_width == 640 && camera_context->capture_input_height == 480 )
    	ovSizeFormat = OV_SIZE_VGA;
    else
	    ovSizeFormat = OV_SIZE_VGA;

	status = OV7660SetFormat( ovSizeFormat, ovFormat, camera_context->capture_mode);//, camera_context->light, camera_context->zoom );
	        
    // set capture width/height and timing 
    if (ovFormat == OV_FORMAT_RAW8){
        timing.BFW = 0x00;
        timing.BLW = 0x01; /*GRBG to RGGB*/           
    }else{
    timing.BFW = 0x00;
    timing.BLW = 0x00;            
    }
    XllpCIConfigureMP(camera_context->ci_reg_base, camera_context->capture_input_width-1, camera_context->capture_input_height-1, &timing);

    return XLLP_STATUS_SUCCESS;
}

XLLP_STATUS_T CameraFuncOV7660StartCapture(  P_XLLP_Camera_Context_T camera_context, unsigned int frames )
{   
    (void) camera_context;
    
	// turn auto function on only doing continues capture
	if (frames == 0) {
		OV7660AutoFunctionOn();
	}
	else {
		OV7660AutoFunctionOff();
	}

    // turn viewfinder on
    OV7660ViewfinderOn();
    return XLLP_STATUS_SUCCESS;
}

XLLP_STATUS_T CameraFuncOV7660StopCapture(  P_XLLP_Camera_Context_T camera_context )
{
    (void) camera_context;
    
	// turn auto function off
	OV7660AutoFunctionOff();	

	// turn viewfinder off
    OV7660ViewfinderOff();

    return XLLP_STATUS_SUCCESS;
}

XLLP_STATUS_T CameraFuncOV7660SetPowerMode(P_XLLP_Camera_Context_T camera_context, XLLP_UINT8_T mode)
{
    (void) camera_context;

    OV7660PowerDown( camera_context->gpio_reg_base, camera_context->ost_reg_base, mode );
    return XLLP_STATUS_SUCCESS;
}

XLLP_STATUS_T CameraFuncOV7660Read8bit( P_XLLP_Camera_Context_T camera_context, XLLP_UINT8_T reg_addr,  XLLP_UINT8_T *reg_val)
{
    (void) camera_context;

    OV7660ReadSensorReg( reg_addr, reg_val );
	return XLLP_STATUS_SUCCESS;
}

XLLP_STATUS_T CameraFuncOV7660Write8bit( P_XLLP_Camera_Context_T camera_context, XLLP_UINT8_T reg_addr,  XLLP_UINT8_T reg_val)
{
    XLLP_UINT8_T buffer;
    XLLP_STATUS_T status;

    (void) camera_context;

    buffer = reg_val;
    status = (OV7660WriteSensorReg( reg_addr, &buffer ) == 0) ?
              XLLP_STATUS_SUCCESS : XLLP_STATUS_FAILURE;

    return status;
}

XLLP_STATUS_T CameraFuncOV7660SetContrast( P_XLLP_Camera_Context_T camera_context, XLLP_UINT8_T mode, XLLP_UINT32_T value)
{
    (void) camera_context;

    if (mode == XLLP_SENSOR_MANUAL_CONTRAST)
        return OV7660SetContrast(value);
    else 
    	return XLLP_STATUS_SUCCESS;    	
}

XLLP_STATUS_T CameraFuncOV7660SetExposure( P_XLLP_Camera_Context_T camera_context, XLLP_UINT8_T mode, XLLP_UINT32_T value)
{
    (void) camera_context;
    
    if (mode == XLLP_SENSOR_MANUAL_EXPOSURE)
        return OV7660SetExposure(value);
    else
    	return XLLP_STATUS_SUCCESS;
}

XLLP_STATUS_T CameraFuncOV7660SetWhiteBalance( P_XLLP_Camera_Context_T camera_context, XLLP_UINT8_T mode, XLLP_UINT32_T value)
{
    (void) camera_context;
    
    if (mode == XLLP_SENSOR_MANUAL_WHITEBALANCE)
        return OV7660SetWhiteBalance(value);
    else
    	return XLLP_STATUS_SUCCESS;
}
