/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *(C) Copyright 2006 Marvell International Ltd.  
 * All Rights Reserved 
 */

/** 
**  FILENAME:       xllp_pmu.h
**
**	CORE STEPPING:  
**
**  PURPOSE: contains all PMU specific macros, typedefs, and prototypes.
**           Declares no storage.
**                  
**
******************************************************************************/

#ifndef __XLLP_PMU_H__
#define __XLLP_PMU_H__


#include "xllp_defs.h"



/* PMU Performance Monitor Control Register (PMNC) */
#define XLLP_PMU_ID							(0x24u << 24)
#define XLLP_PMU_COUNTERS_DISALBLE			(XLLP_BIT_4)
#define XLLP_PMU_CLOCK_DIVIDER				(XLLP_BIT_3)
#define XLLP_PMU_CLOCK_RESET				(XLLP_BIT_2)
#define XLLP_PMU_COUNTERS_RESET				(XLLP_BIT_1)
#define XLLP_PMU_3_COUNTERS_ENABLE			(XLLP_BIT_0)
#define XLLP_PMU_COUNTERS_ENABLE			(XLLP_BIT_0)

#define XLLP_PMNC_RESERVED_BITS				(0xFFFFFFE0u)

/* INTEN & FLAG Registers bit definition*/
#define XLLP_PMU_CLOCK_COUNT				(XLLP_BIT_0)
#define XLLP_PMU_COUNT_0					(XLLP_BIT_1)
#define XLLP_PMU_COUNT_1					(XLLP_BIT_2)
#define XLLP_PMU_COUNT_2					(XLLP_BIT_3)
#define XLLP_PMU_COUNT_3					(XLLP_BIT_4)

#define XLLP_INTEN_FLAG_RESERVED_BITS		(0xFFFFFFE0u)

/*Events combination*/
/*!evtCount0/2:0x7(instruction count), evtCount1/3:0x0(ICache miss)*/
#define XLLP_PMU_EVTCOUNT_1					(0x0007)
/*!evtCount0/2:0xA(DCache Access), evtCount1/3:0xB(DCache miss)*/
#define XLLP_PMU_EVTCOUNT_2					(0x0B0A)
/*!evtCount0/2:0x1(ICache cannot deliver), evtCount1/3:0x0(ICache miss)*/
#define XLLP_PMU_EVTCOUNT_3					(0x0001)
/*!evtCount0/2:0xB(DBufer stall duration), evtCount1/3:0x9(Dbuffer stall)*/
#define XLLP_PMU_EVTCOUNT_4					(0x090B)
/*!evtCount0/2:0x2(data stall), evtCount1/3:0xC(DCache writeback)*/
#define XLLP_PMU_EVTCOUNT_5					(0x0C02)
/*!evtCount0/2:0x7(instruction count), evtCount1/3:0x3(ITLB miss)*/
#define XLLP_PMU_EVTCOUNT_6					(0x0307)
/*!evtCount0/2:0xA(DCache Access), evtCount/31:0x4(DTLB miss)*/
#define XLLP_PMU_EVTCOUNT_7					(0x040A)

typedef enum 
{
    XLLP_PMU_PMNC = 0,
    XLLP_PMU_CCNT,
    XLLP_PMU_PMN0,
    XLLP_PMU_PMN1,
	XLLP_PMU_PMN2,
	XLLP_PMU_PMN3,
	XLLP_PMU_INTEN,
	XLLP_PMU_FLAG,
	XLLP_PMU_EVTSEL
}XLLP_PMU_REGNUMBER;

/**
  PMU and PML Event
*/
typedef enum
{
	XLLP_PMU_EVENT_INVALIDATE=0xFFFFFFFFu,
	XLLP_PMU_EVENT_L1_INSTRUCTION_MISS=0x0u,			/*!< L1 Instruction cache miss requires fetch from external memory */

	XLLP_PMU_EVENT_L1_INSTRUCTION_NOT_DELIVER,	/*!< L1 Instruction cache cannot deliver an instruction. this  indicate
												 an instruction cache or TLB miss. This event will occur eveyr cycle in 
												 which the condition is present */
	
	XLLP_PMU_EVENT_STALL_DATA_DEPENDENCY,		/*!< Stall due to a data dependency. This event will occur every cycle in which 
												 the condition is present */

	XLLP_PMU_EVENT_INSTRUCTION_TLB_MISS,			/*!< Instruction TLB miss*/
	
	XLLP_PMU_EVENT_DATA_TLB_MISS,				/*!< Data TLB miss*/

	XLLP_PMU_EVENT_BRANCH_RETIRED,				/*!< Branch instruction retired, branch may or many not have changed program flow.
												 (Counts only B and BL instruction, in both ARM and Thumb mode)*/

	XLLP_PMU_EVENT_BRANCH_MISPREDICTED,			/*!< Branch mispredicted. Counts only B and BL instructions, in both ARM and Thumb mode*/

	XLLP_PMU_EVENT_INSTRUCTION_RETIRED,			/*!< Instruction retired. This event will occur every cycle in which 
												 the condition is present*/
	
	XLLP_PMU_EVENT_L1_DATA_STALL,				/*!< L1 Data cache buffer full stall. This event will occur every cycle in which 
												 the condition is present. */

	XLLP_PMU_EVENT_L1_DATA_STALL_C,				/*!< L1 Data cache buffer full stall. This event occur for each contiguous sequence of 
												 this type of stall*/

	XLLP_PMU_EVENT_L1_DATA_ACCESS,				/*!< L1 Data cache access, not including Cache Operations. All data accesses are 
												 treated as cacheable accessses and are counted here even if the cache is not enabled*/
	
	XLLP_PMU_EVENT_L1_DATA_MISS,					/*!< L1 Data cache miss, not including Cache Operations. All data accesses are treated as 
												 cachedable accesses and are counted as misses if the data cache is not enable*/
	
	XLLP_PMU_EVENT_L1_DATA_WRITE_BACK,			/*!< L1 data cache write-back. This event occures once for each line that is written back
												 from the cache*/

	XLLP_PMU_EVENT_SOFTWARE_CHANGED_PC,			/*!< Software changed the PC(b bx bl blx and eor sub rsb add adc sbc rsc orr mov bic mvn ldm pop)
											   will be counted. The count does not increment when an exception occurs and the PC changed to 
											   the exception address(e.g.. IRQ, FIR, SWI,...)*/

	XLLP_PMU_EVENT_BRANCH_RETIRED_ALL,			/*!< Branch instruction retired, branch may or may noot have chanaged program flow.(Count
											    ALL branch instructions, indirect as well as direct*/

	XLLP_PMU_EVENT_INSTRUCTION_CYCLE_RETIRED,	/*!< Instruction issue cycle of retired instruction. This event is a count of the number of 
												core cycle each instruction requires to issue*/

	XLLP_PMU_EVENT_ALL_CHANGED_PC=0x18,				/*!< All change to the PC. (includes software changes and exceptions*/

	XLLP_PMU_EVENT_PIPE_FLUSH_BRANCH,			/*!< Pipe line flush due to branch mispredict or exception*/

	XLLP_PMU_EVENT_BACKEND_STALL,				/*!< The core could not issue an instruction due to a backed stall. This event will occur every 
											     cycle in which the condition is present*/
	
	XLLP_PMU_EVENT_MULTIPLIER,					/*!< Multiplier in use. This event will occur every cycle in which the multiplier is active*/

	XLLP_PMU_EVENT_MULTIPLIER_STALL_PIPE,		/*!< Multiplier stalled the instruction pipelien due to resource stall. This event will occur
												 every cycle in which the condition is present*/

	XLLP_PMU_EVENT_COPROCESSOR_STALL_PIPE,		/*!< Coprocessor stalled the instruction pipeline. This event will occur every cycle in which the 
												 condition is present*/

	XLLP_PMU_EVENT_DATA_CACHE_STALL_PIPE,		/*!< Data cache stalled the instruction pipeline. This event will occur every cycle in which the 
												 condition is present*/

	XLLP_PMU_EVENT_L2_REQUEST=0x20,				/*!< Unified L2 Cache request, not including cache operations. This event includes table walks, 
												 data and instruction reqeusts*/

	XLLP_PMU_EVENT_L2_MISS=0x23,					/*!< Unified L2 cache miss, not including cache operations*/

	XLLP_PMU_EVENT_ADDRESS_BUS=0x40,				/*!< Address bus transcation*/

	XLLP_PMU_EVENT_SELF_INITIATED_ADDRRES,		/*!< Self initiated(Core Generated) address bus transaction*/

	XLLP_PMU_EVNET_BUS_CLOCK=0x43,				/*!< Bus clock. This event occurs onece for each bus cycle*/

	XLLP_PMU_EVENT_SELF_INITIATED_DATA=0x47,			/*!< Data bus transaction. This event occurs once for each data bus cycle*/

	XLLP_PMU_EVENT_BUS_TRANSACTION,				/*!< Data bus transaction. This event occures once for each data bus cycle*/


	XLLP_PMU_EVENT_ASSP_0=0x80,
	XLLP_PMU_EVENT_ASSP_1,
	XLLP_PMU_EVENT_ASSP_2,
	XLLP_PMU_EVENT_ASSP_3,
	XLLP_PMU_EVENT_ASSP_4,
	XLLP_PMU_EVENT_ASSP_5,
	XLLP_PMU_EVENT_ASSP_6,
	XLLP_PMU_EVENT_ASSP_7,
	
	XLLP_PMU_EVNET_POWER_SAVING=0xFF,			/*!< Power Saving event. This event deactivates the corresponding PMU event counter*/

	XLLP_MONAHANS_EVENT_MASK=0x80000000,
	XLLP_MONAHANS_EVENT_CORE_INSTRUCTION_FETCH=XLLP_MONAHANS_EVENT_MASK,  
												/*!< Core is performing a new instruction fetch. e.g. an L2 cache miss.*/
	XLLP_MONAHANS_EVENT_CORE_DATA_FETCH,				/*!< Core is performing a new data fetch*/
	XLLP_MONAHANS_EVENT_CORE_READ,					/*!< Core read request count*/
	XLLP_MONAHANS_EVENT_LCD_READ,					/*!< LCD read request cout*/
	XLLP_MONAHANS_EVENT_DMA_READ,					/*!< DMA read request count*/
	XLLP_MONAHANS_EVENT_CAMERA_READ,					/*!< Camera interface read request cout*/
	XLLP_MONAHANS_EVENT_USB20_READ,					/*!< USB 2.0 read request count*/
	XLLP_MONAHANS_EVENT_2D_READ,						/*!< 2D grahpic read request count*/
	XLLP_MONAHANS_EVENT_USB11_READ,					/*!< USB1.1 host read reqeust count*/
	XLLP_MONAHANS_EVENT_PX1_UNITIZATION,				/*!< PX1 bus unitization. the number of cycles durring which the PX1 bus is occupied*/
	XLLP_MONAHANS_EVENT_PX2_UNITIZATION,				/*!< PX2(sidecar) bus unitization. the number of cycles durring which the PX2 bus is occupied*/
	XLLP_MONAHANS_EVENT_DMC_NOT_EMPTY=XLLP_MONAHANS_EVENT_MASK|14,			
												/*!< Dynamic memory queue for Mandris occupied. the number of cycles when the DMC queue is not empty*/
	XLLP_MONAHANS_EVENT_DMC_2,						/*!< Dynamic memory queue for Mandris occupied by more than 1 request. the number of cycles when the DMC
													 queue has 2 or more requests*/

	XLLP_MONAHANS_EVENT_DMC_3,						/*!< Dynamic memory queue for Mandris occupied by more than 2 request. the number of cycles when the DMC
													 queue has 3 or more requests*/

	XLLP_MONAHANS_EVENT_DMC_FULL,					/*!< Dynamic memory queue for Mandris occupied by more than 3 request. the number of cycles when the DMC
													 queue is full*/										
	
	XLLP_MONAHANS_EVENT_SMC_NOT_EMPTY,				/*!< Static memory queue for Mandris occupied. the number of cycles when the SMC queue is not empty*/
	XLLP_MONAHANS_EVENT_SMC_2,						/*!< Static memory queue for Mandris occupied by more than 1 request. the number of cycles when the SMC
													 queue has 2 or more requests*/

	XLLP_MONAHANS_EVENT_SMC_3,						/*!< Static memory queue for Mandris occupied by more than 2 request. the number of cycles when the SMC
													 queue has 3 or more requests*/

	XLLP_MONAHANS_EVENT_SMC_FULL,					/*!< Static memory queue for Mandris occupied by more than 3 request. the number of cycles when the SMC
													 queue is full*/				
	
	XLLP_MONAHANS_EVENT_CORE_WRITE=XLLP_MONAHANS_EVENT_MASK|36,	
												/*!< Core write request count*/
	XLLP_MONAHANS_EVENT_DMA_WRITE,					/*!< DMA write request count*/
	XLLP_MONAHANS_EVENT_CAMERA_WRITE,				/*!< Camera interface write request cout*/
	XLLP_MONAHANS_EVENT_USB20_WRITE,					/*!< USB 2.0 write request count*/
	XLLP_MONAHANS_EVENT_2D_WRITE,					/*!< 2D grahpic write request count*/
	XLLP_MONAHANS_EVENT_USB11_WRITE,					/*!< USB1.1 host write reqeust count*/

	XLLP_MONAHANS_EVENT_PX1_REQUEST,					/*!< PX1 bus reqeust. length of time that at least one bus request is asserted on PX bus 1*/
	XLLP_MONAHANS_EVENT_PX2_REQUEST,					/*!< PX2 bus reqeust.  length of time that at least one bus request is asserted on PX bus 2*/

	XLLP_MONAHANS_EVENT_PX1_RETRIES,					/*!< PX1 bus retries. number of retries on PX bus 1*/
	XLLP_MONAHANS_EVENT_PX2_RETRIES,					/*!< PX2 bus retries. number of retries on PX bus 2*/

	
	XLLP_MONAHANS_EVENT_TEMPERATURE_1,				/*!< Temperature leve 1. time the part has spent in temperature range 1*/
	XLLP_MONAHANS_EVENT_TEMPERATURE_2,				/*!< Temperature leve 1. time the part has spent in temperature range 2*/
	XLLP_MONAHANS_EVENT_TEMPERATURE_3,				/*!< Temperature leve 1. time the part has spent in temperature range 3*/
	XLLP_MONAHANS_EVENT_TEMPERATURE_4,				/*!< Temperature leve 1. time the part has spent in temperature range 4*/

	XLLP_MONAHANS_EVENT_CORE_LATENCY_1,				/*!< Core read/write latency measurement. amount of time when core have more than 1 read/write
												  request outstanding*/

	XLLP_MONAHANS_EVENT_CORE_LATENCY_2,				/*!< Core read/write latency measurement. amount of time when core have more than 2 read/write
												  request outstanding*/
	XLLP_MONAHANS_EVENT_CORE_LATENCY_3,				/*!< Core read/write latency measurement. amount of time when core have more than 3 read/write
												  request outstanding*/
	XLLP_MONAHANS_EVENT_CORE_LATENCY_4,				/*!< Core read/write latency measurement. amount of time when core have more than 4 read/write
												  request outstanding*/

	XLLP_MONAHANS_EVENT_PX1_IM_1,					/*!<  PX1 to IM read/write latency measurement. Amount of time when PX1 to IM has more than 1 
												   read/write requests outstanding.*/
	XLLP_MONAHANS_EVENT_PX1_IM_2,					/*!<  PX1 to IM read/write latency measurement. Amount of time when PX1 to IM has more than 2 
												   read/write requests outstanding.*/
	XLLP_MONAHANS_EVENT_PX1_IM_3,					/*!<  PX1 to IM read/write latency measurement. Amount of time when PX1 to IM has more than 3 
												   read/write requests outstanding.*/
	XLLP_MONAHANS_EVENT_PX1_IM_4,					/*!<  PX1 to IM read/write latency measurement. Amount of time when PX1 to IM has more than 4 
												   read/write requests outstanding.*/

	XLLP_MONAHANS_EVENT_PX1_MEM_1,					/*!<  PX1 to DMEM/SMEM read/write latency measurement. Amount of time when PX1 to DMEM/SMEM has more than 1 
												   read/write requests outstanding.*/
	XLLP_MONAHANS_EVENT_PX1_MEM_2,					/*!<  PX1 to DMEM/SMEM read/write latency measurement. Amount of time when PX1 to DMEM/SMEM has more than 2 
												   read/write requests outstanding.*/
	XLLP_MONAHANS_EVENT_PX1_MEM_3,					/*!<  PX1 to DMEM/SMEM read/write latency measurement. Amount of time when PX1 to DMEM/SMEM has more than 3 
												   read/write requests outstanding.*/
	XLLP_MONAHANS_EVENT_PX1_MEM_4,					/*!<  PX1 to DMEM/SMEM read/write latency measurement. Amount of time when PX1 to DMEM/SMEM has more than 4 
												   read/write requests outstanding.*/

	XLLP_MONAHANS_EVENT_PX2_IM_1,					/*!<  PX2 to IM read/write latency measurement. Amount of time when PX2 to IM has more than 1 
												   read/write requests outstanding.*/
	XLLP_MONAHANS_EVENT_PX2_IM_2,					/*!<  PX2 to IM read/write latency measurement. Amount of time when PX2 to IM has more than 2 
												   read/write requests outstanding.*/
	XLLP_MONAHANS_EVENT_PX2_IM_3,					/*!<  PX2 to IM read/write latency measurement. Amount of time when PX2 to IM has more than 3 
												   read/write requests outstanding.*/
	XLLP_MONAHANS_EVENT_PX2_IM_4,					/*!<  PX2 to IM read/write latency measurement. Amount of time when PX2 to IM has more than 4 
												   read/write requests outstanding.*/

	XLLP_MONAHANS_EVENT_PX2_MEM_1,					/*!<  PX2 to DMEM/SMEM read/write latency measurement. Amount of time when PX2 to DMEM/SMEM has more than 1 
												   read/write requests outstanding.*/
	XLLP_MONAHANS_EVENT_PX2_MEM_2,					/*!<  PX2 to DMEM/SMEM read/write latency measurement. Amount of time when PX2 to DMEM/SMEM has more than 2 
												   read/write requests outstanding.*/
	XLLP_MONAHANS_EVENT_PX2_MEM_3,					/*!<  PX2 to DMEM/SMEM read/write latency measurement. Amount of time when PX2 to DMEM/SMEM has more than 3 
												   read/write requests outstanding.*/
	XLLP_MONAHANS_EVENT_PX2_MEM_4					/*!<  PX2 to DMEM/SMEM read/write latency measurement. Amount of time when PX2 to DMEM/SMEM has more than 4 
												   read/write requests outstanding.*/

}XLLP_PMU_EVENT_TYPE;

typedef P_XLLP_VUINT32_T P_XLLP_PML_T;
typedef XLLP_VUINT32_T XLLP_PML_T;

#ifdef __cplusplus
extern "C"
{
#endif

/**
	This routine reads the designated PMU register via CoProcessor 14 

@param	 aReg		PMU register number to read define in XLLP_PMU_REGNUMBER
@return				32-bit value read from register
*/
extern unsigned int XllpReadPMUReg(unsigned int aReg);
/**
	This routine Writes the designated PMU register via CoProcessor 14 

@param	 aReg		PMU register number to read define in XLLP_PMU_REGNUMBER
		 aValue		Value to write to PMU register
@return				
*/
extern void XllpWritePMUReg(unsigned int aReg, unsigned int aValue);

extern XLLP_PMU_EVENT_TYPE XllpPMUEventSelect(int counter, XLLP_PMU_EVENT_TYPE type, P_XLLP_PML_T base);

#ifdef __cplusplus
}
#endif


#endif //__PMU_H__ 

