/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *(C) Copyright 2006 Marvell International Ltd.  
 * All Rights Reserved 
 */

#include "xllp_ac97acodec.h"
#include "xllp_ost.h"


static XLLP_BOOL_T XllpAc97ACODECLinkLock(XLLP_AC97ACODEC_T * pAc97);
static XLLP_ACODEC_ERROR_T  XllpAc97ACODECColdReset(XLLP_ACODEC_CONTEXT_T * pAc97ctxt);
static XLLP_ACODEC_ERROR_T  XllpAc97ACODECShutdownAclink(XLLP_AC97ACODEC_T * pAc97Reg, P_XLLP_OST_T pOstRegs);


/*******************************************************************************
 *
 * FUNCTION:         XllpAc97ACodecInit
 *
 * DESCRIPTION:      Perform a cold reset of the AC'97 controller and codec(s).
 *                   
 *                   Note: Does not enable any interrupt types within the 
 *                     ACUNIT, so no interrupts should occur at this point.
 *
 * INPUT PARAMETERS: 
 *       pAc97ctxt   a pointer to a XLLP_ACODEC_CONTEXT_T struct, which contains 
 *						necessary setup information.
 *
 * RETURNS:   
 *       Success:    0 
 *       Failure:    Refer to XllpAc97ACodecColdReset
 *
 * GLOBAL EFFECTS:   
 *
 * ASSUMPTIONS:      - No other systems, such as debuggers, are using the 
 *                     AC '97 Controller or codecs, or the AC Link.
 *                   - The software system will be unharmed by a cold reset of 
 *                     the entire AC '97 subsystem and any associated devices.
 *                   - Ac97 Clock should be enabled before the function called
 *
 * CALLS:            XllpAc97ACodecColdReset      
 *
 * CALLED BY:        
 *
 * PROTOTYPE:        XLLP_ACODEC_ERROR_T  XllpAc97ACodecInit(XLLP_ACODEC_CONTEXT_T *pAc97ctxt)
 *
 *******************************************************************************/

XLLP_ACODEC_ERROR_T  XllpAc97ACodecInit(XLLP_ACODEC_CONTEXT_T *pAc97ctxt)
{
    XLLP_ACODEC_ERROR_T	status ;
	
    status = XllpAc97ACODECColdReset(pAc97ctxt);
    
    return (status);
}

/*******************************************************************************
 *
 * FUNCTION:         XllpAc97ACODECDeInit
 *
 * DESCRIPTION:      Shutdown AC97 unit clearly and thoroughly.
 *                   Release GPIOs used by AC97.
 *                   Disable clock to AC97 unit. 
 *
 * INPUT PARAMETERS: 
 *       pAc97ctxt   a pointer to a XLLP_ACODEC_CONTEXT_T struct, which contains 
 *						necessary information.
 *
 * RETURNS:
 *       Success:    0 (XLLP_AC97_NO_ERROR)
 *       Failure:    Refer to XllpAc97ACODECShutdownAclink
 *
 * CALLS:            XllpAc97ACODECShutdownAclink
 *
 * CALLED BY:        
 *
 * PROTOTYPE:        XLLP_ACODEC_ERROR_T  XllpAc97ACODECDeInit(P_XLLP_ACODEC_CONTEXT_T pAc97ctxt);
 *
 *******************************************************************************/

XLLP_ACODEC_ERROR_T  XllpAc97ACodecDeInit(XLLP_ACODEC_CONTEXT_T * pAc97ctxt)
{
    XLLP_ACODEC_ERROR_T	status ;
	
    status = XllpAc97ACODECShutdownAclink((P_XLLP_AC97ACODEC_T)(pAc97ctxt->pCtrlReg), pAc97ctxt->pOSTRegs);

    return (status);
}



XLLP_ACODEC_ERROR_T XllpAc97ACodecWrite (XLLP_ACODEC_CONTEXT_T *pDevContext, XLLP_UINT16_T offset, XLLP_UINT16_T data)
{
    XLLP_ACODEC_ERROR_T	status = XLLP_ACODEC_SUCCESS;
    XLLP_BOOL_T			gotLink;
    XLLP_UINT32_T		timeRemaining;
    P_XLLP_VUINT32_T	pCodecReg;
    P_XLLP_AC97ACODEC_T pAc97Reg = (P_XLLP_AC97ACODEC_T)(pDevContext->pCtrlReg); 
	P_XLLP_OST_T pOstRegs = pDevContext->pOSTRegs;
	XLLP_UINT32_T maxRWTimeOutUs = (pDevContext->uMaxReadWriteTimeOutMs) * 1000; 
	
    
	if(offset == XLLP_AC97_CR_E_MDM_GPIO_PIN_STAT)
	{//it is a special register and sent out on slot 12
    	pCodecReg = &(pAc97Reg->CodecRegsPrimaryMdm[0]);
    	pCodecReg += offset / XLLP_AC97_CODEC_REGS_PER_WORD;
        // The data will be sent out on slot 12.
        *pCodecReg = (XLLP_UINT32_T)data;
        goto done;
	}
        
    // Point to specified register within area mapped to target codec regs
	pCodecReg = &(pAc97Reg->CodecRegsPrimaryAud[0]);
	pCodecReg += offset / XLLP_AC97_CODEC_REGS_PER_WORD;

	
	//Lock the ACLINK
    timeRemaining = XLLP_AC97_LOCK_TIMEOUT_DEF;
    do
    {
        gotLink = XllpAc97ACODECLinkLock(pAc97Reg);
        if (XLLP_FALSE == gotLink)	// 1 usec is a long time.  Skip delay if possible.
        {
			XllpOstDelayMicroSeconds(pOstRegs, 1);
        }
    }        // Wait while time remaining and ACLINK not available
    while (timeRemaining-- && (XLLP_FALSE == gotLink));

    if (XLLP_FALSE == gotLink)	// Didn't get the ACLINK
    {
        status = XLLP_ACODEC_CONTROLLER_INTERFACE_TIMEOUT;
    }
    else	// We got the link. Perform the write operation and don't wait.
    {
        // First, clear old write status indication CDONE by writing a ONE to that bit.
        pAc97Reg->gsr = XLLP_AC97_GSR_CDONE_MSK;

        *pCodecReg = (XLLP_UINT32_T)data;       // Now the write!

        // Wait until write cycle is complete. There should be a way
        //  to do this speculatively at the beginning of the procedure.
        //  Need to discover it. Too inefficient to always wait.

        timeRemaining = maxRWTimeOutUs;
        do
        {
			XllpOstDelayMicroSeconds(pOstRegs, 1);
        }     // Wait while time remaining and command I/O still incomplete.
        while ( (timeRemaining--) && !(pAc97Reg->gsr & XLLP_AC97_GSR_CDONE_MSK));
		if (!(pAc97Reg->gsr & XLLP_AC97_GSR_CDONE_MSK))
		{
			status = XLLP_ACODEC_CONTROLLER_INTERFACE_TIMEOUT;
            pAc97Reg->car = XLLP_AC97_CAR_CAIP_CLEAR;
        }
    }  // Got AC link

done:
    return(status);
} // Ac97CtrlCodecWrite()


XLLP_ACODEC_ERROR_T XllpAc97ACodecRead  (XLLP_ACODEC_CONTEXT_T *pDevContext, XLLP_UINT16_T offset,  XLLP_UINT16_T *pdata)
{
    XLLP_ACODEC_ERROR_T	status = XLLP_ACODEC_SUCCESS;
    XLLP_BOOL_T			gotLink;
	XLLP_UINT32_T		timeRemaining;
    P_XLLP_VUINT32_T	pCodecReg;
    P_XLLP_AC97ACODEC_T pAc97Reg = (P_XLLP_AC97ACODEC_T)(pDevContext->pCtrlReg); 
	P_XLLP_OST_T pOstRegs = pDevContext->pOSTRegs;
	XLLP_UINT32_T maxRWTimeOutUs = (pDevContext->uMaxReadWriteTimeOutMs) * 1000; 
/*    if(offset == XLLP_AC97_CR_E_MDM_GPIO_PIN_STAT)
    {//it is a special register. it does not need dummy read and must be read in modem IO space
        // Select the Primary or Secondary modem IO address space
        pCodecReg = &(pAc97Reg->CodecRegsPrimaryMdm[0]);
        pCodecReg += offset / XLLP_AC97_CODEC_REGS_PER_WORD;

        // The data is received on Slot 12 and stored by the
        // ACUNIT so we can read back straight away.
        *pdata = (XLLP_UINT16_T)(*pCodecReg);
            
        goto done;
    }*/
    
    // Point to specified register within area mapped to target codec regs
	pCodecReg = &(pAc97Reg->CodecRegsPrimaryAud[0]);
	pCodecReg += offset / XLLP_AC97_CODEC_REGS_PER_WORD;

	//Lock the ACLINK
    timeRemaining = XLLP_AC97_LOCK_TIMEOUT_DEF;
    do
    {
        gotLink = XllpAc97ACODECLinkLock(pAc97Reg);
        if (XLLP_FALSE == gotLink)	// 1 usec is a long time.  Skip delay if possible.
        {
			XllpOstDelayMicroSeconds(pOstRegs, 1);
        }
    }        // Wait while time remaining and ACLINK not available
    while (timeRemaining-- && (XLLP_FALSE == gotLink));

    if (XLLP_FALSE == gotLink)	// Didn't get the ACLINK
    {
        status = XLLP_ACODEC_CONTROLLER_INTERFACE_TIMEOUT;
    }
    else	// We got the link. Perform the write operation and don't wait.
    {
         // First, clear old read status indications.
        pAc97Reg->gsr = XLLP_AC97_GSR_SDONE_MSK | XLLP_AC97_GSR_RCS_ERR_MSK;

        *pdata = (XLLP_UINT16_T)(*pCodecReg); // This is THE DUMMY READ.

         // Wait for read I/O with codec to complete before doing real read.
        timeRemaining = maxRWTimeOutUs;
        do
        {
			XllpOstDelayMicroSeconds(pOstRegs, 1);
        }   // Wait while time remaining and read I/O still incomplete
        while( (timeRemaining--) && (!(pAc97Reg->gsr & XLLP_AC97_GSR_SDONE_MSK)) );

        if ((pAc97Reg->gsr & XLLP_AC97_GSR_SDONE_MSK) && (!(pAc97Reg->gsr & XLLP_AC97_GSR_RCS_ERR_MSK)) )
        {
            if (pAc97Reg->gsr & XLLP_AC97_GSR_RCS_ERR_MSK)
            {//timeout indicated by RCS bit
                status = XLLP_ACODEC_CONTROLLER_INTERFACE_TIMEOUT;                
            }
            // succeed in reading. clear status bits first.
            pAc97Reg->gsr = XLLP_AC97_GSR_SDONE_MSK | XLLP_AC97_GSR_RCS_ERR_MSK;
            *pdata = (XLLP_UINT16_T)(*pCodecReg);	// THE REAL READ.
            if (*pdata == 0xffff)
            {//timeout indicated by returned value
                status = XLLP_ACODEC_CONTROLLER_INTERFACE_TIMEOUT;                
            }
			//check later: is second waiting really needed?
     		timeRemaining = maxRWTimeOutUs;
    	    do
	        {
				XllpOstDelayMicroSeconds(pOstRegs, 1);
  		    }   // Wait while time remaining and read I/O still incomplete
    	    while( (timeRemaining--) && (!(pAc97Reg->gsr & XLLP_AC97_GSR_SDONE_MSK)) );
        }
        else	// failed
        {
            status = XLLP_ACODEC_CONTROLLER_INTERFACE_TIMEOUT;
            pAc97Reg->car = XLLP_AC97_CAR_CAIP_CLEAR;
        } // else  (OK to do real read)

    } // else  (We got the link.  Perform the read operations.)

//done:    
    return (status);
} 


static XLLP_ACODEC_ERROR_T  XllpAc97ACODECColdReset(XLLP_ACODEC_CONTEXT_T * pAc97ctxt)  
{ 
    XLLP_ACODEC_ERROR_T	status = XLLP_ACODEC_SUCCESS;
    P_XLLP_AC97ACODEC_T pAC97 = (P_XLLP_AC97ACODEC_T)(pAc97ctxt->pCtrlReg);
    P_XLLP_OST_T 		pOstRegs = pAc97ctxt->pOSTRegs;
    XLLP_BOOL_T         priCodecReady;
    XLLP_UINT32_T		timeRemaining; 

    pAC97->gcr = 0;    
    pAC97->gcr |= XLLP_AC97_GCR_CLKBPB_MSK;
    // Hold reset active for a minimum time
	XllpOstDelayMicroSeconds(pOstRegs, XLLP_AC97_COLD_HOLDTIME);
    pAC97->gcr &= ~XLLP_AC97_GCR_CLKBPB_MSK;

    // Deactivate cold reset condition
	pAC97->gcr |= XLLP_AC97_GCR_COLD_RESET_MSK; 
    

    priCodecReady = XLLP_FALSE;
    timeRemaining = (pAc97ctxt->uMaxSetupTimeOutMs) * 10;
    do
    {
		XllpOstDelayMicroSeconds(pOstRegs, 1);
        if (pAC97->gsr & XLLP_AC97_GSR_PCRDY_MSK)
            priCodecReady = XLLP_TRUE;
    }
    while (timeRemaining-- && (priCodecReady == XLLP_FALSE));

    // Timeout status if some of the devices weren't ready.
    if (priCodecReady == XLLP_FALSE)
    {
        status = XLLP_ACODEC_CONTROLLER_INTERFACE_TIMEOUT;
    }

    return (status);
} 


static XLLP_ACODEC_ERROR_T  XllpAc97ACODECShutdownAclink(P_XLLP_AC97ACODEC_T pAc97Reg, P_XLLP_OST_T pOstRegs)
{
    XLLP_ACODEC_ERROR_T	status = XLLP_ACODEC_SUCCESS;
    XLLP_UINT32_T timeRemaining = XLLP_AC97_LINKOFF_TIMEOUT_DEF; 

    pAc97Reg->gcr |= XLLP_AC97_GCR_LINK_OFF_MSK;
    pAc97Reg->gcr |= XLLP_AC97_GCR_CLKBPB_MSK;
	
	while (!(pAc97Reg->gsr & XLLP_AC97_GSR_ACOFFD_MSK))
	{
		timeRemaining --;
		if (0 == timeRemaining)
		{
    		status = XLLP_ACODEC_CONTROLLER_INTERFACE_TIMEOUT;
    		break;
		}
		XllpOstDelayMicroSeconds(pOstRegs, 1);
    }
    pAc97Reg->gcr |= XLLP_AC97_GCR_FRCRST_MSK;
	//check later: any delay needed
    pAc97Reg->gcr &= ~XLLP_AC97_GCR_FRCRST_MSK;
    pAc97Reg->gcr &= ~XLLP_AC97_GCR_CLKBPB_MSK;
	
    return(status);
}

static XLLP_BOOL_T XllpAc97ACODECLinkLock(P_XLLP_AC97ACODEC_T pAc97Reg)
{
    XLLP_BOOL_T		status = XLLP_TRUE;
    XLLP_VUINT32_T	carTmp;

    carTmp = pAc97Reg->car;
    if (carTmp & XLLP_AC97_CAR_CAIP_MSK)	// "1" in CAIP bit means lock failed.
    {
        status = XLLP_FALSE;
    }
    return (status);
} // XllpAc97LinkLock()


XLLP_ACODEC_ERROR_T XllpAc97ACodecSaveContext (XLLP_ACODEC_CONTEXT_T *pDevContext, XLLP_AC97_SAVE_CONTEXT_T *pAc97SaveContext)
{
    P_XLLP_AC97ACODEC_T pAc97Reg = (P_XLLP_AC97ACODEC_T)(pDevContext->pCtrlReg); 
    pAc97SaveContext->pocr = pAc97Reg->pocr;
    pAc97SaveContext->picr = pAc97Reg->picr;
    pAc97SaveContext->mccr = pAc97Reg->mccr;
    pAc97SaveContext->pcscr = pAc97Reg->pcscr;
    pAc97SaveContext->pcclcr = pAc97Reg->pcclcr;
    pAc97SaveContext->mocr = pAc97Reg->mocr;
    pAc97SaveContext->micr = pAc97Reg->micr;
    pAc97SaveContext->gcr = pAc97Reg->gcr;
   
    return XLLP_ACODEC_SUCCESS;
}

XLLP_ACODEC_ERROR_T XllpAc97ACodecRestoreContext (XLLP_ACODEC_CONTEXT_T *pDevContext, XLLP_AC97_SAVE_CONTEXT_T *pAc97SaveContext)
{
    P_XLLP_AC97ACODEC_T pAc97Reg = (P_XLLP_AC97ACODEC_T)(pDevContext->pCtrlReg); 
   
    pAc97Reg->pocr = pAc97SaveContext->pocr;
    pAc97Reg->picr = pAc97SaveContext->picr;
    pAc97Reg->mccr = pAc97SaveContext->mccr;
    pAc97Reg->pcscr = pAc97SaveContext->pcscr;
    pAc97Reg->pcclcr = pAc97SaveContext->pcclcr;
    pAc97Reg->mocr = pAc97SaveContext->mocr;
    pAc97Reg->micr = pAc97SaveContext->micr;
    //check later: any side effect to restore ac97 GCR?
    pAc97Reg->gcr = pAc97SaveContext->gcr;
   
    return XLLP_ACODEC_SUCCESS;
}
