/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *(C) Copyright 2006 Marvell International Ltd.  
 * All Rights Reserved 
 */

/*++

Module Name:  $Workfile: xllp_dmac.c $

Abstract:  
     contains all primitive functions for Monahans DMA Controller register 
     access and control

Notes:Only valid for processor code named Monahans.

--*/

#include "xllp_dmac.h"    

#define XLLP_DMAC_DDADR_RESERVED_MASK     0xFFFFFFF0
#define XLLP_DMAC_DRCMR_ENABLE            0x80
#define XLLP_DMAC_DRCMR_DISABLE           0x00
 
/******************************************************************************
XLLP_DOC_HDR_BEGIN

  Function Name: XllpDmacInit

  Description: XllpDmacInit is used to initialize the DMA controller.  
                 
  Global Registers Modified: 
    DINT, DRCMRx register, DRCMRx register, DSADRx, DTADRx, DCMDx, where x denotes the channel number.  
  
  Input Arguments:
    pDmacHandle:  Specifies DMA unit base address.
  Output Arguments:
    None
  Return Value: 
    XLLP_STATUS_SUCCESS: indicates the DMA controller is initialized successfully. 

XLLP_DOC_HDR_END
*******************************************************************************/
XLLP_STATUS_T XllpDmacInit(P_XLLP_DMAC_T pDmacHandle)
{
    XLLP_UINT32_T        i;
    XLLP_UINT32_T        tmpDcsrClear;
    
    //
    // clear all of the DMA CSR registers
    //
    for (i=0; i < XLLP_DMAC_CHANNEL_NUM; i++)
    {
        // Sticky bits can be cleared by writing 1 to specific bit
        tmpDcsrClear = (pDmacHandle->dcsr[i] & XLLP_DCSR_WRITABLES_MSK);
        pDmacHandle->dcsr[i] = tmpDcsrClear;
        
        // Clear other bits
        pDmacHandle->dcsr[i] = 0x0;
    }
    
    //
    // Clear the interrupt register
    //
    pDmacHandle->dint  = 0x0;
    
    //
    // clear the DMA Request to Channel Map Registers
    //
    for (i=0; i < XLLP_DMAC_DRCMR1_NUM; i++)
    {
        pDmacHandle->drcmr1[i] = 0x0;
    }

    for (i=0; i < XLLP_DMAC_DRCMR2_NUM; i++)
    {
        pDmacHandle->drcmr2[i] = 0x0;
    }
    
    //
    // clear the descriptor registers
    //
    for(i=0; i < XLLP_DMAC_CHANNEL_NUM; i++)
    {
        pDmacHandle->ddg[i].ddadr = 0x0;
        pDmacHandle->ddg[i].dsadr = 0x0;
        pDmacHandle->ddg[i].dtadr = 0x0;
        pDmacHandle->ddg[i].dcmd  = 0x0;
    }
 
    return XLLP_STATUS_SUCCESS;
}


/******************************************************************************
XLLP_DOC_HDR_BEGIN

  Function Name: XllpDmacCfgChannelDescTransfer

  Description: XllpDmacCfgChannelDescTransfer configures the specified DMA 
                 channel for a Descriptor Mode transfer operation.  To configure 
                 a channel for a no descriptor mode transfer, use the 
                 XllpDmacCfgChannelNoDescTransfer primitive.  It is expected that 
                 the developer has already get an available channel from OS and 
                 bound it to a service routine for interrupt that may occur on the 
                 channel, and XllpDmacFillLinkedDesc to configure the descriptor 
                 before calling this primitives.
                 
  Global Registers Modified: 
    DDADRx, DRCMRx registers.  DSADRx, DTADRx, and DCMDx register will be modified
    when the descriptor is loaded.  x indicates the channel number        
  
  Input Arguments:
      pDmacHandle:  Specifies DMA unit base address.
    pDesc:  Hold the pointer to the descriptor to be used in configuring the 
            specified DMA channel.  The pointer cannot be null.
    aChannel:  Specifies the channel to be configured for the DMA transfer.
    aDeviceDrcmr:  Specifies the device to be mapped to the DMA channel for the transfer
                     operation.
                     
  Output Arguments:
      None
  Return Value: 
    None 

XLLP_DOC_HDR_END
*******************************************************************************/
void XllpDmacCfgChannelDescTransfer(
                                   P_XLLP_DMAC_T             pDmacHandle,
                                   P_XLLP_DMAC_DESCRIPTOR_T  pDesc,
                                   XLLP_DMAC_CHANNEL_T       aChannel,
                                   XLLP_DMAC_DEVICE_T        aDeviceDrcmr
                                   )
{
//    XLLP_VUINT32_T aTargetData; 

    //
    // Map device to channel
    //

    //
    // No mapping required for mem2mem moves
    //
    if (aDeviceDrcmr != XLLP_DMAC_MEM2MEM_MOVE)
    {
        XllpDmacMapDeviceToChannel(pDmacHandle, aDeviceDrcmr, aChannel);
    }

    XllpDmacLoadDescriptor(pDmacHandle, pDesc, aChannel);
}

/******************************************************************************
XLLP_DOC_HDR_BEGIN

  Function Name: XllpDmacCfgChannelNoDescTransfer

  Description: XllpDmacCfgChannelNoDescTransfer configures the specified DMA 
                 channel for a No Descriptor Mode transfer operation.  To configure
                 a channel for a descriptor mode transfer, use the 
                 XllpDmacCfgChannelDescTransfer primitive.  It is expected that the
                 developer has already obtained an available channel from OS and bound it to 
                 a service routine for interrupt that may occur on the channel.
                 
  Global Registers Modified: 
    DRCMRx, DSADRx, DTADRx, DCSRx, and DCMDx registers, where x indicate the 
    channel number  
  
  Input Arguments:
    pDmacHandle:  Specifies DMA unit base address.
    aSourceAddr:  Source address of  the DMA transfer to be configured 
    aTargetAddr:  Target address of the DMA transfer to be configured
    pCmd:  Commands for the channel to be configured.  Users are required to 
           initialize the various fields of pCmd data structure before making this
           call.
    aChannel: Specifies the channel to be configured for the DMA transfer.
    aDeviceDrcmr:  Specifies the device to be mapped to the DMA channel for the transfer
                     operation.               
                     
  Output Arguments:
      None
  Return Value: 
    None 

XLLP_DOC_HDR_END
*******************************************************************************/
void XllpDmacCfgChannelNoDescTransfer(
                                     P_XLLP_DMAC_T             pDmacHandle,
                                     XLLP_UINT32_T             aSourceAddr,      
                                     XLLP_UINT32_T             aTargetAddr,      
                                     XLLP_DMAC_COMMAND_T*      pCmd,
                                     XLLP_DMAC_CHANNEL_T       aChannel,
                                     XLLP_DMAC_DEVICE_T        aDeviceDrcmr
                                     )
{
    XLLP_VUINT32_T aCommand, aTargetData;

    //
    // Map device to channel                 
    //

    //
    // No mapping required for mem2mem moves 
    //
    if (aDeviceDrcmr != XLLP_DMAC_MEM2MEM_MOVE)
    {
        XllpDmacMapDeviceToChannel(pDmacHandle, aDeviceDrcmr, aChannel);
    }    

    //
    // Set the No Descriptor Fetch bit in DCSR
    //
    
    //
    // Protect the write-1-to-clear bits 
    // from being changed inadvertently.
    //
    aTargetData  = (pDmacHandle->dcsr[aChannel] & XLLP_DCSR_WRITABLES_MSK) & (~XLLP_DMAC_DCSR_WRITE_ONE_TO_CLEAR_BITS);;
    aTargetData |= XLLP_DMAC_DCSR_NO_DESC_FETCH;

    //
    // Protect the RUN bit since we have no atom access here.
    //
    pDmacHandle->dcsr[aChannel] = (aTargetData | XLLP_DMAC_DCSR_MASK_RUN);

    //
    // Construct command value from pCmd      
    //
    aCommand = (
                (pCmd->aLen)                               | 
                ((pCmd->aWidth)      << XLLP_BIT_FIELD_14) |
                ((pCmd->aSize)       << XLLP_BIT_FIELD_16) |
                ((pCmd->aEndian)     << XLLP_BIT_FIELD_18) | 
                ((pCmd->aEndIrqEn)   << XLLP_BIT_FIELD_21) |
                ((pCmd->aStartIrqEn) << XLLP_BIT_FIELD_22) |
                ((pCmd->aFlowTrg)    << XLLP_BIT_FIELD_28) |
                ((pCmd->aFlowSrc)    << XLLP_BIT_FIELD_29) |
                ((pCmd->aIncTrgAddr) << XLLP_BIT_FIELD_30) |
                ((pCmd->aIncSrcAddr) << XLLP_BIT_FIELD_31)
               );           

    pDmacHandle->ddg[aChannel].dsadr  =    aSourceAddr;
    pDmacHandle->ddg[aChannel].dtadr  =    aTargetAddr;
    pDmacHandle->ddg[aChannel].dcmd   =    aCommand;
}

/******************************************************************************
XLLP_DOC_HDR_BEGIN

  Function Name: XllpDmacStartTransfer

  Description: XllpDmacStartTransfer starts the DMA transfer process for the 
                 specified channel.
                 
  Global Registers Modified: 
    DCSRx register, where x indicate the channel number
  
  Input Arguments:
    pDmacHandle:  Specifies DMA unit base address.
    aChannel:  Specifies the channel to start the DMA transfer on.                     
  
  Output Arguments:
      None
  Return Value: 
    None 

XLLP_DOC_HDR_END
*******************************************************************************/
void XllpDmacStartTransfer(P_XLLP_DMAC_T  pDmacHandle, XLLP_DMAC_CHANNEL_T aChannel )
{
    XLLP_UINT32_T val = (pDmacHandle->dcsr[aChannel] & XLLP_DCSR_WRITABLES_MSK) & (~XLLP_DMAC_DCSR_WRITE_ONE_TO_CLEAR_BITS);
    
    pDmacHandle->dcsr[aChannel] = (val & (~XLLP_DMAC_DCSR_MASK_RUN)) | XLLP_DMAC_DCSR_RUN;
}

/******************************************************************************
XLLP_DOC_HDR_BEGIN

  Function Name: XllpDmacStopTransfer

  Description: XllpDmacStopTransfer stops the DMA transfer process for the 
                 specified channel.
                 
  Global Registers Modified: 
    DCSRx register, where x indicate the channel number
  
  Input Arguments:
    pDmacHandle:  Specifies DMA unit base address.  
    aChannel:  Specifies the channel to stop the DMA transfer on.  
  
  Output Arguments:
      None
  Return Value: 
    None 

XLLP_DOC_HDR_END
*******************************************************************************/

void XllpDmacStopTransfer(P_XLLP_DMAC_T  pDmacHandle,  XLLP_DMAC_CHANNEL_T aChannel )
{
    XLLP_UINT32_T val = (pDmacHandle->dcsr[aChannel] & XLLP_DCSR_WRITABLES_MSK) & (~XLLP_DMAC_DCSR_WRITE_ONE_TO_CLEAR_BITS);

    if (!(val & XLLP_DMAC_DCSR_STOP_INTR)) 
    {
        //
        // issue stop command
        //
        pDmacHandle->dcsr[aChannel] = ((val & (~XLLP_DMAC_DCSR_MASK_RUN)) & (~XLLP_DMAC_DCSR_RUN));
    }
}

/******************************************************************************
XLLP_DOC_HDR_BEGIN

  Function Name: XllpDmacMapDeviceToChannel

  Description: XllpDmacMapDeviceToChannel is used to map the specified device
                 to the specified DMA channel. 
                 
  Global Registers Modified: 
    DRCMRx register, where x denotes the channel number.  
  
  Input Arguments:
    pDmacHandle:  Specifies DMA unit base address.
    aDeviceDrcmr:  Specifies the device to be mapped to the DMA channel.  Valid 
              parameters are values of type  XLLP_DMAC_DRCMR_T, otherwise
              invalid.
    aChannel: Specifies the DMA channel to map the specified device to.  Valid 
              channel values range from 0-31.
  
  Output Arguments:
    None
  Return Value: 
    None 

XLLP_DOC_HDR_END
*******************************************************************************/
void XllpDmacMapDeviceToChannel(
                               P_XLLP_DMAC_T       pDmacHandle, 
                               XLLP_DMAC_DRCMR_T   aDeviceDrcmr,
                               XLLP_DMAC_CHANNEL_T aChannel
                               )
{
    if (aDeviceDrcmr < XLLP_DMAC_DRCMR1_NUM)
    {
        //
        // Enable device-to-channel mapping 
        //
        pDmacHandle->drcmr1[aDeviceDrcmr] = (XLLP_DMAC_DRCMR_ENABLE | aChannel);
    }
    else
    {
        //
        // Enable device-to-channel mapping 
        //
        pDmacHandle->drcmr2[aDeviceDrcmr-XLLP_DMAC_DRCMR1_NUM] = (XLLP_DMAC_DRCMR_ENABLE | aChannel);
    }    
}

/******************************************************************************
XLLP_DOC_HDR_BEGIN

  Function Name: XllpDmacUnMapDeviceToChannel

  Description: XllpDmacUnMapDeviceToChannel is used to unmap the specified 
                 device from the specified DMA channel.  
                 
  Global Registers Modified: 
    DRCMRx register, where x denotes the channel number.  
  
  Input Arguments:
    pDmacHandle:  Specifies DMA unit base address.  
    aDeviceDrcmr:  Specifies the device to be unmapped from the DMA channel.  Valid
              parameters are values of type XLLP_DMAC_DRCMR_T, otherwise invalid.
    aChannel: Specifies the DMA channel to unmap the specified device from.  
              Valid channel values range from 0-31.

  Output Arguments:
    None
  Return Value: 
    None 

XLLP_DOC_HDR_END
*******************************************************************************/
void XllpDmacUnMapDeviceToChannel(
                                 P_XLLP_DMAC_T       pDmacHandle, 
                                 XLLP_DMAC_DRCMR_T   aDeviceDrcmr,
                                 XLLP_DMAC_CHANNEL_T aChannel
                                 )
{
    if (aDeviceDrcmr < XLLP_DMAC_DRCMR1_NUM)
    {
        //
        // Disable device-to-channel mapping 
        //
        pDmacHandle->drcmr1[aDeviceDrcmr] = (XLLP_DMAC_DRCMR_DISABLE | aChannel);
    }
    else
    {
        //
        // Disable device-to-channel mapping 
        //
        pDmacHandle->drcmr2[aDeviceDrcmr-XLLP_DMAC_DRCMR1_NUM] = (XLLP_DMAC_DRCMR_DISABLE | aChannel);
    }    
}

/******************************************************************************
XLLP_DOC_HDR_BEGIN

  Function Name: XllpDmacFetchDescriptor

  Description:  This function is used to load descriptor address into DMA controller 
                in DMA fetch-descriptor mode.  
                 
  Global Registers Modified: 
    DDADRx register, where x denotes the channel number.  
  
  Input Arguments:
    pDmacHandle:  Specifies DMA unit base address.
    pDesc:  Hold the pointer to the descriptor to be used in configuring the 
            specified DMA channel.  The pointer cannot be null.
    aChannel:  Specifies the channel to be configured for the DMA transfer
  Output Arguments:
    None
  Return Value: 
    None 

XLLP_DOC_HDR_END
*******************************************************************************/
void XllpDmacLoadDescriptor(P_XLLP_DMAC_T             pDmacHandle,
                             P_XLLP_DMAC_DESCRIPTOR_T  pDesc,
                             XLLP_DMAC_CHANNEL_T       aChannel
                            )
{
    pDmacHandle->ddg[aChannel].ddadr = (XLLP_UINT32_T)pDesc;
}
