/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *(C) Copyright 2006 Marvell International Ltd.  
 * All Rights Reserved 
 */

/** 
**  FILENAME:       xllp_intc.c
**
** CORE STEPPING:  
**
**  PURPOSE: contains all INTC C function.
**           
**                  
**
*/

#include "xllp_intc.h"

/**
This function enable/disable special interrupt source.

@param
 source    Interrupt souce , @see XLLP_INTC_SOURCE_T.
 enable  - true : Enable interrupt
    - false: Disable interrupt
@return
 Old status before call this function

@remarks
 Assume the system already in required kernel/supervisor mode. 
 NOT guarantee serialization

*/

XLLP_BOOL_T XllpINTCEnableSource(XLLP_INTC_SOURCE_T source,XLLP_BOOL_T enable)
{
 
 XLLP_INT32_T reg,old,bit;
 XLLP_INT32_T tem_source =(XLLP_INT32_T)source;
 if(source>=32)
 {
  reg=XLLP_INTC_ICMR2;
  tem_source-=32;
 }
 else
 {
  reg=XLLP_INTC_ICMR;
 }

 bit = 1u << tem_source;

 old=XllpReadINTCReg(reg);

 if(enable)
 {
  XllpWriteINTCReg(reg,old|bit);

 }else
 {
  XllpWriteINTCReg(reg,old&(~bit));
 }

 return (XLLP_BOOL_T)(old&bit);

}



/**
This function setl interrupt source fiq or irq

@param
 source    Interrupt souce , @see XLLP_INTC_SOURCE_T.
 enable  - true : FIQ
    - false: IRQ
@return
 Old status before call this function

@remarks
 Assume the system already in required kernel/supervisor mode. 
 NOT guarantee serialization 

*/

XLLP_BOOL_T XllpINTCSetSourceLevel(XLLP_INTC_SOURCE_T source,XLLP_BOOL_T enable)
{
 
 XLLP_INT32_T reg,old,bit;
 XLLP_INT32_T tem_source = (XLLP_INT32_T)source;
 if(source>=32)
 {
  reg=XLLP_INTC_ICLR2;
  tem_source -= 32;
 }
 else
 {
  reg=XLLP_INTC_ICLR;
 }

 bit = 1u << tem_source;

 old=XllpReadINTCReg(reg);

 if(enable)
 {
  XllpWriteINTCReg(reg,old|bit);

 }else
 {
  XllpWriteINTCReg(reg,old&(~bit));
 }

 return (XLLP_BOOL_T)(old&bit);

}



/// Save state before system enter sleep state
///
/// @param 
///     pVirtual    INTC base register addresses
///     pSave       Buff to save status
///
/// @return
///     0   -   Successful
///     XLLP_STATUS_WRONG_PARAMETER   -   Fail
/// @remarks
///  Assume the system already in required kernel/supervisor mode. 
///  NOT guarantee serialization 

XLLP_STATUS_T XllpINTCSave(P_XLLP_INTC_T pVirtual, P_XLLP_INTC_SAVE_T pSave)
{
    int i=0;

    if(!pVirtual)
        return XLLP_STATUS_WRONG_PARAMETER;

    if(!pSave)
        return XLLP_STATUS_WRONG_PARAMETER;

    pSave->iccr = pVirtual->iccr;
    
    for(i=0;i<32;i++)
        pSave->ipr[i] = pVirtual->ipr[i];  // Interrupt Priority Registers [0-31]
    
    for(i=0;i<21;i++)
        pSave->ipr[i+32] = pVirtual->ipr2[i];  // Interrupt Priority Registers [32-53]
    

    pSave->icmr = XllpReadINTCReg(XLLP_INTC_ICMR);
    pSave->icmr2 = XllpReadINTCReg(XLLP_INTC_ICMR2);
    
    pSave->iclr = XllpReadINTCReg(XLLP_INTC_ICLR);
    pSave->iclr2 = XllpReadINTCReg(XLLP_INTC_ICLR2);


    return XLLP_STATUS_SUCCESS;
}

/// restore state After system wakeup
///
/// @param 
///     pVirtual    INTC base register addresses
///     pSave       Buff to save status
///
/// @return
///     0   -   Successful
///     XLLP_STATUS_WRONG_PARAMETER   -   Fail
///
/// @remarks
///  Assume the system already in required kernel/supervisor mode. 
///  NOT guarantee serialization 
///
XLLP_STATUS_T XllpINTCRestore(P_XLLP_INTC_T pVirtual, P_XLLP_INTC_SAVE_T pSave)
{
    int i=0;

    if(!pVirtual)
        return XLLP_STATUS_WRONG_PARAMETER;

    if(!pSave)
        return XLLP_STATUS_WRONG_PARAMETER;

    pVirtual->iccr = pSave->iccr;
    
    for(i=0;i<32;i++)
        pVirtual->ipr[i] = pSave->ipr[i];  // Interrupt Priority Registers [0-31]
    
    for(i=0;i<21;i++)
        pVirtual->ipr2[i] = pSave->ipr[i+32];  // Interrupt Priority Registers [32-52]
    

    XllpWriteINTCReg(XLLP_INTC_ICMR, pSave->icmr );
    XllpWriteINTCReg(XLLP_INTC_ICMR2,pSave->icmr2);
    
    XllpWriteINTCReg(XLLP_INTC_ICLR,pSave->iclr);
    XllpWriteINTCReg(XLLP_INTC_ICLR2,pSave->iclr2);

    return XLLP_STATUS_SUCCESS;
}
