/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *(C) Copyright 2006 Marvell International Ltd.  
 * All Rights Reserved 
 */

/**
**
**   xllp_tsi.c
**
**   Valid for Monahans touch screen interface
**
*****************************************************************************/
#include "xllp_tsi.h"
#include "xllp_intc.h"
#include "xllp_ost.h"

//fix me:
//The OST is much fast than normal in now platform (prealpha0 silicon). 
//Then the function XllpOstDelayMilliSeconds() may faster than expected.
//So after the OST works properly. function XllpOstDelayMilliSeconds() need to be changed.

//-----------------------------------------------------------------------------------------
// Function: InitializeTSI
// Purpose:  Initialize the TSI 
// 1.	clear all the bits of register ADCD
// 2.	enable pen down event detection 
// 3.	enable valid conversion event detection
// 4.	disable consecutive X and Y measurements
// 5.	set PRED and ADC_DELAY time
// 6.	set resolution  value
//Input Parameters:
// XLLP_TSI_REGS *pTSIregs __ the TSI register base address
//Output Parameters:
//Returns:
//--------------------------------------------------------------------------------------------
void XllpInitializeTSI(XLLP_TSI_REGS_T *pTSIregs)
{
	XLLP_UINT32_T tem;
	pTSIregs->XLLP_ADCD = 0;
	pTSIregs->XLLP_ADCS = 0;
	pTSIregs->XLLP_ADCS = 
		XLLP_TSI_SD_EN | //stylus down/lift interrupt enable
			//when the pen down to the screen or lift from screen 
			//the interrupt will be driven 
		XLLP_TSI_VAL_EN | // conversion valid interrupt enable
		XLLP_TSI_XYM | //enable consecutive x and y measurements 
		XLLP_TSI_PRED |  
		XLLP_TSI_ADC_DELAY; 
	XllpINTCEnableSource(XLLP_INTC_S_TSI,XLLP_TRUE);//enable TSI interrupt
	//read the registers to avoid the split transaction problem.
	tem = pTSIregs->XLLP_ADCD;
	tem = pTSIregs->XLLP_ADCS;
}

//-----------------------------------------------------------------------------------------
// Function: PowerOnTSI
// Purpose:  Power on /off the ADC of TSI 
// Input Parameters:
// XLLP_TSI_REGS *pTSIregs __   TSI register base address
// XLLP_BOOL_T enable __  XLLP_TRUE: power on the ADC;
//  XLLP_FALSE: power off the ADC  
// Output Parameters:
// Returns:
//--------------------------------------------------------------------------------------------
void XllpPowerOnTSI(XLLP_TSI_REGS_T *pTSIregs, XLLP_BOOL_T enable)
{
	XLLP_UINT32_T tem;
	if(enable){
		pTSIregs->XLLP_ADCE |= XLLP_TSI_BIAS_EN;
	}
	else{
		pTSIregs->XLLP_ADCE &= (~XLLP_TSI_BIAS_EN);
	}
	tem =  pTSIregs->XLLP_ADCE;
}

//-----------------------------------------------------------------------------------------
// Function: XllpGetPenEvent
// Purpose:  This function is used to decide if the pen touch/lift event has happened. 
// Input Parameters:
// XLLP_TSI_REGS *pTSIregs __   TSI register base address
// Output Parameters:
//  XLLP_BOOL_T * pPenEvent  __ XLLP_TSI_EVENT_TOUCH: pen down to the screen
//                              XLLP_TSI_EVENT_LIFT: pen lift from the screen. 

// Returns:
//XLLP_STATUS_UNSUPPORTED __ the pen touch/lift event detection is not enabled.
//XLLP_STATUS_SUCCESS

//--------------------------------------------------------------------------------------------
XLLP_STATUS_T XllpGetPenEvent(XLLP_TSI_REGS_T *pTSIregs, XLLP_TSI_EVENT_T *pPenEvent)
{
	XLLP_UINT32_T tem;
	if(pTSIregs->XLLP_ADCS & XLLP_TSI_SD_EN){
		if(pTSIregs->XLLP_ADCD & XLLP_TSI_SD){
			if(pTSIregs->XLLP_ADCD & XLLP_TSI_SV){
				//pen touch
				*pPenEvent = XLLP_TSI_EVENT_TOUCH ;
			}else{
				//pen lift from screen
				*pPenEvent = XLLP_TSI_EVENT_LIFT;
 			}
			pTSIregs->XLLP_ADCD &= (~XLLP_TSI_SD);//clear the SD bit
			tem = pTSIregs->XLLP_ADCD;//read to avoid the spit transaction problem
		}
		else{
			*pPenEvent = XLLP_TSI_EVENT_NONE; //pen event did not happend
		}
	}else{
		pTSIregs->XLLP_ADCD &= (~XLLP_TSI_SD); //clear the SD bit
		return XLLP_STATUS_UNSUPPORTED; //pen down event detection is not enable
	}
	return XLLP_STATUS_SUCCESS;
}

//-----------------------------------------------------------------------------------------
// Function:ReadTSI 
//
// Purpose: 
// The function can get the position value(x , y) and pen state from TSI. 
// The operation process is:
// 1. get the pen state
// 2. disable the pen event detection 
// 3. start a sample routine
// 4. poll the x and y value 
// 5. enabled pen event detection.
//
// Input Parameters:
// XLLP_TSI_REGS *pTSIregs __   TSI register base address
// P_XLLP_OST_T pOstregs __ OST base address
// 
// Output Parameters:
// XLLP_UINT16_T *x __ return x value; if something is wrong or the pen is up, 0xffff will be set to the value
// XLLP_UINT16_T *y __ return y value; if something is wrong or the pen is up, 0xffff will be set to the value
// XLLP_TSI_PEN_STATE_T *pPenState __ XLLP_TSI_PEN_DOWN 
//			 		XLLP_TSI_PEN_UP
// Returns:
//XLLP_STATUS_INVALID_STATE: the pen is opn
//XLLP_STATUS_TIME_OUT 
//XLLP_STATUS_SUCCESS: 
//-----------------------------------------------------------------------------------------
XLLP_STATUS_T XllpReadTSI(
			XLLP_TSI_REGS_T *pTSIregs, 
			P_XLLP_OST_T pOstregs, 
			XLLP_UINT16_T *x, 
			XLLP_UINT16_T *y,
			XLLP_TSI_PEN_STATE_T *pPenState
			)
{
	XLLP_INT_T times=5000;
	XLLP_UINT_T saved_tsi_irq_state;
	*pPenState =(pTSIregs->XLLP_ADCD & XLLP_TSI_SV) ? XLLP_TSI_PEN_DOWN : XLLP_TSI_PEN_UP;
	if(*pPenState == XLLP_TSI_PEN_UP){
		*x =  0xffff;
		*y =  0xffff;
		return XLLP_STATUS_INVALID_STATE;
	}
	saved_tsi_irq_state = XllpReadINTCReg(XLLP_INTC_ICMR2) & (0x10);
	//if the tsi irq is unmask, 
	//before start sample routine we must mask it 
	if(saved_tsi_irq_state){
		XllpINTCEnableSource(XLLP_INTC_S_TSI,XLLP_FALSE); 
	}
	pTSIregs->XLLP_ADCS &= (~XLLP_TSI_SD_EN); //disable the style down event detection
	pTSIregs->XLLP_ADCD &=(~XLLP_TSI_VAL);
	pTSIregs->XLLP_ADCS |= XLLP_TSI_RUN; //start the routine of sample
	while( !(pTSIregs->XLLP_ADCD & XLLP_TSI_VAL) ) {
		XllpOstDelayMilliSeconds(pOstregs, 1);
		times--;
		if (times < 0){
			*x = 0xffff;
			*y = 0xffff;
			
			//fix me: this XLLP_TSI_RUN bit 
			//should be cleard automatically 
			//when synchronized to ADC clock.
			//However I never find this bit 
			//be cleared automatically. So
			//I clear it here.
			pTSIregs->XLLP_ADCS &=(~XLLP_TSI_RUN);
			
			//enable pen event dection
			pTSIregs->XLLP_ADCS |= (XLLP_TSI_SD_EN);
			
			//if the TSI interrupt is not mask before 
			//the sample routine, we must 
			//restore it to unmask
			if(saved_tsi_irq_state){
				XllpINTCEnableSource(XLLP_INTC_S_TSI,XLLP_TRUE);
			}
			return XLLP_STATUS_TIME_OUT;
		}
	}
	pTSIregs->XLLP_ADCD &= (~XLLP_TSI_VAL); //clear the VAL bit
	*x = (XLLP_UINT16_T)(pTSIregs->XLLP_ADCD & MASK_XLLP_TSI_ADC_DATA0);
	*y = (XLLP_UINT16_T)((pTSIregs->XLLP_ADCD & MASK_XLLP_TSI_ADC_DATA1) >> 15);
	//wait for the pen detection circurit to be stable.
	//without the delay, after the pen detection is enable
	//unwanted interrupt may happens.
	//I think it is because that after the conversion of the ADC,the pen 
        //detection circurit will be turn from disable to enable.
	XllpOstDelayMilliSeconds(pOstregs, 5);	

	//fix me: this XLLP_TSI_RUN bit should be cleard automatically 
	//when synchronized to ADC clock.
	//However I never find this bit be cleared automatically. So
	//I clear it here.
	pTSIregs->XLLP_ADCS &=(~XLLP_TSI_RUN);
	//enable the stylus down event dection
	pTSIregs->XLLP_ADCS |= (XLLP_TSI_SD_EN);
	
	//if the TSI interrupt is not mask before the sample routine we must 
	//restore it to unmask
	if(saved_tsi_irq_state){
		XllpINTCEnableSource(XLLP_INTC_S_TSI,XLLP_TRUE);
	}
	
	return XLLP_STATUS_SUCCESS;
}


