/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *(C) Copyright 2006 Marvell International Ltd.  
 * All Rights Reserved 
 */

/**
 **  FILENAME:       xllp_usbohci.c
 **
 **  PURPOSE:        XLLP USB Host APIs - chip specific
 **
 **  Valid for    :  Monahans
 **
 **
 ******************************************************************************/

#include "xllp_usbohci.h"
#include "xllp_usbohci_platform.h"

/*******************************************************************************
 *
 * FUNCTION:         XllpUsbhReset
 *
 * DESCRIPTION:      To Reset PXA27x USB Host
 *                   
 *
 * INPUT PARAMETERS: P_XLLP_USBH_HANDLE_T pUsbhHandle
 * 
 * RETURNS:          
 *       Success:    0 (XLLP_USBH_SUCCESS)
 *       Failure:    
 *
 * CALLS:            XllpOstDelayMicroSeconds, XllpUsbhConfigPlatformSpecificUSBHReset       
 *
 * CALLED BY:        XllpUsbhPowerUp, HCD
 *
 * PROTOTYPE:        XLLP_USBH_ERROR_T XllpUsbhReset(P_XLLP_USBH_HANDLE_T pUsbhHandle);
 *
 *
 *******************************************************************************/

XLLP_USBH_ERROR_T XllpUsbhReset(P_XLLP_USBH_HANDLE_T pUsbhHandle)
{
   // Do the reset for the Bulverde part.
   // Two levels of reset need to be initiated:
   // The OHCI core needs to be reset via the FHR bit,
   // then the OHCI system bus interface needs to be reset via the FSBIR bit.

	// Reset the OHC core and all OHC blocks driven by the 12 MHz clock, eg. write fifo, etc.
	pUsbhHandle->pUSBHRegs->uhchr |=  XLLP_USBOHCI_UHCHR_FHR;

	// Wait for ten micro seconds called for by spec.
	XllpOstDelayMicroSeconds(pUsbhHandle->pOSTRegs, 10) ;

	//Clear FHR bit to enable normal USBH functioning.
	pUsbhHandle->pUSBHRegs->uhchr &= ~XLLP_USBOHCI_UHCHR_FHR;

	// reset the OHC system bus interface, eg. SBI, DMA blocks, fifos, etc.
	pUsbhHandle->pUSBHRegs->uhchr |=  XLLP_USBOHCI_UHCHR_FSBIR;
	
	// auto clears in 3 system bus clocks
	while( pUsbhHandle->pUSBHRegs->uhchr & XLLP_USBOHCI_UHCHR_FSBIR )
	;	    

	//Init Platform Specific Configuration for USBH Reset
	XllpUsbhInitPlatformSpecificConfig(pUsbhHandle);

	//Call for Platform Specific Configuration for USBH Reset
	XllpUsbhConfigPlatformSpecificUSBHReset(pUsbhHandle);

    return (XLLP_USBH_SUCCESS);
}


/*******************************************************************************
 *
 * FUNCTION:         XllpUsbhSelectPowerManagementMode
 *
 * DESCRIPTION:      Used to select power managemen modes like No Power Switching (NPS), 
 *                   PowerSwitchingMode (PSM), Per Port power
 *                   
 *
 * INPUT PARAMETERS: P_XLLP_USBH_HANDLE_T pUsbhHandle
 *          
 *
 * RETURNS:          
 *       Success:    0 (XLLP_USBH_SUCCESS)
 *       Failure:    
 *
 * CALLS:            None. 
 *
 * CALLED BY:        XllpUsbhPowerUp
 *
 * PROTOTYPE:        XLLP_USBH_ERROR_T XllpUsbhSelectPowerManagementMode (
 *								   P_XLLP_USBH_HANDLE_T pUsbhHandle , 
 *								   XLLP_UINT32_T  PowerMode,                           
 *                                 XLLP_UINT32_T  NumPorts, 
 *                                 P_XLLP_UINT32_T  pPortMode)
 *
 *
 *******************************************************************************/

XLLP_USBH_ERROR_T XllpUsbhSelectPowerManagementMode (
								   P_XLLP_USBH_HANDLE_T pUsbhHandle , 
								   XLLP_UINT32_T  PowerMode,                           
                                   XLLP_UINT32_T  NumPorts, 
                                   P_XLLP_UINT32_T  pPortMode)
{
	switch(PowerMode)
	{
	case	XLLP_USBOHCI_PPM_NPS:
				// set NO Power Switching mode
				pUsbhHandle->pUSBHRegs->uhcrhda |= XLLP_USBOHCI_UHCRHDA_NPS;			
				break;

	case	XLLP_USBOHCI_PPM_GLOBAL:
				// make sure the NO Power Switching mode bit is OFF so Power Switching can occur
				// make sure the PSM bit is CLEAR, which allows all ports to be controlled with 
				// the GLOBAL set and clear power commands
				//pUsbhHandle->pUSBHRegs->uhcrhda &= ~(XLLP_USBOHCI_UHCRHDA_NPS|XLLP_USBOHCI_UHCRHDA_PSM);
		        pUsbhHandle->pUSBHRegs->uhcrhda &= ~(XLLP_USBOHCI_UHCRHDA_NPS);
				pUsbhHandle->pUSBHRegs->uhcrhda &= ~(XLLP_USBOHCI_UHCRHDA_PSM);
				break;

	case	XLLP_USBOHCI_PPM_PERPORT:
				// make sure the NO Power Switching mode bit is OFF so Power Switching can occur
				// make sure the PSM bit is SET, which allows all ports to be controlled with 
				// the PER PORT set and clear power commands
				pUsbhHandle->pUSBHRegs->uhcrhda &= ~XLLP_USBOHCI_UHCRHDA_NPS;
				pUsbhHandle->pUSBHRegs->uhcrhda |=  XLLP_USBOHCI_UHCRHDA_PSM;

				// set the power management mode for each individual port to Per Port.
				{
					XLLP_UINT32_T		p;

					for( p = 0; p < NumPorts; p++ )
					{
						pUsbhHandle->pUSBHRegs->uhcrhdb |= (unsigned int)( 1u << (p+17) );	// port 1 begins at bit 17
					}
				}

				break;

	case	XLLP_USBOHCI_PPM_MIXED:
				// make sure the NO Power Switching mode bit is OFF so Power Switching can occur
				// make sure the PSM bit is SET, which allows all ports to be controlled with 
				// the PER PORT set and clear power commands
				pUsbhHandle->pUSBHRegs->uhcrhda &= ~XLLP_USBOHCI_UHCRHDA_NPS;
				pUsbhHandle->pUSBHRegs->uhcrhda |=  XLLP_USBOHCI_UHCRHDA_PSM;

				// set the power management mode for each individual port to Per Port.
				// if the value in the pPortMode array is non-zero, set Per Port mode for the port.
				// if the value in the pPortMode array is zero, set Global mode for the port
				{
					XLLP_UINT32_T		p;

					for( p = 0; p < NumPorts; p++ )
					{
						if( pPortMode[p] )
						{
							pUsbhHandle->pUSBHRegs->uhcrhdb |= (unsigned int)( 1u << (p+17) );	// port 1 begins at bit 17
						}
						else
						{
							pUsbhHandle->pUSBHRegs->uhcrhdb &= ~(unsigned int)( 1u << (p+17) );	// port 1 begins at bit 17
						}

					}
				}

				break;

	}
    return (XLLP_USBH_SUCCESS);
}

/*******************************************************************************
 *
 * FUNCTION:         XllpUsbhPowerUp
 *
 * DESCRIPTION:      Called to power on USB Host at bootup and at resume from sleep
 *                   
 *
 * INPUT PARAMETERS: P_XLLP_USBH_HANDLE_T pUsbhHandle
 *          
 *
 * RETURNS:          
 *       Success:    0 (XLLP_USBH_SUCCESS)
 *       Failure:    
 *
 * CALLS:            XllpUsbhTurnOnUSBHostClock,  XllpUsbhPortEnableOverCurrentNotification, 
 *                   XllpUsbhPortEnablePower, XllpUsbhSelectPowerManagementMode,
 *                   XllpUsbhReset
 *
 * CALLED BY:        HCD.
 *
 * PROTOTYPE:        XLLP_USBH_ERROR_T XllpUsbhPowerUp(P_XLLP_USBH_HANDLE_T pUsbhHandle);
 *
 *
 *******************************************************************************/

XLLP_USBH_ERROR_T XllpUsbhPowerUp(P_XLLP_USBH_HANDLE_T pUsbhHandle)
{
	// this sets up Port 1 Pwr notification using gpio3_2  as input in alternate function 1 mode
	XllpUsbhPortEnableOverCurrentNotification(pUsbhHandle, XLLP_USBH_PORT_1);

	// this sets up Port 1 Pwr enable using gpio2_2 as output in alternate function 2 mode
	XllpUsbhPortEnablePower(pUsbhHandle, XLLP_USBH_PORT_1);

	XllpUsbhSelectPowerManagementMode(pUsbhHandle, XLLP_USBOHCI_PPM_NPS, 0, 0 );

	XllpUsbhReset(pUsbhHandle);

    return (XLLP_USBH_SUCCESS);
}

/*******************************************************************************
 *
 * FUNCTION:         XllpUsbhPowerDown
 *
 * DESCRIPTION:      Called to put USB Host into sleep mode
 *                   
 *
 * INPUT PARAMETERS: P_XLLP_USBH_HANDLE_T pUsbhHandle
 *          
 *
 * RETURNS:          
 *       Success:    0 (XLLP_USBH_SUCCESS)
 *       Failure:    
 *
 * CALLS:            XLLP_USBOHCI_UHCRHS_MWRITE   
 *
 * CALLED BY:        HCD.
 *
 * PROTOTYPE:        XLLP_USBH_ERROR_T XllpUsbhPowerDown(P_XLLP_USBH_HANDLE_T pUsbhHandle);
 *
 *
 *******************************************************************************/

XLLP_USBH_ERROR_T XllpUsbhPowerDown(P_XLLP_USBH_HANDLE_T pUsbhHandle)
{

	//Check this whether its enough
	//set global power switch mode
	pUsbhHandle->pUSBHRegs->uhcrhda  &= ~ (XLLP_USBOHCI_UHCRHDA_NPS | XLLP_USBOHCI_UHCRHDA_PSM);  

    //Clear global power
	XLLP_USBOHCI_UHCRHS_MWRITE( pUsbhHandle, XLLP_USBOHCI_UHCRHS_CGP); 

    return (XLLP_USBH_SUCCESS);
}

#if 0
static XLLP_UINT32_T xllp_usbh_port3_mfp_list[] = {
	XLLP_MFP_PIN_GPIO75_OFFSET,
	XLLP_MFP_PIN_GPIO76_OFFSET,
	XLLP_MFP_PIN_GPIO77_OFFSET,
	XLLP_MFP_PIN_GPIO78_OFFSET,
	XLLP_MFP_PIN_GPIO79_OFFSET,
	XLLP_MFP_PIN_GPIO80_OFFSET,
	XLLP_MFP_PIN_EOLIST_MARKER
};

static XLLP_MFP_ALT_FN_T  xllp_usbh_port3_af_list[] = {
	XLLP_MFP_PIN_GPIO75_AF_USB_P3_1,
	XLLP_MFP_PIN_GPIO76_AF_USB_P3_2,
	XLLP_MFP_PIN_GPIO77_AF_USB_P3_3,
	XLLP_MFP_PIN_GPIO78_AF_USB_P3_4,
	XLLP_MFP_PIN_GPIO79_AF_USB_P3_5,
	XLLP_MFP_PIN_GPIO80_AF_USB_P3_6
};

static XLLP_MFP_DRIVE_STRENGTH_T xllp_usbh_port3_ds_list[] = {
	XLLP_MFP_DEFAULT_DS,
	XLLP_MFP_DEFAULT_DS,
	XLLP_MFP_DEFAULT_DS,
	XLLP_MFP_DEFAULT_DS,
	XLLP_MFP_DEFAULT_DS,
	XLLP_MFP_DEFAULT_DS
};
#endif	

/*******************************************************************************
 *
 * FUNCTION:         XllpUsbhConfigPort3
 *
 * DESCRIPTION:      Called to configure port3
 *                   
 *
 * INPUT PARAMETERS: P_XLLP_USBH_HANDLE_T pUsbhHandle
 *		     XLLP_USBH_PORT3_MODE_T mode	
 *          
 *
 * RETURNS:          
 *       Success:    0 (XLLP_USBH_SUCCESS)
 *       Failure:    XLLP_USBH_PORT3_MODE_UNVALID
 *
 * CALLS:            XllpMfpSetAfDs_List
 *
 * CALLED BY:        HCD.
 *
 * PROTOTYPE:        XLLP_USBH_ERROR_T XllpUsbhConfigPort3(P_XLLP_USBH_HANDLE_T pUsbhHandle, XLLP_USBH_PORT3_MODE_T mode);
 *
 *
 *******************************************************************************/
#if 0
XLLP_USBH_ERROR_T XllpUsbhConfigPort3(P_XLLP_USBH_HANDLE_T pUsbhHandle, XLLP_USBH_PORT3_MODE_T mode)
{
	XLLP_STATUS_T status;
	
	if ((XLLP_USBH_PORT3_EXT_TRANSCEIVER == mode) ||\
			(XLLP_USBH_PORT3_EXT_DEVICE == mode)) {
		pUsbhHandle->pUdcRegs->up3ocr = mode;
		
		status=XllpMfpSetAfDs_List( pUsbhHandle->pMfpRegs, xllp_usbh_port3_mfp_list, 
				xllp_usbh_port3_af_list, xllp_usbh_port3_ds_list);
		return status;
	}

    return (XLLP_USBH_PORT3_MODE_UNVALID);
}

#endif
/*******************************************************************************
 *
 * FUNCTION:         XllpUsbhPortEnableSleepStandbyEnable
 *
 * DESCRIPTION:      Enable or disable power to the USB single-ended receivers
 *                   and the USB port [1, 2 or 3] power supply 
 *                   
 *
 * INPUT PARAMETERS: P_XLLP_USBH_HANDLE_T pUsbhHandle
 *                   XLLP_USBH_PORT_T Port -     Port 1  (XLLP_USBH_PORT_1)
 *                                             or Port 2 (XLLP_USBH_PORT_2)
 *											   or Port 3 (XLLP_USBH_PORT_3)
 *
 * RETURNS:          
 *       Success:    0 (XLLP_USBH_SUCCESS)
 *       Failure:    
 *
 * CALLS:            None 
 *
 * CALLED BY:        XllpUsbhConfigPlatformSpecificUSBHReset
 *
 * PROTOTYPE:        XLLP_USBH_ERROR_T XllpUsbhPortEnableSleepStandbyEnable(
 *                                                     P_XLLP_USBH_HANDLE_T pUsbhHandle, 
 *													   XLLP_USBH_PORT_T Port, 
 *													   XLLP_BOOL_T enable);
 *
 *
 *******************************************************************************/

XLLP_USBH_ERROR_T XllpUsbhPortEnableSleepStandbyEnable(P_XLLP_USBH_HANDLE_T pUsbhHandle, 
													   XLLP_USBH_PORT_T Port, 
													   XLLP_BOOL_T enable)
{
	switch(Port)
	{
	  case	XLLP_USBH_PORT_1:		
		if (XLLP_TRUE == enable)
		{
		  pUsbhHandle->pUSBHRegs->uhchr &= ~XLLP_USBOHCI_UHCHR_SSEP1;
		}
		else
		{
		  pUsbhHandle->pUSBHRegs->uhchr |= XLLP_USBOHCI_UHCHR_SSEP1;
		}
		break;


	  case	XLLP_USBH_PORT_2:		
		if (XLLP_TRUE == enable)
		{
		  pUsbhHandle->pUSBHRegs->uhchr &= ~XLLP_USBOHCI_UHCHR_SSEP2;
		}
		else
		{
		  pUsbhHandle->pUSBHRegs->uhchr |= XLLP_USBOHCI_UHCHR_SSEP2;
		}
		break;


	  case	XLLP_USBH_PORT_3:		
		if (XLLP_TRUE == enable)
		{
		  pUsbhHandle->pUSBHRegs->uhchr &= ~XLLP_USBOHCI_UHCHR_SSEP3;
		}
		else
		{
		  pUsbhHandle->pUSBHRegs->uhchr |= XLLP_USBOHCI_UHCHR_SSEP3;
		}
		break;
	}
    return (XLLP_USBH_SUCCESS);
}

XLLP_USBH_ERROR_T XllpUsbhConfigureGlobalSleepStandbyEnable(P_XLLP_USBH_HANDLE_T pUsbhHandle, 
				  XLLP_BOOL_T enable)
{
	if (XLLP_TRUE == enable)
    {
      pUsbhHandle->pUSBHRegs->uhchr &= ~XLLP_USBOHCI_UHCHR_SSE;
	}
	else
	{
	  pUsbhHandle->pUSBHRegs->uhchr |= XLLP_USBOHCI_UHCHR_SSE;
	}

    return (XLLP_USBH_SUCCESS);
}



XLLP_USBH_ERROR_T XllpUsbhConfigurePowerSensePolarity(P_XLLP_USBH_HANDLE_T pUsbhHandle, 
					  XLLP_LEVEL_T level)
{
	if (XLLP_HI == level)
    {
      pUsbhHandle->pUSBHRegs->uhchr &= ~XLLP_USBOHCI_UHCHR_PSPL;
	}
	else
	{
	  pUsbhHandle->pUSBHRegs->uhchr |= XLLP_USBOHCI_UHCHR_PSPL;
	}

    return (XLLP_USBH_SUCCESS);
}


XLLP_USBH_ERROR_T XllpUsbhConfigurePowerControlPolarity(
			P_XLLP_USBH_HANDLE_T pUsbhHandle, 
			XLLP_LEVEL_T level)
{
	if (XLLP_HI == level)
    {
      pUsbhHandle->pUSBHRegs->uhchr &= ~XLLP_USBOHCI_UHCHR_PCPL;
	}
	else
	{
	  pUsbhHandle->pUSBHRegs->uhchr |= XLLP_USBOHCI_UHCHR_PCPL;
	}

    return (XLLP_USBH_SUCCESS);
}



