/*
 * usb_serial_download.c: serial over usb download command
 *
 * Copyright (c) 2004, Intel Corporation (alek.du@intel.com)
 *
 * (C) Copyright 2006 Marvell International Ltd.  
 * All Rights Reserved 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ident "$Id: usb_serial_download.c,v 1.6 2004/09/07 02:44:16 bvd Exp $"

#ifdef HAVE_CONFIG_H
# include <blob/config.h>
#endif

#include <blob/download.h>
#include <blob/command.h>
#include <blob/errno.h>
#include <blob/error.h>
#include <blob/md5.h>
#include <blob/md5support.h>
#include <blob/serial_usb.h>
#include <blob/partition.h>

#ifdef CONFIG_MD5_SUPPORT
#define MD5TAG	"md5:\r"
#define MD5SIZE	16
#endif

#define STARTAG "start"
#define ENDTAG "end\r"
#define EXITAG "exit"
#define ACKTAG "ACK\r"
#define NAKTAG "NAK\r"
#define COMMANDSIZE 64

#define strsizeof(x) (sizeof(x)-1)

/* receive a image file to dedicated memory address
    support md5 check now, if you send image as this
    |data of packet1|16 bytes md5 digest|md5:\r|
    |data of packet2|16 bytes md5 digest|md5:\r| 
    			...
    |data of packetn|16 bytes md5 digest|md5:\r|
			....
*/

int UsbSerialReceive(char *addr,int len)
{
	int size,total=0;// for image size 
	int lastsize=0;// for progress bar				
	int packsize=0;//current packsize
	char command[COMMANDSIZE+1];// keep received command
	int comlen=0;// current command len

	while(1){
		comlen=0;
		while(comlen<COMMANDSIZE){
			size=usb_serial_read(command+comlen,COMMANDSIZE-comlen);
			if (size){
				comlen+=size;
				command[comlen]='\0';
				if (command[comlen-1]=='\r'){
					comlen--;
					command[comlen]='\0';
					break;
				}
				else printf("%s",command+comlen-size);
			}
		}
		printf("\ncurrent command:(%s)\n",command);	
		
		if (comlen==strsizeof(STARTAG) && !strncmp(command,STARTAG, strsizeof(STARTAG))){
			usb_serial_write(ACKTAG,strsizeof(ACKTAG));
			printf("start to receive binary images\n");
			printf("please send 'end\\r' as finish tag\n");//fix me: hope no end\\r in image file
			total=lastsize=packsize=0;
			while (total+packsize<len){
#ifdef CONFIG_MD5_SUPPORT				
				u32 md5_count[MD5SIZE/4];
#endif				
				size=usb_serial_read(addr+total+packsize,len-(total+packsize));
				if (size){
					packsize+=size;
					if (packsize-lastsize>=65535)
						lastsize=packsize,printf(".");
					if (*(addr+total+packsize-1)=='\r'){
#ifdef CONFIG_MD5_SUPPORT						
						if (packsize>=MD5SIZE+strsizeof(MD5TAG)&& !strncmp(addr+total+packsize-strsizeof(MD5TAG),MD5TAG,strsizeof(MD5TAG))){
							printf("*");
							md5_buffer(addr+total,packsize-(MD5SIZE+strsizeof(MD5TAG)),md5_count);//md5
							if (!memcmp(addr+total+packsize-(MD5SIZE+strsizeof(MD5TAG)),md5_count,MD5SIZE)){//md5 right
								total+=packsize-(MD5SIZE+strsizeof(MD5TAG));													
								usb_serial_write(ACKTAG,strsizeof(ACKTAG));
							}
							else {
								printf("x");
								usb_serial_write(NAKTAG,strsizeof(NAKTAG));
							}
							packsize=lastsize=0;
						}
						else 
#endif							
						if ((unsigned int)packsize>=strsizeof(ENDTAG)&& !strncmp(addr+total+packsize-strsizeof(ENDTAG),ENDTAG,strsizeof(ENDTAG))){
							total+=packsize-strsizeof(ENDTAG);//last packet no MD5? or all data just in one packet and without md5
							usb_serial_write(ACKTAG,strsizeof(ACKTAG));
							break;
						}
					}
				}
			}
			printf("end tag received, total (%d) bytes got\n",total);
		}		
		else if (comlen==strsizeof(EXITAG) && !strncmp(command,EXITAG, strsizeof(EXITAG))){
			usb_serial_write(ACKTAG,strsizeof(ACKTAG));
			printf("exit tag received, total (%d) bytes got finially\n",total);
			break;
		}
		else {
			usb_serial_write(NAKTAG,strsizeof(NAKTAG));
			printf("Unknown command.\n");
		}
	}
	return total;
}

int udownload_cmd(int argc, char *argv[])
{
	u32 startAddress;
	int bufLen;
	int *numRead;
	int retval;
	u32 *digest;
	static int firstin=1;
	
	if(argc < 2)
		return -ENOPARAMS;

	retval = set_download_parameters(argv[1], &startAddress,
					 &bufLen, &numRead, &digest);
	if (retval < 0)
		return 0;

	printf("Entering serial over usb commandline... \n");
	printf("App protocol:\n"
		"APP                     BLOB\n"
		"start\\r   -------->            \n"
		"               <--------ACK\\r \n"
		"binary image ------->        \n"
		"end\\r  --------->             \n"
		"               <--------ACK\\r \n"
		"exit\\t ---------->            \n"
		"               <--------ACK\\r \n"
		"\n"
		"\n"
		"App first send start\\r to request to start transfer image \n"
		"after receive \"ACK\\r\" from BLOB then app send binary image.\n"
		"App send end\\r to indicate end of image.\n"
		"Also BLOB will send \"ACK\\r\" to App.\n"
		"App send exit\\r to leave udownload. \n");

	if (firstin){
		usb_serial_init();
		firstin=0;
	}
	//max 16M
	bufLen=0x1000000;
#ifdef CONFIG_MD5_SUPPORT
	bufLen+=MD5SIZE+strsizeof(MD5TAG);
#endif
	*numRead = UsbSerialReceive((char *) startAddress, bufLen);

	printf("Received %i (0x%08x) bytes ", *numRead, *numRead);
	printf("at 0x%08x\n", (unsigned int)startAddress);
	return 0;
}

char udownload_help[] = "udownload {blob|param|kernel|ramdisk}\n"
"Download <argument> image to RAM using serial over usb\n";


