/*
 * ether-s3c2500.c: Low-level ethernet polling functions
 *
 * Copyright (c) 2003  Russ Dill <Russ.Dill@asu.edu>
 *
 * smc9194.c was used as a skeleton for this file
 * Original Author:
 * 	Erik Stahlman				( erik@vt.edu )
 * contributors:
 *      Arnaldo Carvalho de Melo <acme@conectiva.com.br>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <util.h>
#include <time.h>
#include <errno.h>
#include <serial.h>
#include <command.h>
#include <net/s3c2500.h>
#include <net/ether.h>

#undef SE_DEBUG

#ifdef SE_DEBUG
#	define DBG( x, args... )	if ( dbg>=x ) printf( args )
#else
#	define DBG( x, args... ) ;
#endif

/* 
 * Because this is being done with polling, and packets can really
 * pile up, especially little packets on big networks, make a ring
 * of 16 little buffers (maybe more later). Packets under 160 will
 * fit in one, larger packets will fit across multiple buffers.
 * This way, we are much less likely to miss out precious tftp
 * packets
 */

#define MAX_FRAME (6 + 6 + 2 + 1500 + 22)
#define PKT_SIZE 160
#define NFRAMESB 4
#define NFRAMES (1 << NFRAMESB)

struct buffer_t tx_buffer, rx_buffer[NFRAMES];

/* rx buffers */
u8 pkts[NFRAMES][PKT_SIZE];

static int dbg = 0;
static u32 s3c2500_ether_chip_base = 0;

/* the current rx descriptor that the ethernet chip is about to or
 * currently filling */
static int curr_desc;

/* Write out a word */
static inline void se_outw(u32 addr, u32 data)
{
	*((volatile u32 *) addr) = data;
}

/* Read in a word */
static inline u32 se_inw(u32 addr)
{
	return *((volatile u32 *) addr);
}

/* Set bits in a register */
static inline void se_bset(u32 addr, u32 bit)
{
	se_outw(addr, se_inw(addr) | bit);
}


/* retreive the address from the CAM */
static int s3c2500_ether_get_addr(u8 dev_addr[6])
{
	u32 base;

	if (!(base = s3c2500_ether_chip_base))
		return -EINVAL;

	memcpy(dev_addr, (u8 *) (base + CAM), 6);
	DBG(1, "%s: got eth addr %02x:%02x:%02x:%02x:%02x:%02x\n", __FUNCTION__,
			dev_addr[0], dev_addr[1], dev_addr[2], dev_addr[3], dev_addr[4],
			dev_addr[5]);

	return 0;
}


/* Write a hardware address to the CAM */
static int s3c2500_ether_set_addr(u8 dev_addr[6])
{
	u32 base;

	if (!(base = s3c2500_ether_chip_base))
		return -EINVAL;

	memcpy((u8 *) (base + CAM), dev_addr, 6);
	DBG(1, "%s: set eth addr to %02x:%02x:%02x:%02x:%02x:%02x\n", __FUNCTION__,
			dev_addr[0], dev_addr[1], dev_addr[2], dev_addr[3], dev_addr[4],
			dev_addr[5]);
	return 0;
}


/* Receive a (one) packet up to length bytes, otherwise, discard */
static int s3c2500_ether_rcv(u8 *data, int length)
{
	u32 stat, len, base, pos, desc, copy_len;
	int ret = 0, i;
	u8 *copy_base;

	if (!(base = s3c2500_ether_chip_base))
		return -EINVAL;

	i = curr_desc;

	/* do we own it yet? */
	if (rx_buffer[i].desc & BDMA_owner) return 0;

	/* Read/clear stat */
	stat = se_inw(base + BMRXSTAT);
	se_outw(base + BMRXSTAT, 0);

	/* do we have a packet? */
	if (!(rx_buffer[i].desc & BRxDone)) return 0;

	pos = 0;
	len = rx_buffer[i].desc & 0xffff;
	len -= 4; /* Strip the FCS */

	do {
		desc = rx_buffer[i].desc;

		/* only do the copying if it will fit */
		if (len <= length) {
			copy_base = rx_buffer[i].ptr;
			if (desc & DescEOF) copy_len = len - pos;
			else copy_len = PKT_SIZE;
			if (!pos) {
				copy_base += 2;
				copy_len -= 2;
			}
			memcpy(data + pos, copy_base, copy_len);
			pos += copy_len;
		}
		i = (i + 1) % NFRAMES;
	} while (!(desc & DescEOF));

	/* Reenable used (or discarded) for reception */
	do {
		rx_buffer[curr_desc++].desc = BDMA_owner;
		if (curr_desc >= NFRAMES) curr_desc = 0;
	} while (i != curr_desc);

	/* Make sure transmission is still enabled */
	se_bset(base + BDMARXCON, BRxEn);
	se_bset(base + MACRXCON, MRxEn);

	return (len <= length) ? len : 0;
}


/* Send a packet of length bytes (wait on completion) */
static int s3c2500_ether_snd(u8 *data, int length)
{
	u32 base;

	if (!(base = s3c2500_ether_chip_base))
		return -EINVAL;

	memcpy(data + 6, hwaddress, 6);

	if (length > 1514) {
		printf("%s: Too big packet error.\n", __FUNCTION__);
		return -1;
	}

	/* wait for the descriptor to free up */
	while (tx_buffer.desc & BDMA_owner);

	/* Setup the descriptor */
	tx_buffer.ptr = (u8 *) data;
	tx_buffer.desc = length | ((((u32) data) % 4) << 16) | BDMA_owner;

	/* Make sure everything is ready to go */
	se_bset(base + BDMATXCON, BTxEn);
	se_bset(base + MACTXCON, MTxEn);

	/* Wait and see if it goes out */
	while (!(se_inw(base + BMTXSTAT) & TxComp));
	
	/* reset the stats */
	se_outw(base + BMTXSTAT, 0x0);

	/* give it back */
	tx_buffer.desc = 0;

	return 0;
}


/* Reset everything */
static void s3c2500_ether_reset(u32 base)
{
	int i;
	/* Clear out all the registers left by an improper reset
	 * (see ethernet errata) or a simple jump to 0x0 */
	for (i = BDMATXCON; i <= BDMARXLEN; i += 4)
		se_outw(base + i, 0);
	for (i = MACCON; i < STACON; i += 4)
		se_outw(base + i, 0);
	se_outw(base + STACON, 0x00006000);
	se_outw(base + CAMEN, 0);
	se_outw(base + MISSCNT, 0);
	for (i = CAM; i < CAM + 0x80; i += 4)
		se_outw(base + i, 0);

	/* Reset the TX and RX block */
	se_outw(base + BDMATXCON, BTxRS);
	se_outw(base + BDMARXCON, BRxRS);

	/* Reset the MII */
	se_outw(base + MACCON, MReset);

	DBG(1, "%s: reset done\n", __FUNCTION__);
}


/* Enable everything */
static void s3c2500_ether_enable(u32 base)
{
	int i;
	/* Set the MAC to full duplex */
	se_outw(base + MACCON, MACFD);

	/* give it the size of our receive buffer */
	se_outw(base + BDMARXLEN, PKT_SIZE | MAX_FRAME << 16);

	/* point the chip to our buffers */
	se_outw(base + BDMATXDPTR, (u32) &tx_buffer);
	se_outw(base + BDMARXDPTR, (u32) &rx_buffer);
	
	/* setup our buffer descriptors */
	tx_buffer.ptr = 0;
	tx_buffer.desc = 0;
	for (i = 0; i < NFRAMES; i++) {
		rx_buffer[i].ptr = (u8 *) pkts[i];
		rx_buffer[i].desc = BDMA_owner;
	}
	curr_desc = 0;
	
	/* put an address in the CAM */
	s3c2500_ether_set_addr(hwaddress);

	/* Enable reception of our CAM */
	se_outw(base + CAMEN, 1);

	/* Enable reception modes (MAC matching) */
	se_outw(base + CAMCON, MBroad | MCompEn | MStation);

	/* setup BDMA and enable reception/transmition */
	se_outw(base + BDMATXCON, 0x6 << BTxMSL);
	se_outw(base + MACTXCON, MTxEn);

	se_outw(base + BDMARXCON, BRxEn | (0x2 << BRxMSL) | 
				  (NFRAMESB << BRxNBD));
	se_outw(base + MACRXCON, MRxEn);

	DBG(1, "%s: chip enabled.\n", __FUNCTION__);
}


static int s3c2500_ether_init(u32 base)
{

	DBG(1, "%s(%08x)\n", __FUNCTION__, base);

	s3c2500_ether_chip_base = base;

	s3c2500_ether_reset(base);
	s3c2500_ether_enable(base);

	return 0;
}


/* export ethernet driver */
ether_driver_t s3c2500_ether_driver = {
	init:		s3c2500_ether_init,
	rcv:		s3c2500_ether_rcv,
	snd:		s3c2500_ether_snd,
	get_addr:	s3c2500_ether_get_addr,
	set_addr:	s3c2500_ether_set_addr,
	base:		0xF00A0000,
};
