/**********************************************************************
 * blob partition reading for GIO
 *
 * Copyright (C) 2003, Stefan Eletzhofer <stefan.eletzhofer@eletztrick.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: gio_part_blob.c,v 1.1 2003/04/03 14:32:50 seletz Exp $
 *
 */

/**********************************************************************
 * Includes
 */
#ifdef HAVE_CONFIG_H
# include <blob/config.h>
#endif

#include <blob/types.h>
#include <blob/errno.h>
#include <blob/util.h>
#include <blob/command.h>

#include <blob/init.h>
#include <blob/partition.h>
#include <blob/generic_io.h>

/**********************************************************************
 * defines
 */
#define DEBUG 0

#if DEBUG
#	define DBG( x, args... )	if ( dbg>=x ) printf( args );
#else
#	define DBG( x, args... )	
#endif

/**********************************************************************
 * module globals
 */
#if DEBUG
static int dbg = 1;
#else
static int dbg = 0;
#endif

static char module_version[] = "$Id: gio_part_blob.c,v 1.1 2003/04/03 14:32:50 seletz Exp $";

/**********************************************************************
 * static functions
 */

/**********************************************************************
 * blob partition io driver
 *
 * - default blob partition is registered as a default io driver
 * - the configure call "selects" a partition by it's name.
 *   (just like the tar io driver)
 */
static int part_io_init( io_driver_t *io );
static int part_io_conf( io_driver_t * io, void *conf );
static int part_io_read( unsigned char *dest, unsigned char *src, size_t amount, io_driver_t *io );
static int part_io_write( unsigned char *dest, unsigned char *src, size_t amount, io_driver_t *io );

/* initialize driver struct */
static int part_io_init( io_driver_t *io )
{
	if ( !io )
		return -EINVAL;

	io->private_data = NULL;
	io->io_size = 0;
	io->conf = part_io_conf;
	io->read = part_io_read;
	io->write = part_io_write;
	io->child_io = NULL;

	return 0;
}

/* configure: select a partition by name */
static int part_io_conf( io_driver_t * io, void *conf )
{
	char *name;
	const blob_partition_t *part;

	if ( !io || !conf )
		return -EINVAL;

	name = (char *)conf;
	if ( !name )
		return -EINVAL;

	part = pt_find_by_name( name );
	if ( !part )
		return -EINVAL;

	DBG( 1, "%s: found partition '%s': offset=%x, size=%x, flags=%x\n", __FUNCTION__,
			part->name, part->offset, part->size, part->flags );


	io->io_size=part->size;
	io->private_data=(void*)part;

	return 0;
}

static int part_io_read( unsigned char *dest, unsigned char *src, size_t amount, io_driver_t *io )
{
	blob_partition_t *part;
	io_driver_t *child;

	DBG( 1, "%s: dest=%p, src=%p, amount=%d, io=%p\n", __FUNCTION__, 
			dest, src, amount, io );

	if ( !io || !dest )
		return -EINVAL;

	child = io_get_child( io );
	if ( !child )
		return -EINVAL;

	part = (blob_partition_t *)io->private_data;
	if ( !part )
		return -EINVAL;

	DBG( 1, "%s: using partition '%s'.\n", __FUNCTION__, part->name );
	DBG( 1, "%s: dest=%p, src=%p(=%p), amount=%d, io=%p\n", __FUNCTION__, 
			dest, src, (void*)(src + part->offset), amount, io );

	/* defer reading to our child */
	return io_read(dest, (void *)((u32)src + part->offset), amount, child);
}

static int part_io_write( unsigned char *dest, unsigned char *src, size_t amount, io_driver_t *io )
{
	blob_partition_t *part;
	io_driver_t *child;


	if ( !io || !src )
		return -EINVAL;

	child = io_get_child( io );
	if ( !child )
		return -EINVAL;

	part = (blob_partition_t *)io->private_data;
	if ( !part )
		return -EINVAL;

	DBG( 1, "%s: using partition '%s'.\n", __FUNCTION__, part->name );
	DBG( 1, "%s: dest=%p(=%p), src=%p, amount=%d, io=%p\n", __FUNCTION__, 
			dest, (void*)(dest + part->offset), src, amount, io );

	/* defer writing to our child */
	return io_write((void *)((u32)dest + part->offset), src, amount, child);
}


/* register default blob partition io driver */
void init_part_io( void )
{
	int ret;
	static io_driver_t io_part;
	
	ret = part_io_init( &io_part );
	if ( ret ) return ;

	ret = io_register( &io_part, "BLOB_PART" ); 
	if ( ret ) return;

	return;
}
