/****************************************************************************/
/* FILENAME:		RB_Check.c		*/
/*					*/
/* DESCRIPTION:	Sample code for Memory (RAM + Flash) 	*/
/*			check used by Red Bend UPI	*/
/*					*/
/* Copyright Red Bend Software.  All rights reserved.	*/
/****************************************************************************/

#include "RB_Check.h"
#include "RB_ImageUpdate.h"
#include "fota_const.h"

const Byte memValue = 0x55;
const Byte imageValue= 0xdd;
const Byte backupImgValue = 0xee;
const Dword diff = 0x10;
const Dword oddSize = 0x15;
const Dword evenSize = 0x16;

#define NULL 0
#define SUCCESSFUL 0
#define FAIL 1
#define RANDOM_VALUE 0x1234



int check()
{
   RB_Trace(NULL, "start fota check.\r\n");
   RB_Trace(NULL, "%s", "Test string output.\r\n");
   RB_Trace(NULL, "The value you should see is 0x00004000, 0x%08x\r\n", 0x4000);	
  
   Byte* pRam = (Byte*)RAM_START_ADDRESS;
   Dword ramSize = RAM_SIZE;
   Byte* pInner = (Byte*)0x5c008000;
   Dword innerSize = RAM_INTERNAL_SIZE;
   Byte* pBack[3];
      
   pBack[0] = UPDATE_BACKUP_START_ADDRESS;
   pBack[1] = UPDATE_BACKUP_START_ADDRESS + FLASH_BLOCK_SIZE;
   pBack[2] = UPDATE_BACKUP_START_ADDRESS + 2 * FLASH_BLOCK_SIZE;

   RB_Trace(NULL, "Parameters to the function are: start 0x%08x, pBack0 0x%08x, pBack1 0x%08x, pBack2 0x%08x\r\n",
						 pRam, pBack[0],pBack[1], pBack[2]);
   
   Dword bufCount = 3;
   RB_Check(IMAGE_START_ADDRESS,
        IMAGE_END_ADDRESS,
        pRam,
        ramSize,
        NULL,
        0,
        pBack,
        bufCount);
        
   RB_Trace(NULL, "end fota check.\n");
   
   
   return 0;
}




/****************************************************************************/
/* NAME:		RB_Check		*/
/*					*/
/* DESCRIPTION:			*/
/*	This function returns 0 if all the checks are valid otherwise 	*/
/*	it return a <> 0 value.		*/
/*	It checks that all Red Bend API functions are correctly implemented.	*/
/*	First it checks that all of the RAM memory can be written to	*/
/*	(and also, if exists, internal RAM memory).	*/
/*	Then it checks that the read from flash are implemented correctly	*/
/*	reading a whole sector and then from	*/
/*	odd/even addresses, of different sizes. At the end it checks that we	*/
/*	can read and write to all the given backup blocks.	*/
/*					*/
/* INPUTS:				*/
/*	start - start address of image to update	*/	
/*	end - end address of image to update	*/
/*	pRAM - pointer for RAM to use		*/
/*	dwRAMSize - size of RAM to use  	*/
/*	pbyInternalRAM - pointer for internal RAM to use	*/
/*	dwInternalRAMSize - size of internal RAM to use	*/
/*	buff_start - an array size of [num_of_buff] which points to	*/
/*			the area in the flash where RED BEND can backup 	*/
/*	num_of_buff - number of backup buffers	*/
/*					*/
/* OUTPUTS:		None		*/
/*					*/
/*					*/
/* NOTES: 				*/
/*	If the image has regions that are not cannot be written 	*/
/*	(for ex. Update Agent area) by RB_WriteImage function, you need 	*/
/*	to use RB_DefineWritingExcludeRegions function	*/
/*	to define the number of areas and their start address and size.	*/
/*					*/
/****************************************************************************/

int RB_Check (Dword start, // start address of app to update
				Dword end,	  // end address of app to update
				Byte* pRAM, // pointer for the ram to use
				Dword dwRAMSize,
				Byte* pbyInternalRAM,
				Dword dwInternalRAMSize,
				Dword* buff_start,
				Dword num_of_buff // number of back up buffers
				)

{
	Dword i;
    
     
	/* 	RB_GetBlockSize is an API function which should be implemented 	*/
	RB_Trace(NULL, "before GetBlockSize.\r\n");
	int sectorSize = RB_GetBlockSize();
	RB_Trace(NULL, "After GetBlockSize.\r\n");
	
	/* 	Define areas in memory which you don't want to write to	*/
	RB_DefineWritingExcludeRegions();
    RB_Trace(NULL, "after DefineWritingExcludeRegions.\r\n");

	/* 	check that stack and parsing of trace are working well	*/
	/* 	RB_Trace is an API function which should be implemented	*/
    RB_Trace(NULL, "%s", "Test string output.\r\n");
	RB_Trace(NULL, "The value you should see is 0x00004000, 0x%08x\r\n", 0x4000);	
	RB_Trace(NULL, "Parameters to the function are: start 0x%08x, end 0x%08x, pRam 0x%08x, dwRamSize 0x%08x\r\n",
						 start, end, pRAM, dwRAMSize);
	RB_Trace(NULL, "pbyInternalRAM 0x%08x, dwInternalRAMSize 0x%08x, buff_start 0x%08x, num_of_buff 0x%08x\r\n ",
					pbyInternalRAM, dwInternalRAMSize, buff_start, num_of_buff);
    RB_Trace(NULL,"buff_start1=%x,buff_start2=%d\r\n",buff_start[0],buff_start[1]);

	/* 	check size of image	*/
	RB_Trace(NULL, "!!!!!!!!!Check 1 yx\r\n");
	if ((int)(end - start) < sectorSize)
	{
		RB_Trace(NULL, "error Image size can not be less than sector size [=0x%08x]\r\n", sectorSize);
		return (FAIL);
	}
	
	/* 	check RAM memory		*/
	RB_Trace(NULL, "!!!!!!!!!Check 2 yx\r\n");
	if (RB_CheckRamMemory(pRAM, dwRAMSize))
	{
	     RB_Trace(NULL,"error RB_CheckRamMemory:pRAM=%x,dwRAMSize=%x\r\n",pRAM, dwRAMSize);
		return (FAIL);
	}
	
	/* 	check internal RAM memory	*/
	RB_Trace(NULL, "!!!!!!!!!Check 3 yx\r\n");
	if (RB_CheckInternalRAMMemory(pbyInternalRAM, dwInternalRAMSize))
	{
	           RB_Trace(NULL,"error RB_CheckInternalRAMMemory:pbyInternalRAM=%x,dwInternalRAMSize=%x\r\n",
			   	               pbyInternalRAM, dwInternalRAMSize);
		    return (FAIL);
	}

	
	/* 	check if we can write to the entire image	*/
	RB_Trace(NULL, "!!!!!!!!!Check 4 yx\r\n");
//	RB_memset(pRAM+1024*1024, 98, 1024*1024); 
	if (RB_CheckWriteImage(pRAM, start, start+1024*1024))
	{
	      RB_Trace(NULL,"error RB_CheckWriteImage:pRAM=%x,start=%x,end=%x\r\n",
		  	    pRAM, start, end);
		return (FAIL);
	}
	
/* 	check if we can read image	*/
	RB_Trace(NULL, "!!!!!!!!!Check 5 yx\r\n");
	if (RB_CheckReadImageCases(pRAM, &start))
	{
	      RB_Trace(NULL,"error RB_CheckWriteImage:pRAM=%x,start(%x)=%x\r\n",
		  	pRAM,&start,start);
		return (FAIL);
	}
	
	/*	check if we can write to back up block	*/
	RB_Trace(NULL, "!!!!!!!!!Check 6 yx\r\n");
	for (i=0; i<num_of_buff; i++)
		if (RB_CheckWriteBackupBlock(pRAM, buff_start[i]))
		{
		     RB_Trace(NULL,"error RB_CheckWriteBackupBlock:pRAM=%x,buff_start[%d]=%x\r\n",
			 	pRAM, i, buff_start[i]);
			return (FAIL);
		}

	/* 	check that the address of the backup buffers is not part of image	*/
	RB_Trace(NULL, "!!!!!!!!!Check 7 yx\r\n");
	if (RB_CheckIfBckInImage(start, end, buff_start, num_of_buff))
	{
	    RB_Trace(NULL,"error RB_CheckIfBckInImage:start=%x, end=%x, buff_start=%x, num_of_buff=%x\r\n",
			start, end, buff_start, num_of_buff);
		return (FAIL);
	}
	
	return (SUCCESSFUL);
}

/****************************************************************************/
/* NAME:		int RB_CheckRamMemory(Byte* pRAM, Dword dwRAMSize) 	*/
/*					*/
/* DESCRIPTION:	Checks that writing to RAM works	*/
/*					*/
/* INPUTS:				*/
/*	pRAM - pointer to RAM area 		*/
/*	dwRAMSize - size of RAM to check 	*/
/*					*/
/* OUTPUTS:				*/
/*	SUCCESSFUL or FAIL			*/
/*					*/
/* NOTES:				*/
/*					*/
/****************************************************************************/

int RB_CheckRamMemory(Byte* pRAM, Dword dwRAMSize)
{
		 
	Dword* current_mem = (Dword*)pRAM;
	Dword* mem_end = (Dword*)(pRAM + dwRAMSize);
	
	         RB_Trace(NULL, " RB_CheckRamMemory\r\n");

	/*	check input parameters	*/
	if (!pRAM || !dwRAMSize)
		return (FAIL);
		
    RB_Trace(NULL, " RB_CheckRamMemory step 2\r\n");
    current_mem += 2*1024*1024;
    
	for (; current_mem < mem_end; ++current_mem)
	{
			// 	write to RAM 0x1234 and immediately read it	
			*current_mem = RANDOM_VALUE;
			if (*current_mem != RANDOM_VALUE)
			{
			RB_Trace(NULL, "Cannot write to RAM at address 0x%08x, value is 0x%08x, should be 0x%08x\r\n",
					current_mem, *current_mem, RANDOM_VALUE);
					return (FAIL);
			}
			current_mem+=20*1024;
	}
	
	RB_Trace(NULL, " RB_CheckRamMemory Leave\r\n");
	return (SUCCESSFUL);
}

/****************************************************************************/
/* NAME:	int RB_CheckInternalRAMMemory(	*/
/*			Byte* pbyInternalRAM, Dword dwInternalRAMSize	*/
/*					*/
/* DESCRIPTION:	Checks that writing to internal RAM works	*/
/*					*/
/* INPUTS:				*/
/*	pRAM - pointer to internal RAM area	*/
/*	dwRAMSize - size of internal RAM to check	*/
/*					*/
/* OUTPUTS:				*/
/*	SUCCESSFUL or FAIL			*/
/*					*/
/* NOTES:				*/
/*					*/
/****************************************************************************/

int RB_CheckInternalRAMMemory(Byte* pbyInternalRAM, Dword dwInternalRAMSize)
{

	Dword* current_mem = (Dword*) pbyInternalRAM;
	Dword* mem_end = (Dword*) (pbyInternalRAM + dwInternalRAMSize);
         RB_Trace(NULL, "RB_CheckInternalRAMMemory\r\n");
	/*	check input parameters	*/
if (!dwInternalRAMSize || !pbyInternalRAM)
		return SUCCESSFUL;

		for (; current_mem < mem_end; ++current_mem)
	{
			*current_mem = RANDOM_VALUE;
			if (*current_mem != RANDOM_VALUE)
		{
			RB_Trace(NULL, "Cannot write to Internal RAM at address 0x%08x, value is 0x%08x should be 0x%08x\r\n",
				 current_mem,*current_mem, RANDOM_VALUE);
					return (FAIL);
		}
	}
	return (SUCCESSFUL);
}

/****************************************************************************/
/* NAME:	int RB_CheckReadImage(Byte* pRam, Dword flash, int readSize)	*/
/*					*/
/*					*/
/* DESCRIPTION:	Checks that reading the image works   	  */
/*					*/
/* INPUTS:				*/
/*	pRAM - pointer to RAM area		*/
/*	flash - address in flash to read from	*/
/*	readSize - size in bytes to read    	*/
/*					*/
/* OUTPUTS:			  	*/
/*	SUCCESSFUL or FAIL			*/
/*					*/
/* NOTES:  				*/
/*					*/
/****************************************************************************/

int RB_CheckReadImage(Byte* pRam, Dword flash, int readSize)
{
	Byte *pTempRam1, *pTempRam2;
	int i;
    int diff = 50*1024;
         RB_Trace(NULL, "Enter RB_CheckReadImage\r\n");
	pTempRam1 = pRam+1024*1024;
	pTempRam2 = pRam + 4*1024*1024; 
	int memValue = 80;
	
	/*	 check that size to read is not bigger than max sector size	 */
	if (RB_GetBlockSize() < readSize)
	{
		RB_Trace(NULL, "Given block size 0x%08x is bigger than sector max size 0x%08x", readSize, RB_GetBlockSize());
		return (FAIL);
	}
    
    RB_Trace(NULL, "diff = %08x, pTempRam1 = %08x, readSize= %08x\r\n", diff, pTempRam1, readSize);
	/*from (pTempRam1) till (pTempRam1 + sectorSize + "diff"*2) is memValue	*/
	RB_memset(pTempRam1, memValue, readSize + diff*2);
	RB_Trace(NULL, "pass 1\r\n");	
	/* pTempRam2 looks like this	*/
	/* first "diff" chars memValue	*/
	/* then sectorSize char imageValue	*/
	/* then last "diff" chars memValue	*/
	RB_Trace(NULL, "pass 2\r\n");
	RB_memset(pTempRam2, 98, 1024*1024);
	RB_Trace(NULL, "pass 5\r\n");	
	RB_Trace(NULL, " RB_CheckReadImage after memset\r\n");
 
	/* read from image into pTempRam1	*/
	/* RB_ReadImage is one of the API functions which should be implemented	*/
	//if (RB_ReadImage(NULL, pTempRam1 + 5*FLASH_BLOCK_SIZE, flash, readSize) != 0)
	if (RB_ReadImage(NULL, pTempRam1, flash, readSize) != 0)
	{
		RB_Trace(NULL, "Error read image from address 0x%08x reading size is 0x%08x\r\n", 
			               flash, readSize);
		return (FAIL);
	}
	
	else if (i = RB_memcmp(pTempRam1 + diff, pTempRam2 + diff, readSize)) 
	{
		RB_Trace(NULL, "Error in read image - dirty number byte is 0x%08x,start read address is 0x%08x\r\n",
			                 i, flash);
		return (FAIL);
	}
	else if (i = RB_memcmp(pTempRam1, pTempRam2, diff)) 
	{
		RB_Trace(NULL, "Error in read image - dirty number byte is 0x%08x, start read address is 0x%08xr\\n", 
			         i, flash);
		return (FAIL);
	}
	else if (i = RB_memcmp(pTempRam1 + readSize, pTempRam2 + readSize, diff)) 
	{
		RB_Trace(NULL, "Error in read image - dirty number byte is 0x%08x, start read address is 0x%08x\r\n",
			      i, flash);
		return (FAIL);
	}
	
	return (SUCCESSFUL);
}

/****************************************************************************/
/* NAME:	int RB_CheckReadBackupBlock(Byte *pRam, Dword flash,	*/
/*				Dword readSize)	*/
/*	 				*/
/* DESCRIPTION:	Checks that reading from back up block works	*/
/*					*/
/* INPUTS:				*/
/*	pRAM - pointer to RAM area		*/
/*	flash - address in flash to read from 	*/
/*	readSize - size in bytes to read	*/
/*					*/
/* OUTPUTS:				*/
/*	SUCCESSFUL or FAIL			*/
/*					*/
/* NOTES:				*/
/*					*/
/****************************************************************************/

int RB_CheckReadBackupBlock(Byte *pRam, Dword flash, Dword readSize)
{

	Byte *pTempRam1, *pTempRam2;
	int diff = 1024*1024;
	
    RB_Trace(NULL, " RB_CheckReadBackupBlock\r\n");
	pTempRam1 = pRam;
	pTempRam2 = pRam + readSize + diff * 2;
	
	/* 	from (pTempRam1-diff) till (pTempRam1 + sectorSize) is memValue	*/
	RB_memset(pTempRam1, imageValue, readSize + diff*2);
	/* 	pTempRam2 looks like this: 	*/
	/* 	first "diff" chars memValue	*/
	/* 	then sectorSize char imageValue	*/
	/* 	then last "diff" chars memValue	*/
	RB_memset(pTempRam2, imageValue, diff);
	RB_memset(pTempRam2 + diff, memValue, readSize);
	RB_memset(pTempRam2 + readSize + diff, imageValue, diff);
	
	/* 	read from image into pTempRam1	*/
	/* 	RB_ReadBackupBlock is one of the API functions which should	*/
	/* 	be implemented		*/
	if (RB_ReadBackupBlock(NULL, pTempRam1 + diff, flash, readSize) != 0)
	{
		RB_Trace(NULL, "Error read image from address 0x%08x sectorsize 0x%08x\r\n", 
			flash, readSize);
		return (FAIL);
	}

	
 
	else if (RB_memcmp(pTempRam1 + diff, pTempRam2 + diff, readSize))
	{
		RB_Trace(NULL, "Error comparing pTempRam1 + diff 0x08x and pTempRam2 + diff 0x08x\r\n",
			pTempRam1 + diff, pTempRam2 + diff);
		return (FAIL);
	}
	else if (RB_memcmp(pTempRam1, pTempRam2, diff))
	{
		RB_Trace(NULL, "Error comparing pTempRam1 0x08x and pTempRam2 0x08x\r\n",
			pTempRam1, pTempRam2);
		return (FAIL);
	}
	else if (RB_memcmp(pTempRam1 + readSize, pTempRam2 + readSize, diff))
	{
		RB_Trace(NULL, "Error comparing pTempRam1 + readSize 0x08x and pTempRam2 + readSize 0x08x\r\n",
			pTempRam1 + readSize, pTempRam2 + readSize);
		return (FAIL);
	}
	return (SUCCESSFUL);
}

/****************************************************************************/
/* NAME:		int RB_CheckReadImageCases(Byte *pRam, Dword flash) 	*/
/*					*/
/* DESCRIPTION: 	Checks that reading from image works	*/
/*	 				*/
/* INPUTS:				*/
/*	pRAM - pointer to RAM area		*/
/*	flash - address in flash to read from	*/
/*					*/
/* OUTPUTS: 				*/
/*	SUCCESSFUL or FAIL			*/
/*					*/
/* NOTES:				*/
/*					*/
/****************************************************************************/

int RB_CheckReadImageCases(Byte* pRam, Dword* pFlash)
{

	Dword flash;
	int evenSize = 10240;
	int oddSize = 9987;
         RB_Trace(NULL, " RB_CheckReadImageCases\r\n");
	flash = *pFlash;
	Byte* p = pRam + 1* FLASH_BLOCK_SIZE;
	
	/* 	check for a whole buffer	*/
	if (RB_CheckReadImage(p, flash, RB_GetBlockSize()))
		return (FAIL);
	RB_Trace(NULL, " RB_CheckReadImage whole block pass\r\n");	

	/* 	check for an even part of buffer's size	*/
	/* 	sector size even size; 	*/
	if (RB_CheckReadImage(p, flash, evenSize))
	 return (FAIL);
	RB_Trace(NULL, " RB_CheckReadImage evenSize pass1\r\n");
	/* 	check for an odd part of buffer's size	*/
	/* 	sector size odd size; 	*/
	if (RB_CheckReadImage(p+FLASH_BLOCK_SIZE, flash, oddSize))
	 return (FAIL);
    RB_Trace(NULL, " RB_CheckReadImage oddSize pass1\r\n");
	/* 	check for sector size evenSize; 	*/
	flash ++;
	if (RB_CheckReadImage(p, flash, evenSize))
	 return (FAIL);
    RB_Trace(NULL, " RB_CheckReadImage evenSize pass2\r\n");

	/* 	sector size oddSize;	*/
	if (RB_CheckReadImage(p+FLASH_PAGE_SIZE, flash, oddSize))
	 return (FAIL);
    RB_Trace(NULL, " RB_CheckReadImage evenSize pass2\r\n"); 
	 
	return (SUCCESSFUL);
}

/****************************************************************************/
/* NAME:		int RB_CheckWriteSector(Byte *pRam, Dword flash)	*/
/*					*/
/* DESCRIPTION:	Checks that writing sector works 	*/
/*					*/
/* INPUTS:				*/
/*	pRAM - pointer to RAM area		*/
/*	flash - address in flash to read from	*/
/*					*/
/* OUTPUTS:				*/
/*	SUCCESSFUL or FAIL			*/
/* NOTES:				*/
/****************************************************************************/

int RB_CheckWriteSector(Byte* pRam, Dword flash)
{
    char imageValue = 68;
    char memValue = 68;
	int sectorSize = RB_GetBlockSize();
	Byte *pTempRam1 = pRam+3*1024*1024, *pTempRam2 = pRam + 4*1024*1024;
	
         RB_Trace(NULL, " RB_CheckWriteSector\r\n");
	/* 	don't write to excluded address	*/
	if (RB_CheckIfAddressNotExcluded(flash, sectorSize))
		 return (SUCCESSFUL);
    RB_Trace(NULL,"check success.\r\n");		 
	RB_memset(pTempRam1, imageValue, sectorSize);
	RB_memset(pTempRam2, imageValue, sectorSize);
	/* 	RB_WriteBlock is one of the API functions which should be implemented	*/
	if (RB_WriteBlock(NULL, flash, pTempRam1))
	{
		RB_Trace(NULL, "cannot write to sector 0x%08x\r\n", flash);
			return (FAIL);
	}
	RB_Trace(NULL, "Write block %08x success.\r\n");
	RB_memset(pTempRam1, memValue, sectorSize);
	
	/* 	RB_ReadImage is one of the API functions which should be implemented	*/
	if (RB_ReadImage(NULL, pTempRam1, flash, sectorSize))
	{
		RB_Trace(NULL, "cannot read from address 0x%08x size 08%08x into address 0x%08x\r\n", flash, sectorSize, pTempRam1);
			return (FAIL);
	}
	*(pTempRam1+128) = 0;
	*(pTempRam2+128) = 0;
	RB_Trace(NULL, "pTempRam1 = %c\r\n", pTempRam1+100);
	RB_Trace(NULL, "pTempRam2 = %c\r\n", pTempRam2+100);
//	RB_Trace(NULL, "pTempRam1+1024 = %c, pTempRam2+1024=%c\r\n", pTempRam1[1024], pTempRam2[1024]);
	
	if (RB_memcmp(pTempRam1, pTempRam2, sectorSize))
	{
		RB_Trace(NULL, "compare error after write & read. pTempRam1 = 08%08x pTempRam2 = 0x%08x\r\n", pTempRam1, pTempRam2);
			return (FAIL);
	}
	return (SUCCESSFUL);
}

/****************************************************************************/
/* NAME:	int RB_CheckWriteImage(Byte *pRam, Dword flashStart,	*/
/*		Dword flashEnd)		*/
/*					*/
/*					*/
/* DESCRIPTION:	Checks that writing to image works 	*/
/*				     	*/
/* INPUTS:				*/
/*	pRAM - pointer to RAM area		*/
/*	flashstart - address in flash to read from	*/
/*	flashEnd - end address in flash		*/
/*					*/
/* OUTPUTS:				*/
/*	SUCCESSFUL or FAIL			*/
/*					*/
/* NOTES:				*/
/****************************************************************************/

int RB_CheckWriteImage(Byte* pRam, Dword flashStart, Dword flashEnd)
{

	Dword currentFlash;
	int sectorSize = RB_GetBlockSize();
	
         RB_Trace(NULL, " RB_CheckWriteImage\r\n");
	/*	write to all sectors in image	*/
	for (currentFlash=flashStart; currentFlash+sectorSize<flashEnd && currentFlash<flashEnd; currentFlash+=sectorSize)
	{
		if (RB_CheckWriteSector(pRam, currentFlash))
			 return (FAIL);
	}
	
	return (SUCCESSFUL);
}

/****************************************************************************/
/* NAME:	int RB_CheckWriteBackupBlock(Byte *pRam, Dword flash	*/
/*					*/
/* DESCRIPTION:	Checks that writing to backup block works	*/
/*					*/
/* INPUTS:				*/
/*	pRAM - pointer to RAM area		*/
/*	flashstart - address in flash to read from	*/
/*	flashEnd - end address in flash    	*/
/*					*/
/* OUTPUTS:				*/
/*	SUCCESSFUL or FAIL			*/
/*					*/
/* NOTES:				*/
/*					*/
/****************************************************************************/

int RB_CheckWriteBackupBlock(Byte* pRam, Dword flash)
{

	int sectorSize = RB_GetBlockSize();
	Byte *pTempRam1 = pRam, *pTempRam2 = pRam + sectorSize;

	RB_Trace(NULL, " RB_CheckWriteBackupBlock\r\n");
	RB_memset(pTempRam1, backupImgValue, sectorSize);
	RB_memset(pTempRam2, backupImgValue, sectorSize);
	
	/*	RB_WriteBackupBlock is one of the API functions which should 	*/
	/*	be implemented		*/
	if (RB_WriteBackupBlock(NULL, flash, pTempRam1))
		return (FAIL);
	
	RB_memset(pTempRam1, memValue, sectorSize);
	
	/*	RB_ReadBackupBlock is one of the API functions which should	*/
	/*	be implemented		*/
	if (RB_ReadBackupBlock(NULL, pTempRam1, flash, sectorSize))
		return (FAIL);
	if (RB_memcmp(pTempRam1, pTempRam2, sectorSize))
		return (FAIL);
		
	return (SUCCESSFUL);
}

/****************************************************************************/
/* NAME:		int RB_memcmp(Byte *from,  Byte *to, int length	*/
/*					*/
/* DESCRIPTION:	Red Bend simple implementation for memcmp	*/
/*					*/
/* INPUTS:				*/
/*	from - pointer to first memory area to compare    	*/
/*	to - pointer to second memory area to compare	*/
/*	length - length of memory to compare    	*/
/*					*/
/* OUTPUTS:				*/
/*	0 or 1		 		*/
/*				    	*/
/* NOTES:				*/
/*					*/
/****************************************************************************/

int RB_memcmp(Byte* from, Byte *to, int length)
{
	int i;
	Byte a, b;
	RB_Trace(NULL,"Enter memcmp from=0x%08x, to=0x%08x length = %d\r\n", from, to, length);
	for (i=0; i<length; i++)
	{
	    a = *(from+i);
	    b = *(to+i);
		if ((int)a != (int)b)
		{
		    RB_Trace(NULL, "Not equal found. a=%d b=%d i=%d\r\n", a, b, i);
			return (FAIL);
		}	
	}
	RB_Trace(NULL, "Leave memcmp\r\n");
	return (SUCCESSFUL);
}

/****************************************************************************/
/* NAME:		int RB_memset(Byte* to, Byte value, int size 	*/
/*					*/
/* DESCRIPTION:	Red Bend simple implementation for memset	*/
/*					*/
/* INPUTS:				*/
/*	to - pointer to memory area to set	*/
/*	value- value to set		*/
/*	size - length of memory to set		*/
/*					*/
/* OUTPUTS:		None		*/
/*					*/
/* NOTES:				*/
/****************************************************************************/
void RB_memset(Byte* to, Byte value, int size)
{
    int i;
    RB_Trace(NULL, "Enter memset. to=%08x value=%d size=%d\r\n", to, value, size);
	for (i=0; i<size; i++)
	{
	 if( (i - 1000*(i/1000)) ==0 )  
	      to[i] = value;
	}	
	RB_Trace(NULL, "Leave memset.\r\n");	
}

/****************************************************************************/
/* NAME:	void* RB_memcpy (void *dest, const void *src, Dword count) 	*/
/*					*/
/* DESCRIPTION:	Red Bend simple implementation for memcpy	*/
/*					*/
/* INPUTS:				*/
/*	dest - pointer to memory area to set	*/
/*	src - pointer to memory area to be copied	*/
/*	size - length of memory to set		*/
/*					*/
/* OUTPUTS:				*/
/*	copied buffer	 		*/
/*					*/
/* NOTES:				*/
/****************************************************************************/

void* RB_memcpy(void *dest, const void *src, Dword count)
{
	char *dst8 = (char *)dest;
	char *src8 = (char *)src;

    RB_Trace(NULL, "Enter memcpy, dest=%08x src=%08x count=%d\r\n", dest, src, count);
	while (count--)
	{
		*dst8++ = *src8++;
	}
	RB_Trace(NULL, "Leave memcpy\r\n");
	return dest;
}

/****************************************************************************/
/* NAME:	int RB_CheckIfAddressNotExcluded(Dword flashAddress,	*/
/*			int sectorSize)	*/
/*					*/
/* DESCRIPTION:	Checks if the given address are not excluded	*/
/*					*/
/* INPUTS:				*/
/*	flashAddress C start address to exclude	*/
/*	sectorSize C size of area to exclude	*/
/*					*/
/* OUTPUTS:				*/
/*	0 or 1				*/
/*					*/
/* NOTES:				*/
/*					*/
/****************************************************************************/

int RB_CheckIfAddressNotExcluded(Dword flashAddress, int sectorSize)
{

	int i=0;
         RB_Trace(NULL, " RB_CheckIfAddressNotExcluded\r\n");
	if (!num_of_exclude_regions)
		return 0;
	for (; i<num_of_exclude_regions; i++)
		if ((flashAddress >= exclude_regions_array_start[i] &&
			flashAddress < exclude_regions_array_start[i] + exclude_regions_array_length[i])
			|| (flashAddress+sectorSize >= exclude_regions_array_start[i] &&
			flashAddress+sectorSize <= exclude_regions_array_start[i] + exclude_regions_array_length[i]))
			  return (FAIL);

    RB_Trace(NULL, "Leave CheckIfAddressNotExclude\r\n");
	 return (SUCCESSFUL);
}

/****************************************************************************/
/* NAME:	int RB_CheckIfBckInImage(Dword start, Dword end,	*/
/*			Dword* buff_start, Dword num_of_buff)	*/
/*					*/
/*					*/
/* DESCRIPTION:	Checking if backup block is part of the image	*/
/*					*/
/* INPUTS:				*/
/*	start C start address of backup		*/
/*	end C end of backup buffer		*/
/*	num_of_buff C number of backup buffers	*/
/*					*/
/* OUTPUTS:				*/
/*	0 or 1				*/
/*					*/
/* NOTES:				*/
/*					*/
/****************************************************************************/

int RB_CheckIfBckInImage(Dword start, Dword end, Dword* buff_start, Dword num_of_buff)
{

//	Dword current;
	Dword i;
	int sectorSize = RB_GetBlockSize();
	
         RB_Trace(NULL, " RB_CheckIfBckInImage\r\n");
	for (i=0; i<num_of_buff; i++)
	{
		if (buff_start[i] >= start && buff_start[i] < end)
			if (!RB_CheckIfAddressNotExcluded(buff_start[i], sectorSize))
		{
				RB_Trace(NULL, "Probably backup block %d are part of image\r\n", i);
				  return (FAIL);
		}
	}
	return (SUCCESSFUL);
}


/****************************************************************************/
/* NAME:		void RB_DefineWritingExcludeRegions	*/
/*					*/
/*					*/
/* DESCRIPTION:	define exclude regions	*/
/*					*/
/* INPUTS:		None		*/
/*					*/
/* OUTPUTS:		None		*/
/*					*/
/* NOTES:				*/
/*					*/
/****************************************************************************/

void RB_DefineWritingExcludeRegions(void)
{
         RB_Trace(NULL, " RB_DefineWritingExcludeRegions\r\n");

	//if you have exclude regions set their start address and length here
	num_of_exclude_regions = 1;
	exclude_regions_array_start[0] = 0x00AC0000;
	exclude_regions_array_length[0] = 0x00540000;
}




