/******************************************************************************
**  (C) Copyright 2007 Marvell International Ltd. 
**  All Rights Reserved
******************************************************************************/

/******************************************************************************
**
**  COPYRIGHT (C) 2006 Intel Corporation.
**
**  This software as well as the software described in it is furnished under
**  license and may only be used or copied in accordance with the terms of the
**  license. The information in this file is furnished for informational use
**  only, is subject to change without notice, and should not be construed as
**  a commitment by Intel Corporation. Intel Corporation assumes no
**  responsibility or liability for any errors or inaccuracies that may appear
**  in this document or any software that may be provided in association with
**  this document.
**  Except as permitted by such license, no part of this document may be
**  reproduced, stored in a retrieval system, or transmitted in any form or by
**  any means without the express written consent of Intel Corporation.
**
**  FILENAME:
**
**  PURPOSE:
**              USB 2.0 Client Device Controller DMA transfer routines
**
**
**  LAST MODIFIED:
**
******************************************************************************/

#include "xllp_defs.h"
#include "misc.h"
#include "xllp_u2d.h"

//--------------------------------------
// Global Memory pointers and Structures
//--------------------------------------


#define U2D_ERROR_LIST_SIZE 128
XLLP_UINT8_T u2DErrList[U2D_ERROR_LIST_SIZE];
XLLP_UINT32_T u2DErrListLocation = 0;

static P_XLLP_U2D_REGISTERS_T      pRegs = (P_XLLP_U2D_REGISTERS_T)XLLP_MONAHANS_U2D_REG_BASE;

//------------------------------------------------
// In-memory circular error log for U2D
//-------------------------------------------------
XLLP_UINT8_T U2DLogError( XLLP_UINT8_T Error )
{
    if( u2DErrListLocation >= U2D_ERROR_LIST_SIZE )
      u2DErrListLocation = 0;

    u2DErrList[u2DErrListLocation++] = Error;

    return Error;
}


//----------------------------------------------------------
// XllpU2DEndPointEnabled()
//
//----------------------------------------------------------
XLLP_UINT8_T XllpU2DEndPointEnabled( XLLP_U2D_EP_T EndPtId )
{
  // Endpoint 0: because of the U2D register definitions,
  // and USB defines that EP 0 is always enabled, the
  // corresponding U2DCR_X register for EP0 does not exist, thus
  // it also lacks an EE bit ...
  //------------------------------------------------------------------
  if( EndPtId == U2D_ENDPOINT_0 )
    return 1;

  return
  (pRegs->U2DCR_X[EndPtId]) & XLLP_U2DCRX_EE;
}

//---------------------------------------------------------------------
// XllpU2DGetEndPointMaxPacketSize()
//
//---------------------------------------------------------------------
XLLP_UINT32_T XllpU2DGetEndPointMaxPacketSize( XLLP_U2D_EP_T EndPtId )
{
   return
   (pRegs->U2DNE_X[EndPtId] & XLLP_U2DNEX_MPS_MASK ) >> XLLP_U2DNEX_MPS_SHIFT;
}

//------------------------------------------------------------------------
// XllpU2DGetEndPointTransferDirection()
//
//  Check the USB endpoint direction from the
//  endpoint information register.
//  The bit set as '1' indicates a USB IN type
//  endpoint
//
//------------------------------------------------------------------------
XLLP_UINT32_T XllpU2DGetEndPointTransferDirection( XLLP_U2D_EP_T u2dEndPtId )
{
   return
   (pRegs->U2DNE_X[u2dEndPtId] & XLLP_U2DNEX_ED) >> XLLP_U2DNEX_ED_SHIFT;
}

//----------------------------------------------------------------------------
// XllpU2DGetUSBEndpointId
//
//----------------------------------------------------------------------------
XLLP_U2D_EP_T XllpU2DGetUSBEndpointId(XLLP_U2D_EP_T u2dEp )
{
  XLLP_U2D_EP_T USBid;

  if( u2dEp == 0 )
    return U2D_ENDPOINT_0;

  USBid = (pRegs->U2DNE_X[u2dEp] & XLLP_U2DNEX_EN_MASK);
  USBid >>= XLLP_U2DNEX_EN_SHIFT;

  return USBid;
}

//--------------------------------------------------
// XllpU2DGetBinding()
//       Find the mapping from the USB endpoint
//       to the U2D endpoint
//
//--------------------------------------------------
XLLP_U2D_EP_T XllpU2DGetBinding( XLLP_U2D_EP_T USBEndPtId )
{
   XLLP_U2D_EP_T i;

   for( i = U2D_ENDPOINT_A; i <= U2D_ENDPOINT_G; i++ )
   {
     if( (XllpU2DEndPointEnabled(i)) &&
         (XllpU2DGetUSBEndpointId(i)) == USBEndPtId )
     {
       return i;
     }
   }

   U2DLogError(U2D_ERR_ENDPOINT_BIND);

   return U2D_ENDPOINT_BAD;
}

//----------------------------------------------------------------------------
//  USBEndPointPacketInterruptReceived()
//
//     Determine if there is an outstanding interrupt on the USB endpoint
//
//----------------------------------------------------------------------------
XLLP_UINT8_T USBEndPointPacketInterruptReceived( XLLP_U2D_EP_T usbEp )
{
  XLLP_U2D_EP_T u2dEp;

  // get the mapping from USB to U2D
  //--------------------------------
  u2dEp = XllpU2DGetBinding( usbEp );

  if( u2dEp != U2D_ENDPOINT_BAD )
    return pRegs->U2DISR & (XLLP_U2DISR_IEx_MASK << (XLLP_U2DISR_CHAN_SHIFT * u2dEp));

  return 0;
}

//--------------------------------------------------
// XllpU2DWaitForStoppedDMAChan()
//
//   Wait for the Channel to be in the Stopped state
//   as indicated by the STOPINTR in the appropriate
//   channels control/status register
//--------------------------------------------------
void XllpU2DWaitForStoppedDMAChan( XLLP_U2D_EP_T u2dEndPtId )
{
   XLLP_UINT32_T runMask;

   // Wait for the DMA channel to be in the STOPPED state
   //----------------------------------------------------
   do
   {
       runMask = pRegs->U2DMACSR_X[u2dEndPtId];

   }while( ((runMask & XLLP_U2DMACSRX_STOPINTR) == 0) &&
           ((runMask & XLLP_U2DMACSRX_RUN     ) != 0)
         );
}

//-----------------------------------------------------------
// XllpU2DWaitForEndOfReceiveOrStopped()
//
//   Wait for the ENDPOINT FIFO to indicate that the endpoint
//   is finished receiving
//-----------------------------------------------------------
void XllpU2dWaitforEndOfReceiveOrStopped( XLLP_U2D_EP_T u2dEpId )
{
  // Wait for the Endpoint to be in the EOR or STOPPED state
  //--------------------------------------------------------
  while( (pRegs->U2DMACSR_X[u2dEpId] & (XLLP_U2DMACSRX_EORINTR | XLLP_U2DMACSRX_STOPINTR)) == 0 )
  {}

  // This code covers a silicon bug... Tibet ID 2017404
  // we must clear
  // the descriptor Start, Stop and/or EOR interrupts
  // even though we may not be enabling these interrupts...
  // write 1 to clear the interrupt status bits
  //--------------------------------------------------------
  pRegs->U2DMACSR_X[u2dEpId] &= pRegs->U2DMACSR_X[u2dEpId];
}

//-----------------------------------------------------
// U2DSetRunBit()
//
//
//
//-----------------------------------------------------
void U2DSetRunBit( XLLP_U2D_EP_T EndPtId )
{
   XLLP_VUINT32_T temp;

   // Clear any outstanding interrupt bits first
   // set-up a proper mask to prevent writing 1's
   // back to certain bits in the register
   //--------------------------------------------
   temp = pRegs->U2DMACSR_X[EndPtId];
   temp &= ( XLLP_U2DMACSRX_EORINTR    |
             XLLP_U2DMACSRX_RASINTR    |
             XLLP_U2DMACSRX_ENDINTR    |
             XLLP_U2DMACSRX_STARTINTR  |
             XLLP_U2DMACSRX_BUSERRINTR   );

   // Now set the Run bit
   //--------------------
   temp |= XLLP_U2DMACSRX_RUN;
   pRegs->U2DMACSR_X[EndPtId] = temp;
   temp = pRegs->U2DMACSR_X[EndPtId];
}

//--------------------------------------------------
// XllpU2DStartEndpointTX()
//
//
//--------------------------------------------------
void XllpU2DStartEndpointTx( XLLP_U2D_EP_T              EndPtId,
                             P_XLLP_U2D_DMA_RAM_DESCR_T DescrPtr )
{
   XLLP_VUINT32_T temp;

   // Ensure the U2D has locked in the descriptor address
   //----------------------------------------------------
   pRegs->U2DMACHAN_X[EndPtId].descriptorAddr = DescrPtr->descriptorAddr;
   temp = pRegs->U2DMACHAN_X[EndPtId].descriptorAddr;

   // Ok, once we set DME for a channel, never touch the
   // U2DCSR_X register again, unless we reconfigure the USB
   // controller.  The DMA and U2D act in concert thru
   // internal state machines, so meddliing with bits is
   // not advised ...
   //------------------------------------------------------
   if( !(pRegs->U2DCSR_X[EndPtId] & XLLP_U2DCSRX_DME) )
   {
      pRegs->U2DCSR_X[EndPtId] = (XLLP_U2DCSRX_DME) ;
      temp = pRegs->U2DCSR_X[EndPtId];
   }

   U2DSetRunBit( EndPtId );
}

//----------------------------------------------------------------
// XllpU2DStartEndpointRX()
//
//
//----------------------------------------------------------------
void XllpU2DStartEndpointRx( XLLP_U2D_EP_T              EndPtId,
                             P_XLLP_U2D_DMA_RAM_DESCR_T DescrPtr )
{
   XLLP_VUINT32_T temp;

   // Ensure the U2D has locked in the descriptor address
   // before starting the DMA transfer
   //----------------------------------------------------
   pRegs->U2DMACHAN_X[EndPtId].descriptorAddr = DescrPtr->descriptorAddr;
   temp = pRegs->U2DMACHAN_X[EndPtId].descriptorAddr;

   // Ok, once we set DME for a channel, never touch the
   // U2DCSR_X register again, unless we recongiure the USB
   // controller.  The DMA and U2D act in concert thru
   // internal state machines, so meddliing with bits is
   // not advised ...
   //------------------------------------------------------
   if( !(pRegs->U2DCSR_X[EndPtId] & XLLP_U2DCSRX_DME) )
   {
      pRegs->U2DCSR_X[EndPtId] = (XLLP_U2DCSRX_DME);
      temp = pRegs->U2DCSR_X[EndPtId];
   }

   // Now start the U2 DMA transfer
   //------------------------------
   U2DSetRunBit(EndPtId);
}

//--------------------------------------------------
// XllpU2DTearDownFinishedTransfer()
//
//
//--------------------------------------------------
void XllpU2DTearDownFinishedTransfer( XLLP_U2D_DMA_CHANNEL_ID EndPtId )
{
  XLLP_UINT32_T temp;

  // DMA RUN MUST MUST MUST BE OFF !!!
  // Before loading new descriptor...
  // This also clears interrupt status bits such as:
  // EORINTR... VERY important
  //----------------------------------------------------------------------------
  pRegs->U2DMACSR_X[EndPtId] &= ~( XLLP_U2DMACSRX_RUN | XLLP_U2DMACSRX_MASKRUN);
  temp = pRegs->U2DMACSR_X[EndPtId];

#if 0
  // ATTEMPT TO TURN OFF DME BIT & SET FTF
  //------------------------------------------
  {
     temp = pRegs->U2DCSR_X[EndPtId];
     temp &= ~XLLP_U2DCSRX_DME;
     pRegs->U2DCSR_X[EndPtId] = temp;
     temp = pRegs->U2DCSR_X[EndPtId];
     temp |= XLLP_U2DCSRX_FEF;
     pRegs->U2DCSR_X[EndPtId] = temp;
     temp = pRegs->U2DCSR_X[EndPtId];
  }
#endif

}

//--------------------------------------------------------------------------------
// DMA Tx/Rx Routine for EndPoints A->G
//
//      Transfer the data, alleviating the caller from USB knowledge/complexities.
//      The transfer is broken up into the required number of USB packets here.
//
//      The max transfer size is limited by the number of descriptors available;
//      if needed, increase the number of descriptors if the transfer cannot be
//      accomodated. (SYS_U2D_NUM_DMA_DESCRIPTORS)
//
//      This service does not allow multiple transfers to be queued on the same
//      DMA channel. So transfers are serialized here by waiting for any preceding
//      transfer on the channel to complete before starting the next transfer on
//      the USB.
//
//   Assumptions:
//    The RAM buffer must be on an 8 byte alignment
//    We don't support using multiple descriptors for a single packet.
//    We only accomodate USB BULK and INTERRUPT packets.
//
//--------------------------------------------------------------------------------
XLLP_UINT8_T XllpU2DMAxfer( XLLP_U2D_EP_T  USBEndPtId, // USB Endpoint
                            P_XLLP_UINT8_T     BufPtr,     // Memory buffer
                            XLLP_UINT32_T      BufSize,    // Buffer Size
                            XLLP_UINT8_T       SynchronusTransfer, // return after transmit complete?
                            XLLP_INT32_T       *RxByteCount        // actual # bytes received
                          )
{
   XLLP_UINT32_T              cmdRegMask, maxPacketSize, i, numDescriptors, packetLen, temp;
   XLLP_UINT32_T              BufPtrStart, BufSizeStart, firstPacketSize = 0;
   XLLP_U2D_EP_T              u2dEndPtId;      // U2D End point
   P_XLLP_U2D_DMA_RAM_DESCR_T descrPtr, chainPtr;
   XLLP_UINT8_T               retVal = U2D_ERR_NO_ERROR;
   static __attribute__((aligned(16))) XLLP_U2D_DMA_RAM_DESCR_T descrBuf[30];

   // Get an available DMA descriptor chain
   //--------------------------------------
   chainPtr = descrPtr = descrBuf;

   // Check input args: RAM Buf must be 8 byte aligned
   //-------------------------------------------------
   if( ((XLLP_UINT32_T)BufPtr & 0x07) != 0 )
     return U2DLogError(U2D_ERR_BAD_BUFFER_ALIGN);

   // Figure out which U2D endpoint the USB endpoint is bound to
   // also checks that the endpoint is enabled and configured
   //-----------------------------------------------------------
   u2dEndPtId = XllpU2DGetBinding( USBEndPtId );
   if( u2dEndPtId == U2D_ENDPOINT_BAD )
     return U2DLogError(U2D_ERR_ENDPOINT_FAIL);

   // Get max packet size
   //-------------------------------------
   maxPacketSize = XllpU2DGetEndPointMaxPacketSize( u2dEndPtId );
   if( maxPacketSize == 0)
     return U2DLogError(U2D_ERR_ENDPOINT_NO_FIFO);

   // Ensure any prior operation on channel is complete
   //--------------------------------------------------
   XllpU2DWaitForStoppedDMAChan( u2dEndPtId );
   XllpU2DTearDownFinishedTransfer( u2dEndPtId );

   // If it's a USB IN endpoint, set XFIRDIR and PACKCOMP bits in descriptor
   // Otherwise we're receiving, see what the Byte count register reads...
   //----------------------------------------------------------------------
   cmdRegMask = 0;   // side effect, if == 0 -> implies OUT EndPoint!
   if( XllpU2DGetEndPointTransferDirection(u2dEndPtId) )
     cmdRegMask = XLLP_U2DMACMDX_XFRDIR_TX | XLLP_U2DMACMDX_PACKCOMP;
   else
     firstPacketSize = pRegs->U2DBCR_X[u2dEndPtId] & XLLP_U2DBCRX_BC_MASK;

   // Each Descriptor can only describe a transfer for
   // the max packet size for the endpoint, but the Transfer
   // buffer could be much larger... so figure out how
   // many descriptors we need, and hopefully we can
   // get them allocated...
   //-------------------------------------------------------
   numDescriptors  =  BufSize/maxPacketSize;
   numDescriptors += (BufSize%maxPacketSize) ? 1 : 0;

   // If pulling out the data from a USB packet in multiple steps...
   // Compensate in the # of descriptors
   //---------------------------------------------------------------
   if( (cmdRegMask == 0)                  &&     // if Rx operation ...
       (numDescriptors > 1)               &&
       (firstPacketSize > 0)              &&
       (firstPacketSize != maxPacketSize)    )
   {
     numDescriptors += 1;
   }

   BufPtrStart = (XLLP_UINT32_T)BufPtr;
   BufSizeStart = BufSize;

   // Now set-up the descriptor stuctures for the DMA transfer
   //---------------------------------------------------------
   for( i=0; i<numDescriptors; i++ )
   {
     packetLen = (BufSize <= maxPacketSize) ? BufSize : maxPacketSize;

     // Adjust the first descriptor to match the remaining
     // partial USB packet, if that's the case...
     //------------------------------------------------------------
     if( (cmdRegMask == 0)    &&
         (i == 0)             &&
         (numDescriptors > 1) &&
         (firstPacketSize > 0 )  )
     {
        packetLen = (firstPacketSize < packetLen) ? firstPacketSize : packetLen;
     }

     descrPtr->descriptorAddr = (XLLP_VUINT32_T)descrPtr;
     descrPtr->commandReg     = packetLen | cmdRegMask;
     descrPtr->sourceAddr     = (XLLP_VUINT32_T)BufPtr;
     descrPtr->targetAddr     = (XLLP_VUINT32_T)BufPtr;

     descrPtr++;
     BufPtr   += packetLen;
    BufSize  -= packetLen;
   }

   // point to last descriptor again
   //-------------------------------
   descrPtr -= 1;

   // Set the stop bit in the end of the descriptor chain
   // Enable the IRQ so the descriptors are recovered
   // when the transfer completes
   //----------------------------------------------------
   descrPtr->descriptorAddr |= XLLP_U2DMADADRX_STOP;
   temp = descrPtr->descriptorAddr;

   // DCB: for now, we only support synchronous transfers...
   //      life is tough enough with the U2D on B0 silicon
   //--------------------------------------------------------

   // Wait here until the transfer completes if required
   //---------------------------------------------------
   //if( SynchronusTransfer )
   {
      //-----------------------------------------------------
      // Set the "STOP" bit in all descriptors in the Chain...
      // and process one at a time...
      //-----------------------------------------------------
      descrPtr = chainPtr;

      if( cmdRegMask == 0)
      {

         // This clears the packet complete or short packet complete interrupts
         // on the channel
         //---------------------------------------------------------------------
         pRegs->U2DISR = ( (XLLP_U2DISR_IEx_PCI_BIT | XLLP_U2DISR_IEx_SPI_BIT)
                                        << (XLLP_U2DISR_CHAN_SHIFT*u2dEndPtId)  );

         //----------------------------------------------
         // BEGIN CLUDGE ...  This is done this way
         // because the U2DMA does not "play nice"
         // with a descriptor chain greater than size "1"
         //----------------------------------------------
         for( i = 0; i < numDescriptors; i++, descrPtr++ )
         {
           descrPtr->descriptorAddr |= XLLP_U2DMADADRX_STOP;
           XllpU2DStartEndpointRx( u2dEndPtId, descrPtr);
           XllpU2dWaitforEndOfReceiveOrStopped( u2dEndPtId );
           XllpU2DTearDownFinishedTransfer( u2dEndPtId );
         }

         *RxByteCount = pRegs->U2DMACHAN_X[u2dEndPtId].targetAddr_ro - BufPtrStart;
      }
      else
      {
         XllpU2DStartEndpointTx( u2dEndPtId, descrPtr );

         // Wait for the transfer to complete
         //--------------------------------------
         XllpU2DWaitForStoppedDMAChan( u2dEndPtId );
         XllpU2DTearDownFinishedTransfer( u2dEndPtId );
      }
   }

   return retVal;
}



//--------------------------------------------------
// xllpU2dDmaInit
//
// U2D DMA SYSTEM INIT
//--------------------------------------------------
XLLP_UINT32_T XllpU2DDmaInit( void )
{
#if 0
    XLLP_UINT32_T size;

   size =  SYS_U2D_NUM_DMA_DESCRIPTORS * sizeof(XLLP_U2D_DMA_RAM_DESCR_T);

   // Get DMA Descriptor array allocated on 16 byte boundary
   //-------------------------------------------------------
   dmaDescrBasePtr = (P_XLLP_U2D_DMA_RAM_DESCR_T)U2DGlobalMemoryAlloc( size, 16);
   if( dmaDescrBasePtr == NULL )
     return U2DLogError( U2D_ERR_DESCRIPTOR_ALLOC_FAIL );

   memset( dmaDescrBasePtr, 0, size );

   // Allocate tracking array for descriptor allocataions
   // this array is a single byte per DMA descriptor
   //----------------------------------------------------
   dmaDescrAllocArray = U2DGlobalMemoryAlloc( SYS_U2D_NUM_DMA_DESCRIPTORS, 4);
   memset(dmaDescrAllocArray, 0, SYS_U2D_NUM_DMA_DESCRIPTORS);
#endif

   // Init the DMA H/W and clear descriptor channel regs
   //---------------------------------------------------

   // Enable bus error interrupts...
   // Primarily for debugging at the moment
   //----------------------------------------
   pRegs->U2DMACR = 0x04;

   return 0;
}

//-------------------------------------------------------------------------------
//
//-------------------------------------------------------------------------------
XLLP_UINT8_T XllpU2DInit( P_XLLP_U2D_T pHandle )
{
  // set-up the initial error/debugging log
  //----------------------------------------
  memset( u2DErrList, 0, sizeof( u2DErrList ) );
  u2DErrListLocation = 0;

  //Allocate DMA descriptor Area
  //----------------------------
  XllpU2DDmaInit();

  return 1;
}
