/*--------------------------------------------------------------------------------------------------------------------
(C) Copyright 2006, 2007 Marvell DSPC Ltd. All Rights Reserved.
-------------------------------------------------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------------------------------------------------
INTEL CONFIDENTIAL
Copyright 2006 Intel Corporation All Rights Reserved.
The source code contained or described herein and all documents related to the source code (Material) are owned
by Intel Corporation or its suppliers or licensors. Title to the Material remains with Intel Corporation or
its suppliers and licensors. The Material contains trade secrets and proprietary and confidential information of
Intel or its suppliers and licensors. The Material is protected by worldwide copyright and trade secret laws and
treaty provisions. No part of the Material may be used, copied, reproduced, modified, published, uploaded, posted,
transmitted, distributed, or disclosed in any way without Intels prior express written permission.

No license under any patent, copyright, trade secret or other intellectual property right is granted to or
conferred upon you by disclosure or delivery of the Materials, either expressly, by implication, inducement,
estoppel or otherwise. Any license under such intellectual property rights must be express and approved by
Intel in writing.
-------------------------------------------------------------------------------------------------------------------*/

/*********************************************************************
*                      M O D U L E     B O D Y                       *
**********************************************************************
* Title: I2C package                                                 *
*                                                                    *
* Filename: I2C.c                                                    *
*                                                                    *
* Target, platform: Common Platform, HW platform                     *
*                                                                    *
* Description:                                                       *
*                                                                    *
* Notes:                                                             *
**********************************************************************/


#include "typedef.h"
#include "i2c.h"
#include "PinMux_defs.h"
#include "protocolmanager.h"


/***************************   macros   *******************************************/

#define NULL        				0

#define I2C_SLAVE_WRITE(slv)        ( (slv) | 0x00000000 )			/* Master is writing to the slave */
#define I2C_SLAVE_READ(slv)         ( (slv) | 0x00000001 )      	/* Master is reading from the slave */

#define I2C_REG_WRITE(reg,wval) \
        ( (* ( (volatile UINT32*)(I2C_REGISTER_BASE_ADDRESS + (reg)) ) ) = wval);

#define I2C_REG_READ(reg,rval) \
        rval = (* ( (volatile UINT32*)(I2C_REGISTER_BASE_ADDRESS + (reg)) ) );



#define I2C_REG_BIT_WRITE(reg,btNm,sc) \
            {   UINT32 regVl; \
                I2C_REG_READ(reg,regVl); \
                sc ? ( regVl = (1 << btNm) | regVl ) : (regVl = ((regVl) & (~(1 << btNm)))); \
                I2C_REG_WRITE(reg,regVl); \
            }


#define I2C_STATUS_REG_CLEAR_BIT(bitNum)   			\
            {   UINT32 regVal; 						\
                regVal = (1 << bitNum); 			\
                I2C_REG_WRITE(I2C_ISR_REG,regVal); 	\
			}



#define I2C_REG_BIT_READ(regVl,btNm)  ( (regVl >> btNm) & 0x00000001 )

#define I2C_REG_BIT(btNm) (((UINT32)1)<<(btNm))


#define CHECK_IF_CHIP_BUSY(st) \
    {   UINT32_T  rv; \
        I2C_REG_READ(I2C_ISR_REG,rv); \
        st = I2C_REG_BIT_READ(rv,I2C_ISR_UNIT_BUSY_BIT) ? TRUE : FALSE; \
    }


#define CHECK_FOR_EARLY_BUS_BUSY(st) \
    {   UINT32_T  rv; \
        I2C_REG_READ(I2C_ISR_REG,rv); \
        st = I2C_REG_BIT_READ(rv,I2C_ISR_EARLY_IC_BUS_BUSY_BIT) ? TRUE : FALSE; \
    }


#define I2C_RELEASE_BUS_LOOP 	10

/*********************************************************************************************/

typedef struct
{
    UINT8 							activeSlaveAddress;
    UINT8 							*RxBufferPtr;
    UINT16 							dataSize;
    UINT16 							ID;
} I2C_ReceiveRequestParams;



static I2C_ReceiveRequestParams    _receiveReqParams;

/*********************************************************************************************/
#define CHECK_IF_GENERAL_CALL(ad)       ( ((ad <= 0x7) || (ad >= 0xf0)) ? TRUE : FALSE )  // general call is made using the slave address 0x00


#define LOOP_COUNTER_LIMIT              4000L



#define I2C_REGISTER_BASE_ADDRESS   0x40300000
#define I2C_ICR_REG                 0x1690  // Control Register
#define I2C_ISR_REG                 0x1698  // Status Register
#define I2C_ISAR_REG                0x16A0  // Slave Address Register
#define I2C_IDBR_REG                0x1688  // Data Byte Register
#define I2C_IBMR_REG                0x1680  //



#define I2C_REG_CLEAR_VALUE         		0x0000
#define I2C_ICR_CLEAR_ALL_CONTROL_BITS      0xfff0				// Clear the TB, STOP, START, ACK bits in the ICR register

/** bit map of the ICR register **/
#define I2C_ICR_FAST_MODE                           15
#define I2C_ICR_UNIT_RESET_BIT                      14
#define I2C_ICR_IDBR_RECEIVE_FULL_INT_ENABLE_BIT    9
#define I2C_ICR_UNIT_ENABLE_BIT                     6
#define I2C_ICR_SCLEA_ENABLE_BIT                    5
#define I2C_ICR_MASTER_ABORT_BIT                    4
#define I2C_ICR_TRANSFER_BYTE_BIT                   3
#define I2C_ICR_ACK_NACK_CONTROL_BIT                2
#define I2C_ICR_STOP_BIT                            1
#define I2C_ICR_START_BIT                           0


#define I2C_ICR_UNIT_RESET                            (1 << I2C_ICR_UNIT_RESET_BIT)
#define I2C_ICR_IDBR_RECEIVE_FULL_INT_ENABLE          (1 << I2C_ICR_IDBR_RECEIVE_FULL_INT_ENABLE_BIT)
#define I2C_ICR_UNIT_ENABLE                           (1 << I2C_ICR_UNIT_ENABLE_BIT)
#define I2C_ICR_SCLEA_ENABLE                          (1 << I2C_ICR_SCLEA_ENABLE_BIT)
#define I2C_ICR_MASTER_ABORT                          (1 << I2C_ICR_MASTER_ABORT_BIT)
#define I2C_ICR_TRANSFER_BYTE                         (1 << I2C_ICR_TRANSFER_BYTE_BIT)
#define I2C_ICR_ACK_NACK_CONTROL                      (1 << I2C_ICR_ACK_NACK_CONTROL_BIT)
#define I2C_ICR_STOP                                  (1 << I2C_ICR_STOP_BIT)
#define I2C_ICR_START                                 (1 << I2C_ICR_START_BIT)


/** bit map of the ISR register **/
#define I2C_ISR_UNEXPECTED_INTERRUPTS_REPORT			0xff
#define I2C_ISR_UNEXPECTED_INTERRUPTS					0x00000310
#define I2C_ISR_CLEAR_ALL_INTERRUPTS					0x000007f0

#define I2C_ISR_EARLY_IC_BUS_BUSY_BIT					11
#define I2C_ISR_BUS_ERROR_DETECTED_BIT                  10
#define I2C_ISR_SLAVE_ADDRESS_DETECTED_BIT              9
#define I2C_ISR_GENERAL_CALL_ADDRESS_DETECTED_BIT       8
#define I2C_ISR_IDBR_RECEIVE_FULL_BIT                   7
#define I2C_ISR_IDBR_TRANSMIT_EMPTY_BIT                 6
#define I2C_ISR_ARBITRATION_LOSS_DETECTED_BIT           5
#define I2C_ISR_SLAVE_STOP_DETECTED_BIT                 4
#define I2C_ISR_I2C_BUS_BUSY_BIT                        3
#define I2C_ISR_UNIT_BUSY_BIT                           2
#define I2C_ISR_ACK_NACK_STATUS_BIT                     1
#define I2C_ISR_READ_WRITE_MODE_BIT                     0


/** bit map of the IBMR register **/
#define I2C_IBMR_SDA_BIT                     		0
#define I2C_IBMR_SCL_BIT                     		1
#define I2C_IBMR_SCL_AND_SDA_LINES_HIGH				0x3

#define I2C_IBMR_SDA_STATUS_BIT                 (1 << I2C_IBMR_SDA_BIT)
#define I2C_IBMR_SCL_STATUS_BIT                 (1 << I2C_IBMR_SCL_BIT)



#define SET     1
#define CLEAR   0



#define MY_SLAVE_ADD	0x93


/* MASTERS_COLLISION_SILICON_BUG_WORKAROUND*/
/* In 2 chip there was a silicon I2c issue which cause the I2c bus to get stuck
 * The define bellow is a SW W/A for that.
 * For Tavor this should be modify becuase CP14 is not exist on Habell - !!! does not work for Harbell !!!:
 * CP14 - time stamp taken should be replace by time stamp from other free-running timer
 *         resolution should be in 1usec or less, TIME_TO_SAMPLE & CONVERT_TO_MICRO_SEC should be modify.*/

#if defined(_TAVOR_BOERNE_)
   #if !defined(_TAVOR_B0_EARLY_BUS_BUSY_NEW_BIT_)
       #define MASTERS_COLLISION_SILICON_BUG_WORKAROUND
   #endif
#else
	//No need for single chip platfrom , this is only for existing of 2 MASTERS on the I2C bus
#endif




static I2C_BOOL    				_repeat_start;
static I2C_BOOL 				_firstByteToReceive;
static volatile I2C_BOOL        _error_was_detected = FALSE , _arbitration_loss_was_detected = FALSE;


/************************************************************************
* Function: I2CResetUnit
*************************************************************************
* Description:
*
* Parameters:
*
* Return value:
*
* Notes:
************************************************************************/
static void I2CResetUnit ( void )
{
    UINT32_T  	 ICRRegValue;


    // Reset the RX parameter
    _firstByteToReceive = TRUE;

	// Read the control register
	I2C_REG_READ(I2C_ICR_REG,ICRRegValue);

	ICRRegValue &= ~I2C_ICR_MASTER_ABORT;        // Clear the abort bit
  	ICRRegValue &= ~I2C_ICR_START;               // Clear the start bit
  	ICRRegValue &= ~I2C_ICR_STOP;                // Clear the stop bit
	ICRRegValue &= ~I2C_ICR_ACK_NACK_CONTROL;    // Clear the ACK bit
  	ICRRegValue &= ~I2C_ICR_TRANSFER_BYTE;       // Clear the transfer_byte bit

#ifdef MASTERS_COLLISION_SILICON_BUG_WORKAROUND
	ICRRegValue &= ~I2C_ICR_UNIT_ENABLE;         // Clear the unit enable
	ICRRegValue &= ~I2C_ICR_SCLEA_ENABLE;        // Clear the SCL enable
#endif
    // clear the ICR register
    I2C_REG_WRITE(I2C_ICR_REG, I2C_REG_CLEAR_VALUE);

    // Set the RESET bit in the ICR rewgister
    I2C_REG_BIT_WRITE(I2C_ICR_REG, I2C_ICR_UNIT_RESET_BIT, SET);

    // clear the ISR register after reset
    I2C_REG_WRITE(I2C_ISR_REG, I2C_ISR_CLEAR_ALL_INTERRUPTS);

    // Clear the RESET bit in the ICR rewgister
    I2C_REG_BIT_WRITE(I2C_ICR_REG, I2C_ICR_UNIT_RESET_BIT, CLEAR);

    // Restore the ICR register
    I2C_REG_WRITE(I2C_ICR_REG, ICRRegValue);

}
/************************************************************************
* Function: I2CConfigure
*************************************************************************
* Description:
*
* Parameters:
*
* Return value:
*
* Notes:
************************************************************************/
I2C_ReturnCode I2CConfigure( void )
{
    	UINT32 			controlRegValue = I2C_ICR_REG_INIT_VALUE;
        UINT8       	i;


		// reset the I2C unit
    	I2CResetUnit();

#if TAVOR_BOERNE
    	controlRegValue |= FAST_MODE_ENABLE;
#endif

    	I2C_REG_WRITE(I2C_ICR_REG, controlRegValue);

        I2C_REG_WRITE(I2C_ISAR_REG, MY_SLAVE_ADD);

    	return I2C_RC_OK;

}

/************************************************************************
* Function: i2cWaitStatusDirect
*************************************************************************
* Description:  - waits for specific status register value
*               - time is limited
*               - breaks when _error_was_detected indicates error detected by interrupt
*               - breaks when _arbitration_loss_was_detected indicates error detected by interrupt
*
* Parameters:   bitsSet     contains '1' where 1 is expected
*               bitsCleared contains '1' where 0 is expected
* Return value: I2C_RC_OK if got it or error code otherwise
*
* Notes:
************************************************************************/
static I2C_ReturnCode i2cWaitStatusDirect ( UINT32_T    bitsSet , UINT32_T    bitsCleared )
{


	UINT32_T 		statusRegValue;
    UINT32_T 		countLimit = LOOP_COUNTER_LIMIT;
	UINT32_T 		mask = bitsSet|bitsCleared;    // all bits we care of
	UINT32_T 		value = bitsSet;

	do
    {
        I2C_REG_READ(I2C_ISR_REG,statusRegValue);
        countLimit--;
    }
    while ((( statusRegValue & mask ) != value ) && ( _error_was_detected == FALSE ) && (countLimit > 0) && (_arbitration_loss_was_detected == FALSE) );

    if ( I2C_REG_BIT_READ(statusRegValue, I2C_ISR_ARBITRATION_LOSS_DETECTED_BIT) )
		return I2C_ISR_ARBITRATION_LOSS;

    // check for timeout, reset the I2C device for next operations and return error if needed
    if ( countLimit == 0 )
    {
		UINT32_T  ICRRegValue = 0;

		I2C_REG_READ(I2C_ICR_REG,ICRRegValue);
		ICRRegValue |= I2C_ICR_MASTER_ABORT;
		ICRRegValue &= ~I2C_ICR_TRANSFER_BYTE;
    	I2C_REG_WRITE(I2C_ICR_REG, ICRRegValue );
    	return I2C_RC_TIMEOUT_ERROR;
	}
    else
	   if ( ( I2C_REG_BIT_READ(statusRegValue, I2C_ISR_BUS_ERROR_DETECTED_BIT) ))
	        return I2C_ISR_BUS_ERROR;

	return I2C_RC_OK;
}

/************************************************************************
* Function: masterSendDirect
*************************************************************************
* Description: This function is used to write data to I2C slaves in the
* 			   boot stage only while the OS system and INTC are not up.
*
* Parameters:
*
* Return value:
*
* Notes:
************************************************************************/
static I2C_ReturnCode masterSendDirect(UINT8_T *data, UINT16_T dataSize, UINT8_T slaveAddress, I2C_BOOL repeatedStart , I2C_BOOL   MasterReceiveCalled )
{
    UINT32_T  		ICRRegValue = 0;
    int 			i;
	I2C_ReturnCode 	I2CReturnCode = I2C_RC_OK;



	/* Check the bus for free - If there is an arbitration loss, stay until it become free */
    do
	{
        // check if 'Arbitration Loss' was detected
       	if ( I2CReturnCode == I2C_ISR_ARBITRATION_LOSS )
		{
             I2C_STATUS_REG_CLEAR_BIT(I2C_ISR_ARBITRATION_LOSS_DETECTED_BIT);


			 // Read the control register
			 I2C_REG_READ(I2C_ICR_REG,ICRRegValue);

		  	 ICRRegValue &= ~I2C_ICR_START;               // Clear the start bit
		  	 ICRRegValue &= ~I2C_ICR_TRANSFER_BYTE;       // Clear the transfer_byte bit

    		 // Write to the ICR register
    		 I2C_REG_WRITE(I2C_ICR_REG, ICRRegValue);

			 for ( i = 0; i < 50000; i++ );           	/* A little delay to enable others master to finish without interruption */
														/* the bus arbiter, and reduce the arbitration sequences */
		}

    	// write IDBR register: target slave address and R/W# bit=0 for write transaction.
    	I2C_REG_WRITE(I2C_IDBR_REG, I2C_SLAVE_WRITE(slaveAddress) );

    	// write ICR register: set START bit, clear STOP bit, set Transfer Byte bit to initiate the access
        I2C_REG_READ(I2C_ICR_REG,ICRRegValue);
		ICRRegValue &= ~I2C_ICR_STOP;
		ICRRegValue |= (I2C_ICR_START | I2C_ICR_TRANSFER_BYTE);
    	I2C_REG_WRITE(I2C_ICR_REG, ICRRegValue );

	    I2CReturnCode = i2cWaitStatusDirect ( (I2C_REG_BIT(I2C_ISR_IDBR_TRANSMIT_EMPTY_BIT) | I2C_REG_BIT(I2C_ISR_UNIT_BUSY_BIT)),
			    	                     I2C_REG_BIT(I2C_ISR_READ_WRITE_MODE_BIT)  );

    	// clear 'IDBR Transmit Empty' bit (write '1')
    	I2C_STATUS_REG_CLEAR_BIT(I2C_ISR_IDBR_TRANSMIT_EMPTY_BIT);

		if ( ( I2CReturnCode != I2C_ISR_ARBITRATION_LOSS ) && ( I2CReturnCode != I2C_RC_OK ))
		{
        	I2C_REG_READ(I2C_ICR_REG,ICRRegValue);
			ICRRegValue &= 0xfff0;
   			I2C_REG_WRITE(I2C_ICR_REG, ICRRegValue );

			return I2CReturnCode;
		}
    }
    // check if 'Arbitration Loss' was detected
	while ( I2CReturnCode == I2C_ISR_ARBITRATION_LOSS );

    // write ICR register: clear START bit, clear STOP bit
    I2C_REG_READ(I2C_ICR_REG,ICRRegValue);
  	ICRRegValue &= ~I2C_ICR_STOP;
  	ICRRegValue &= ~I2C_ICR_START;
   	I2C_REG_WRITE(I2C_ICR_REG, ICRRegValue );

    /*** Send data bytes - all, except the last byte  ***/
    for ( i = 0; i < ( dataSize - 1 ); i++ )
    {
        // write data byte to the IDBR register
        I2C_REG_WRITE(I2C_IDBR_REG, *data);
        data++;

        // Set 'Tranfer Byte' bit to intiate the access
        I2C_REG_BIT_WRITE(I2C_ICR_REG, I2C_ICR_TRANSFER_BYTE_BIT, SET);

	    if((I2CReturnCode = i2cWaitStatusDirect(I2C_REG_BIT(I2C_ISR_IDBR_TRANSMIT_EMPTY_BIT)|I2C_REG_BIT(I2C_ISR_UNIT_BUSY_BIT),
				                          I2C_REG_BIT(I2C_ISR_READ_WRITE_MODE_BIT))) != I2C_RC_OK)
						return I2CReturnCode;

        // clear 'IDBR Transmit Empty' bit (write '1')
        I2C_STATUS_REG_CLEAR_BIT(I2C_ISR_IDBR_TRANSMIT_EMPTY_BIT);
    }


    /*** Send the last byte with STOP bit ***/

    // write the last data byte to the IDBR register
    I2C_REG_WRITE(I2C_IDBR_REG, *data);

    // write ICR: clear START bit
    I2C_REG_BIT_WRITE(I2C_ICR_REG, I2C_ICR_START_BIT, CLEAR);

    if (( repeatedStart == FALSE ) || ( MasterReceiveCalled == FALSE ))
        I2C_REG_BIT_WRITE(I2C_ICR_REG, I2C_ICR_STOP_BIT, SET);

    I2C_REG_BIT_WRITE(I2C_ICR_REG, I2C_ICR_TRANSFER_BYTE_BIT, SET);

	if ((I2CReturnCode = i2cWaitStatusDirect (I2C_REG_BIT(I2C_ISR_IDBR_TRANSMIT_EMPTY_BIT),
	                                      I2C_REG_BIT(I2C_ISR_READ_WRITE_MODE_BIT)))!=I2C_RC_OK)
				return I2CReturnCode;

	// clear 'IDBR Transmit Empty' bit (write '1')
    I2C_STATUS_REG_CLEAR_BIT(I2C_ISR_IDBR_TRANSMIT_EMPTY_BIT);
    I2C_REG_BIT_WRITE(I2C_ICR_REG, I2C_ICR_STOP_BIT, CLEAR);

    return I2C_RC_OK;
}

/************************************************************************
* Function: I2CBusStateGet
*************************************************************************
* Description: This function is used to read the I2C Bus Monitor
* 			   Register - IBMR.
*
* Parameters:
*
* Return value:
*
* Notes:
************************************************************************/
void I2CBusStateGet ( I2C_IBMR   *bus_state )
{
		UINT32_T     statusRegValue , XCGR_val;



		I2C_REG_READ(I2C_IBMR_REG,statusRegValue);

		bus_state->sda = (statusRegValue & I2C_IBMR_SDA_STATUS_BIT);
		bus_state->scl = ((statusRegValue & I2C_IBMR_SCL_STATUS_BIT ) >> 1);

}
/************************************************************************
* Function: I2CMasterSendDataDirect
*************************************************************************
* Description: This function is used to write data to I2C slaves in the
* 			   boot stage only while the OS system and INTC are not up.
*
* Parameters:
*
* Return value:
*
* Notes:
************************************************************************/
I2C_ReturnCode I2CMasterSendDataDirect( UINT8_T *data , UINT16_T dataSize , UINT8_T slaveAddress )
{
    UINT32_T 			XCGR_val;
    I2C_ReturnCode 		I2CReturnCode;
    I2C_BOOL 			chipBusy;




    CHECK_IF_CHIP_BUSY(chipBusy);
    if ( chipBusy == TRUE )
        return I2C_RC_CHIP_BUSY;

    if ( CHECK_IF_GENERAL_CALL(slaveAddress) )
		return I2C_RC_INVALID_GENERAL_CALL_SLAVE_ADDRESS;

    I2CReturnCode = masterSendDirect(data, dataSize, slaveAddress, _repeat_start , FALSE );

    return I2CReturnCode;
}
/************************************************************************/
/************************************************************************/
static UINT16_T I2CIntLISRDirect ( void )
{
    UINT32_T 			statusRegValue , ICRRegValue = 0;



    //check interrupt source
    I2C_REG_READ(I2C_ISR_REG,statusRegValue);

    if ( I2C_REG_BIT_READ(statusRegValue, I2C_ISR_IDBR_RECEIVE_FULL_BIT) )
    {
        // clear 'IDBR Receive Full' bit (write '1')
        I2C_STATUS_REG_CLEAR_BIT(I2C_ISR_IDBR_RECEIVE_FULL_BIT);
        {
            if ( _firstByteToReceive == TRUE )
            {
            	_firstByteToReceive = FALSE;
            }

            // read the received byte
            I2C_REG_READ(I2C_IDBR_REG,*_receiveReqParams.RxBufferPtr);
            _receiveReqParams.RxBufferPtr++;
            _receiveReqParams.dataSize--;

            // write ICR: clear STOP bit(1), clear ACK/NACK bit(2)
    		I2C_REG_READ(I2C_ICR_REG,ICRRegValue);
			ICRRegValue &= ~I2C_ICR_STOP;
			ICRRegValue &= ~I2C_ICR_ACK_NACK_CONTROL;
    		I2C_REG_WRITE(I2C_ICR_REG, ICRRegValue );


            if ( _receiveReqParams.dataSize > 1 )
            {
            	// write ICR: clear START bit, clear STOP bit, send ACK bit (0 for ACK),
				//             set Transfer Byte bit to initiate the access
    			I2C_REG_READ(I2C_ICR_REG,ICRRegValue);
				ICRRegValue &= ~I2C_ICR_START;
				ICRRegValue &= ~I2C_ICR_STOP;
				ICRRegValue &= ~I2C_ICR_ACK_NACK_CONTROL;
				ICRRegValue |= I2C_ICR_TRANSFER_BYTE;
    			I2C_REG_WRITE(I2C_ICR_REG, ICRRegValue );
            }
            else if ( _receiveReqParams.dataSize == 1 ) //last byte remaing to read (with STOP signal)
            {
            	// write ICR: clear START bit, set STOP bit, send NACK bit (1 for NACK),
				//            set Transfer Byte bit to initiate the access
    			I2C_REG_READ(I2C_ICR_REG,ICRRegValue);
				ICRRegValue &= ~I2C_ICR_START;
				ICRRegValue |= (I2C_ICR_STOP | I2C_ICR_ACK_NACK_CONTROL | I2C_ICR_TRANSFER_BYTE);
    			I2C_REG_WRITE(I2C_ICR_REG, ICRRegValue );
            }
            else
            {
        		// disable 'IDBR Buffer Full' interrupt
        		I2C_REG_BIT_WRITE(I2C_ICR_REG, I2C_ICR_IDBR_RECEIVE_FULL_INT_ENABLE_BIT, CLEAR);
            	_firstByteToReceive = TRUE;
            }
        }
    }

    if ( I2C_REG_BIT_READ(statusRegValue, I2C_ISR_BUS_ERROR_DETECTED_BIT) )
    {
        // clear 'Bus Error Detected' bit (write '1')
        I2C_STATUS_REG_CLEAR_BIT(I2C_ISR_BUS_ERROR_DETECTED_BIT);

        _error_was_detected = TRUE;
    }

    if ( I2C_REG_BIT_READ(statusRegValue, I2C_ISR_ARBITRATION_LOSS_DETECTED_BIT) )
    {
        // clear 'Arbitration Loss Detected' bit (write '1')
        I2C_STATUS_REG_CLEAR_BIT(I2C_ISR_ARBITRATION_LOSS_DETECTED_BIT);

        _arbitration_loss_was_detected = TRUE;
    }

    if ( statusRegValue & I2C_ISR_UNEXPECTED_INTERRUPTS )
    {
        // clear 'I2C_ISR_UNEXPECTED_INTERRUPTS ' bits (write '1')
        I2C_REG_WRITE(I2C_ISR_REG,I2C_ISR_UNEXPECTED_INTERRUPTS);

    }

	return _receiveReqParams.dataSize;

}
/************************************************************************
* Function: I2CMasterReceiveDataDirect
*************************************************************************
* Description: This function is used to read data from I2C slaves in the
* 			   boot stage only while the OS system and INTC are not up.
*
* Parameters:
*
* Return value:
*
* Notes:
************************************************************************/
I2C_ReturnCode I2CMasterReceiveDataDirect ( UINT8_T 						   *cmd,
											UINT16_T 	 						cmdLength,
											UINT8_T 							writeSlaveAddress,
		                                    UINT8_T 						   *designatedRxBufferPtr,
											UINT16_T 	 						dataSize,
											UINT8_T 							readSlaveAddress )
{
    UINT32_T 			XCGR_val;
    I2C_ReturnCode 		I2CReturnCode;
    UINT32_T 			ICRRegValue = 0;
    I2C_BOOL 				chipBusy;



    // store the request parameters (to be used from the ISR and by the I2C task when error)
    _receiveReqParams.RxBufferPtr = designatedRxBufferPtr;
    _receiveReqParams.dataSize = dataSize;


    if ( dataSize == 0 )
        return I2C_RC_INVALID_DATA_SIZE;

    if ( CHECK_IF_GENERAL_CALL(writeSlaveAddress) )
		return I2C_RC_INVALID_GENERAL_CALL_SLAVE_ADDRESS;

    if ( CHECK_IF_GENERAL_CALL(readSlaveAddress) )
		return I2C_RC_INVALID_GENERAL_CALL_SLAVE_ADDRESS;

    CHECK_IF_CHIP_BUSY(chipBusy);
    if(chipBusy)
        return I2C_RC_CHIP_BUSY;

    if ( (cmdLength != 0) && (cmd != NULL) )  /* if command is associated with the received request */
    {
        I2CReturnCode = masterSendDirect(cmd, cmdLength, writeSlaveAddress, _repeat_start, TRUE);

        if( I2CReturnCode != I2C_RC_OK )
            return I2CReturnCode;
    }

    /*** send Read Request ***/
    // write IDBR: target slave address and R/W# bit (1 for read)
    I2C_REG_WRITE(I2C_IDBR_REG, I2C_SLAVE_READ(readSlaveAddress) );

    // write ICR: set START bit, clear STOP bit, set Transfer Byte bit to initiate the access
    I2C_REG_READ(I2C_ICR_REG,ICRRegValue);
	ICRRegValue &= ~I2C_ICR_STOP;
	ICRRegValue |= (I2C_ICR_START | I2C_ICR_TRANSFER_BYTE);
    I2C_REG_WRITE(I2C_ICR_REG, ICRRegValue );

	I2CReturnCode = i2cWaitStatusDirect(I2C_REG_BIT(I2C_ISR_IDBR_TRANSMIT_EMPTY_BIT)|I2C_REG_BIT(I2C_ISR_UNIT_BUSY_BIT)|I2C_REG_BIT(I2C_ISR_READ_WRITE_MODE_BIT), 0 );

    // clear 'IDBR Transmit Empty' bit (write '1')
    I2C_STATUS_REG_CLEAR_BIT(I2C_ISR_IDBR_TRANSMIT_EMPTY_BIT);

	if ( I2CReturnCode != I2C_RC_OK )
        return I2CReturnCode;


    // Initiate the read process:

    if ( dataSize == 1 )   // only one byte to read
    {
        // write ICR: set STOP bit, set ACK/NACK bit (1 for NACK)
        I2C_REG_READ(I2C_ICR_REG,ICRRegValue);
  		ICRRegValue &= ~I2C_ICR_START;
  	    ICRRegValue |= (I2C_ICR_STOP | I2C_ICR_ACK_NACK_CONTROL);
   	    I2C_REG_WRITE(I2C_ICR_REG, ICRRegValue );
    }
    else   // more than one byte to read
    {
        // write ICR: clear STOP bit, clear ACK/NACK bit (0 for ACK)
    	I2C_REG_READ(I2C_ICR_REG,ICRRegValue);
		ICRRegValue &= ~I2C_ICR_STOP;
		ICRRegValue &= ~I2C_ICR_ACK_NACK_CONTROL;
    	I2C_REG_WRITE(I2C_ICR_REG, ICRRegValue );
    }

    I2C_REG_READ(I2C_ICR_REG,ICRRegValue);
  	ICRRegValue &= ~I2C_ICR_START;
  	ICRRegValue |= I2C_ICR_TRANSFER_BYTE;
   	I2C_REG_WRITE(I2C_ICR_REG, ICRRegValue );

	while ( I2CIntLISRDirect() );

    return I2C_RC_OK;
}
/***************************************************************************/
/***************************************************************************/
void I2CDisplayRegsStatus ( void )
{
    UINT32_T  	 	RegValue;
	UINT8_T Msg1[] = "-I2C registers value are :\0";
	UINT8_T Msg[]  = "                    ICR  : 0xxxxxxxxx\0";


 	AddMessage(Msg1);

    I2C_REG_READ(I2C_ICR_REG,RegValue);
 	ConvertIntToBuf8(&Msg[29], RegValue, 8,16);
 	AddMessage(Msg);                               // Display the ICR value

    I2C_REG_READ(I2C_ISR_REG,RegValue);
	Msg[21] = 'S';
 	ConvertIntToBuf8(&Msg[29], RegValue, 8,16);
 	AddMessage(Msg);                               // Display the ISR value

	I2C_REG_READ(I2C_IBMR_REG,RegValue);
	Msg[21] = 'B';
	Msg[22] = 'M';
	Msg[23] = 'R';
 	ConvertIntToBuf8(&Msg[29], RegValue, 8,16);
 	AddMessage(Msg);                               // Display the IBMR value

}
/***************************************************************************/
/***************************************************************************/
UINT8 I2CReleaseBus ( void )
{
    I2C_IBMR   	   	bus_state;
    UINT8       	i = 0;


	while ( i++ < I2C_RELEASE_BUS_LOOP )
	{
		sendClocksOnI2cClkPin();   						   // Send the i2c clock pulse
	}

	I2CBusStateGet ( &bus_state );

	return bus_state.sda ;
}
/***************************************************************************/
/***************************************************************************/
void I2CErrorHandle ( I2C_BUS_CYCLE      bus_cycle )
{
	UINT8_T Msg[] = "********** I2C Micco WD  read error, Reset was done, retrying...\0";


	I2CDisplayRegsStatus();
	I2CConfigure();
	if ( bus_cycle == I2C_WRITE_CYCLE )
	{
		Msg[24] = 'w';
		Msg[26] = 'i';
		Msg[27] = 't';
		Msg[28] = 'e';
	}
 	AddMessage(Msg);
}
/***************************************************************************/
/***************************************************************************/
I2C_ReturnCode I2CMiccoReceiveData ( UINT8_T 	 *dataPtr )
{
    UINT8       		num_of_retry = 0;
	I2C_ReturnCode		rc;


	do
	{
	 	rc = I2CMasterReceiveDataDirect ( &dataPtr[0] ,
			 						 	   MICCO_CMD_1_BYTE_SIZE,
					 					   PMC_I2C_WRITE_SLAVE_ADDRESS,
					                      &dataPtr[1],
					 					   MICCO_READ_1_BYTE,
					 					   PMC_I2C_READ_SLAVE_ADDRESS );
	}
	while (( rc != I2C_RC_OK ) && ( ++num_of_retry < I2C_MICCO_MAX_RETRY ));

	return rc;
}
/***************************************************************************/
/***************************************************************************/
I2C_ReturnCode I2CMiccoSendData ( UINT8_T 	 *dataPtr )
{
    UINT8       		num_of_retry = 0;
	I2C_ReturnCode		rc;


	do
	{
		rc = I2CMasterSendDataDirect ( dataPtr , MICCO_WRITE_2_BYTE , PMC_I2C_WRITE_SLAVE_ADDRESS );
	}
	while (( rc != I2C_RC_OK ) && ( ++num_of_retry < I2C_MICCO_MAX_RETRY ));

	return rc;
}
/***************************************************************************/
/***************************************************************************/
int I2CBusHungHandle ( void )
{
    I2C_IBMR   	   	bus_state;
	int				pin_mux_rc;


    /* In case I2C is stuck in non-idle state, try to release the bus */
    I2CBusStateGet ( &bus_state );
    if ( bus_state.sda == LOW )
    {
	    I2CDisplayRegsStatus();
		AddMessage((UINT8_T*) ("********** I2C bus is NOT in idle state. Waiting to be Idle...\0"));

		setI2CPinsToGpio();

		while ( I2CReleaseBus() == 0 );         // 0 means An Error

		AddMessage((UINT8_T*) ("-I2C bus Return to idle state.\0"));
		pin_mux_rc = setTavorI2CPinMux();
		if ( pin_mux_rc == 0 )              	// 0 means An Error
	 		return GeneralError;
    }

}
/***************************************************************************/
/***************************************************************************/
