/**************************************************************************
**
**  (C) Copyright December 2006 Marvell International Ltd.
**
**  All Rights Reserved.
**
**  This software file (the "File") is distributed by Marvell International  
**  Ltd. under the terms of the GNU General Public License Version 2, June  
**  1991 (the "License").  You may use, redistribute and/or modify this File   
**  in accordance with the terms and conditions of the License, a copy of  
**  which is available along with the File in the license.txt file or by   
**  writing to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   
**  Boston, MA 02111-1307 or on the worldwide web at 
**  http://www.gnu.org/licenses/gpl.txt.
**
**  THE FILE IS DISTRIBUTED AS-IS, WITHOUT WARRANTY OF ANY KIND, AND THE   
**  IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR  
**  PURPOSE ARE EXPRESSLY DISCLAIMED.  The License provides additional
**  details about this warranty disclaimer. 
**
**  Filename: NTBBTIM.cpp
**
**  This program builds the TIM binary image from a TIM descriptor file and all of
**  the binary images that it lists.
**  This is a console application.
**  
******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#if LINUX
  #include <ctype.h>
  #include <sys/types.h>
  #include <fcntl.h>
  #include <sys/stat.h>
  #include <limits.h>
  #include "typedef.h"
#endif
 
#if WINDOWS
  
#endif

#if TRUSTED
  #include "TBBLIB.h"
#endif

#include "NTBBLIB.h"
#include "tim.h"

#define VERBOSE 'v'
#define HELP '?'
#define KEY 'r'
#define TIMIN 'i'
#define TIMOUT 'o'
#define HASH 'h'
#define DS 's'
#define OPTION 'm'
#define TIMVERIFY 'V'
#define RESERVED 'R'
#define CONCATENATE 'C'
#define MIN_ARGS 2

char Version[11] = "3.15";
char Date[11] = "05072008";
const unsigned int MSysSectorSize = 512;
bool bIsReservedDataInFile;
bool bIsKeyFile, bIsHashFile, bIsDsFile, bIsTimInFile, bIsTimOutFile, bIsNonTrusted;
bool bIsTimVerify;
char KeyFileName[MAX_PATH], HashFileName[MAX_PATH], DsFileName[MAX_PATH];
char TimInFileName[MAX_PATH], TimOutFileName[MAX_PATH];
char ReservedFileName[MAX_PATH];
int iOption = 1;
bool bVerbose, bConcatenate;

bool ParseCommandLine (int argc,char* argv[]);
bool ParseDescriptorFile (char *FileName);
bool ParseNonTrustedDescriptorFile (FILE *hKeyFile,CTIM *pTimHeader);
extern bool ScanImageInfo (FILE *hKeyFile,CTIM *pTimHeader,IMAGE_INFO *pImageInfo);
bool GetDWord (FILE *hKeyFile,char *szString,UINT_T *pValue);
bool CheckYourOptions ();
bool CheckImageOverlap(pIMAGE_INFO pImageInfo, unsigned int num_images, unsigned int BootFlashSign);
void PrintUsage ();
bool ParseTrustedDescriptorFile (FILE *hKeyFile,CTIM *pTimHeader);

#pragma warning ( disable : 4096 4081 ) // Disable warning messages


int main (int argc,char *argv[])
{
	bVerbose = FALSE;
	bIsKeyFile = FALSE;
	bIsHashFile = FALSE;
	bIsDsFile = FALSE;
	bIsTimInFile = FALSE;
	bIsTimOutFile = FALSE;
	bIsTimVerify = FALSE;
	bIsNonTrusted = FALSE;
	bIsReservedDataInFile = FALSE;
	bConcatenate = FALSE;

	printf ("\n\nTBB Version: %s\n", Version);
	printf ("TBB Date   : %s\n\n", Date);

	if (argc < MIN_ARGS)
	{
		PrintUsage ();
		return FALSE;
	}
	
	if (ParseCommandLine (argc,argv) == FALSE) return FALSE;

	if ((iOption == 1) || (iOption == 2)) 
	{
		ParseDescriptorFile (KeyFileName);
	}
	else if(iOption == 3)
	{ 
		#if TRUSTED
		BuildDigitalSignature (KeyFileName,HashFileName);
		#endif
	}
	else if(iOption == 4)
	{
		#if TRUSTED 
		AddDigitalSignatureToTim (TimInFileName,TimOutFileName,DsFileName);
		#endif
	}
	else
		printf ("  Error: Invalid Option Mode Selected!");

	return TRUE;
}

bool ParseDescriptorFile (char *FileName)
{
	FILE *hKeyFile;
	CTIM TimHeader;
	char szDword[11];

	if ((hKeyFile = fopen (FileName,"r")) == NULL)
	{
		printf ("  Error: Cannot open file name %s !",FileName);
		return FALSE;
	}

	if (VerifyDescriptorIntegrity (hKeyFile) == FALSE)
	{
		if (hKeyFile != NULL) fclose (hKeyFile);
		return FALSE;
	}

	memset (&TimHeader,0,sizeof (CTIM));
	TimHeader.VersionBind.Identifier = TIMIDENTIFIER;

	if (fscanf (hKeyFile,"Version: %s\n",szDword) != 1)
	{
		printf ("  Error: Key file parsing error reading, Version:\n");
		if (hKeyFile != NULL) fclose (hKeyFile);
		return FALSE;
	}

	TimHeader.VersionBind.Version = Translate (szDword);

	if (fscanf (hKeyFile,"Trusted: %s\n",szDword) != 1)
	{
		printf ("  Error: Key file parsing error reading, Trusted:\n");
		if (hKeyFile != NULL) fclose (hKeyFile);
		return FALSE;
	}

	TimHeader.VersionBind.Trusted = Translate (szDword);

	if (TimHeader.VersionBind.Trusted == 0)
	{
		bIsNonTrusted = TRUE;
		if (CheckYourOptions () == FALSE) return FALSE;
		if (iOption == 1) ParseNonTrustedDescriptorFile (hKeyFile,&TimHeader);
		else
		{
			printf ("////////////////////////////////////////////////////////////\n");
			printf ("  Error: You have selected option mode 2, 3 or 4!\n");
			printf ("You can only build a Non Trusted Image in option mode 1!\n");
			printf ("Example: tbb -m 1 -k ntimkey.txt\n\n");
			printf ("Also check the Trusted directive in the TIM descriptor file.\n");
			printf ("Example: Trusted: 0  (for Non Trusted Image)\n");
			printf ("Example: Trusted: 1  (for Trusted Image)\n");
			printf ("////////////////////////////////////////////////////////////\n");
			if (hKeyFile != NULL) fclose (hKeyFile);
			return FALSE;
		}
	}
	else
	{
		#if TRUSTED
		ParseTrustedDescriptorFile (hKeyFile,&TimHeader);
		#else
		printf("  Error: You can only build a Non Trusted Image \n");
		printf("\twith the Non Trusted only version of the TBB Tool! \n\n");
		printf("Processing has stopped!\n");
		return FALSE;
		#endif
	}

	if (hKeyFile != NULL) fclose (hKeyFile);

	return TRUE;
}

bool ParseNonTrustedDescriptorFile (FILE *hKeyFile,CTIM *pTimHeader)
{
	char szTimFilename[MAX_PATH], szImageFilename[MAX_PATH], szFileNamesList[MAX_IMAGE_FILES][MAX_PATH];
	UINT_T IntParam, NextImageID = 0;
	UINT32_T ImageCrc;
	pIMAGE_INFO pImageInfo;
	FILE *hTimFile, *hImage, *hReservedFile, *hBinaryImage;
	char ReservedBuffer[MAX_PATH];
	unsigned int TimSizeToHash;
	char fLine[100], *pData = NULL;
	int numberOfLinesRead = 0, sizeOfReservedDataFile=0;
	DWORD dwValue;

	printf ("\nProcessing non trusted TIM...\n\n");
	
	if (GetDWord  (hKeyFile,"Issue Date: %s\n",&IntParam) == FALSE) return FALSE;
	pTimHeader->VersionBind.IssueDate = IntParam;
	if (GetDWord (hKeyFile,"OEM UniqueID: %s\n",&IntParam) == FALSE) return FALSE;
	pTimHeader->VersionBind.OEMUniqueID = IntParam;
	if (GetDWord (hKeyFile,"Boot Flash Signature: %s\n",&IntParam) == FALSE) return FALSE;
	pTimHeader->FlashInfo.BootFlashSign = IntParam;
	if (GetDWord (hKeyFile,"Number of Images: %s\n",&IntParam) == FALSE) return FALSE;
	pTimHeader->NumImages = IntParam;
	
	if (bIsReservedDataInFile)
	{
		if ((hReservedFile = fopen (ReservedFileName,"rb")) == NULL)
		{
			printf ("  Error: Cannot open file name %s !",ReservedFileName);
			if (hReservedFile != NULL) fclose (hReservedFile);
			return FALSE;
		}
		fseek (hReservedFile,0L,SEEK_END); // Set position to EOF
		sizeOfReservedDataFile = ftell (hReservedFile); // Get position of file, thus the file size.
		fseek (hReservedFile,0L,SEEK_SET); // Set position to SOF
	}
//	else
//	{
		if (GetDWord (hKeyFile,"Size of Reserved in bytes: %s\n",&IntParam) == FALSE) return FALSE;
		pTimHeader->SizeOfReserved = IntParam;
//	}

	if (bIsReservedDataInFile)
	{
		pTimHeader->SizeOfReserved+= sizeOfReservedDataFile;
	}

	pTimHeader->FlashInfo.WTMFlashSign = 0xFFFFFFFF;
	pTimHeader->FlashInfo.WTMEntryAddr = 0xFFFFFFFF;
	pTimHeader->FlashInfo.WTMEntryAddrBack = 0xFFFFFFFF;
	pTimHeader->FlashInfo.WTMPatchSign = 0xFFFFFFFF;
	pTimHeader->FlashInfo.WTMPatchAddr = 0xFFFFFFFF;

	// allocate space for all of the Image ID blocks
	pImageInfo = (pIMAGE_INFO) calloc((int)pTimHeader->NumImages, sizeof(IMAGE_INFO));

	for (int i = 0; i < (int)pTimHeader->NumImages; i++)
	{
		if (ScanImageInfo (hKeyFile,pTimHeader,&pImageInfo[i]) == FALSE) return FALSE;

		if (i == 0)
		{
			if (pImageInfo[i].ImageID != TIMIDENTIFIER)
			{
				printf ("  Error: The ImageID value is incorrect for the TIM image!\n");
				return FALSE;
			}

			if (fscanf (hKeyFile,"Image Filename: %s\n",szTimFilename) != 1)
			{
				printf ("  Error: Key file parsing error reading, Image Filename:\n");
				return FALSE;
			}
			if ((hTimFile = fopen (szTimFilename,"wb+")) == NULL)
			{
				printf ("  Error: Cannot open file name %s !",szTimFilename);
				return FALSE;
			}

			pImageInfo[i].ImageSize = sizeof (CTIM) + (sizeof (IMAGE_INFO) * pTimHeader->NumImages)
								+ pTimHeader->SizeOfReserved;

			if(pImageInfo[i].ImageSizeToHash == 0xFF || pImageInfo[i].ImageSizeToHash == 0xFFFFFFFF) 
				pImageInfo[i].ImageSizeToHash = pImageInfo[i].ImageSize;
//			else 
//				pImageInfo[i].ImageSizeToHash = 0x0;
			TimSizeToHash = pImageInfo[i].ImageSizeToHash;
			memset (&pImageInfo[i].Hash,0,8);
			fwrite (pTimHeader,sizeof (CTIM),1,hTimFile);

		}
		else
		{
			if (pImageInfo[i].ImageID != NextImageID)
			{
				printf ("  Error: The ImageID doesn't = NextImageID from the previous image!");
				if (hTimFile != NULL) fclose (hTimFile);
				return FALSE;
			}

			if (fscanf (hKeyFile,"Image Filename: %s\n",szImageFilename) != 1)
			{
				printf ("  Error: Key file parsing error reading, Image Filename:\n");
				if (hTimFile != NULL) fclose (hTimFile);
				return FALSE;
			}

			for (int m = i-1; m<=i-1;m++)
			{
				for (int n=0; n<MAX_PATH;n++)
				{
					szFileNamesList[m][n] = szImageFilename[n];
				}
			}

			if ((hImage = fopen (szImageFilename,"rb")) == NULL)
			{
				printf ("  Error: Cannot open file name %s !",szImageFilename);
				if (hTimFile != NULL) fclose (hTimFile);
				return FALSE;
			}

			fseek (hImage,0L,SEEK_END); // Set position to EOF
			pImageInfo[i].ImageSize = ftell (hImage); // Get position of file, thus the file size.
			//printf("Size: %d\n", pImageInfo[i].ImageSize);
			// Output new image with ImageID prepended
			if (PrependID (hImage,szImageFilename,&pImageInfo[i]) == FALSE)
			{
				if (hTimFile != NULL) fclose (hTimFile);
				if (hImage != NULL) fclose (hImage);
				return FALSE;
			}
			if (hImage != NULL) fclose (hImage);
			if ((pImageInfo[i].ImageSizeToHash > pImageInfo[i].ImageSize) ||
				(pImageInfo[i].ImageSizeToHash == 0xFFFFFFFF))
			{
				pImageInfo[i].ImageSizeToHash = pImageInfo[i].ImageSize;
			}
			
			// Set all hash values to 0 then place ImageCrc in first hash entry
			memset (&pImageInfo[i].Hash,0,8);
			if(pImageInfo[i].ImageSizeToHash != 0){
				//disable CRC FOR NOW
				/*if (GenerateImageCrc (hImage,pImageInfo[i].ImageSizeToHash,&ImageCrc) == FALSE)
				{
					if (hTimFile != NULL) fclose (hTimFile);
					if (hImage != NULL) fclose (hImage);
					return FALSE;
				}
				pImageInfo[i].Hash[0] = ImageCrc;
				*/
			}

		}

		fwrite (&pImageInfo[i],sizeof (IMAGE_INFO),1,hTimFile);

		NextImageID = pImageInfo[i].NextImageID;
	}
	//check to make sure the images do not overlap each other
	if(CheckImageOverlap(pImageInfo, pTimHeader->NumImages, pTimHeader->FlashInfo.BootFlashSign) == FALSE)
	{
		return FALSE;
	} 

	// skip the reserved section if size is 0.
	if(pTimHeader->SizeOfReserved != 0){
		if (ProcessReservedData (hKeyFile,hTimFile,pTimHeader->SizeOfReserved) == FALSE)
		{
			if (hTimFile != NULL) fclose (hTimFile);
			return FALSE;
		}
	}

	if (bIsReservedDataInFile)
	{
		memset (ReservedBuffer,0,sizeOfReservedDataFile);
		fread (ReservedBuffer,1,sizeOfReservedDataFile,hReservedFile);
		fwrite (ReservedBuffer,sizeof (char),sizeOfReservedDataFile,hTimFile);
		if (hReservedFile != NULL) fclose (hReservedFile);
	}

	if(TimSizeToHash!=0){
//	if(ImageInfo.ImageSizeToHash != 0){
		;//GenerateTimCrc (hTimFile);  for now disable CRC
	}

	unsigned int location = 0;
	BYTE padding = 0x0;
	BYTE Data[8192];
	int iBlockSize = 8192;

	if (bConcatenate)
	{
		if(pTimHeader->FlashInfo.BootFlashSign == 0x4e414e01)
		{
			printf("Cannot concatenate images for an MDOC device!\n");
			return FALSE;
		}
	 	
	 	for(int j=1;j<(int)pTimHeader->NumImages - 1;j++)
		{
			fseek (hTimFile,0L,SEEK_END); // Set position to EOF
			location = ftell(hTimFile);
			
			//pad out the current binary until we get to the location of where the next image should start
			if(pImageInfo[j].FlashEntryAddr > location)
				fwrite(&padding,1,(pImageInfo[j].FlashEntryAddr - location),hTimFile);

			//open the next image file
			if ((hBinaryImage = fopen (szFileNamesList[j],"rb")) == NULL)
			{
				printf ("  Error: Cannot open file name %s !",szFileNamesList[j]);
				return FALSE;
			}

			//now write out image in 8K chunks
			while (iBlockSize == 8192)
			{
				iBlockSize = fread (Data,1,iBlockSize,hBinaryImage);
				fwrite (Data,1,iBlockSize,hTimFile);
			}
			iBlockSize = 8192;
			//close the image file
			if (hBinaryImage != NULL)
				fclose (hBinaryImage);
		}
	}

	printf ("  Success: Processing has completed successfully!\n");

	if (hTimFile != NULL) fclose (hTimFile);	
	free(pImageInfo);
	return TRUE;
}

bool ScanImageInfo (FILE *hKeyFile,CTIM *pTimHeader,IMAGE_INFO *pImageInfo)
{

	UINT_T IntParam;

	// This is a helper function to parse image info data from the descriptor
	// text file supplied by the user.

	if (GetDWord (hKeyFile,"Image ID: %s\n",&IntParam) == FALSE) return FALSE;
	pImageInfo->ImageID = IntParam;
	if (GetDWord (hKeyFile,"Next Image ID: %s\n",&IntParam) == FALSE) return FALSE;
	pImageInfo->NextImageID = IntParam;
	if (GetDWord (hKeyFile,"Flash Entry Address: %s\n",&IntParam) == FALSE) return FALSE;
	pImageInfo->FlashEntryAddr = IntParam;
	if (GetDWord (hKeyFile,"Load Address: %s\n",&IntParam) == FALSE) return FALSE;
	pImageInfo->LoadAddr = IntParam;
	if(pTimHeader->VersionBind.Trusted){
        if (GetDWord (hKeyFile,"Image Size To Hash in bytes: %s\n",&IntParam) == FALSE) return FALSE;
        pImageInfo->ImageSizeToHash = IntParam;
        if (GetDWord (hKeyFile,"Hash Algorithm ID: %s\n",&IntParam) == FALSE) return FALSE;
        pImageInfo->HashAlgorithmID = IntParam;
	}else{
        if (GetDWord (hKeyFile,"Image Size To CRC in bytes: %s\n",&IntParam) == FALSE) return FALSE;
        pImageInfo->ImageSizeToHash = IntParam;
	}

	return TRUE;
}

bool GetDWord (FILE *hKeyFile,char *szString,UINT_T *pValue)
{
	char szDword[11], szDeclaration[50];
	int i = 0;

	// This is a helper function that converts strings to integers.

	if (fscanf (hKeyFile,szString,szDword) != 1)
	{
		strcpy (szDeclaration,szString);

		for (i = 0; i < 50; i++)
		{
			if (szDeclaration[i] == ':') szDeclaration[i] = '\0';
		}

		printf ("////////////////////////////////////////////////////////////\n");
		printf ("  Error: Declaration statement is missing or corrupted in descriptor file!\n");
		printf ("%s\n",szDeclaration);
		printf ("////////////////////////////////////////////////////////////\n");
		return FALSE;
	}
	*pValue = Translate (szDword);

	return TRUE;
}

bool ParseCommandLine (int argc,char *argv[])
{
	int currentOption = 1;
	BYTE Option;

	while (currentOption < argc)
	{
#if WINDOWS
		if ((*argv[currentOption] == '-') || (*argv[currentOption] == '/'))
#else
		if (*argv[currentOption] == '-')
#endif
		{
			Option = *++argv[currentOption++];

			switch (Option)
			{
				case OPTION:
					iOption = atoi (argv[currentOption]);
					if (iOption > 4 || iOption < 1)
					{
						iOption = 1;
						printf("\n  Error: -m Parameter Invalid or Missing!\n"); 
						printf("Defaulting to Option Mode 1...\n");
					}
					break;
				case KEY:
					if (currentOption == argc)
					{
						printf ("\n  Error:   Command Line Parsing Error:");
						printf (" Missing Key file name parameter.\n");
						return FALSE;
					}
					strcpy (KeyFileName,argv[currentOption]);
					bIsKeyFile = TRUE;
					break;
				case HASH:
					if (currentOption == argc)
					{
						printf ("\n  Error:   Command Line Parsing Error:");
						printf (" Missing hash file name parameter.\n");
						return FALSE;
					}
					strcpy (HashFileName,argv[currentOption]);
					bIsHashFile = TRUE;
					break;
				case DS:
					if (currentOption == argc)
					{
						printf ("\n  Error:   Command Line Parsing Error:");
						printf (" Missing hash file name parameter.\n");
						return FALSE;
					}
					strcpy (DsFileName,argv[currentOption]);
					bIsDsFile = TRUE;
					break;
				case TIMIN:
					if (currentOption == argc)
					{
						printf ("\n  Error:   Command Line Parsing Error:");
						printf (" Missing hash file name parameter.\n");
						return FALSE;
					}
					strcpy (TimInFileName,argv[currentOption]);
					bIsTimInFile = TRUE;
					break;
				case TIMOUT:
					if (currentOption == argc)
					{
						printf ("\n  Error:   Command Line Parsing Error:");
						printf (" Missing hash file name parameter.\n");
						return FALSE;
					}
					strcpy (TimOutFileName,argv[currentOption]);
					bIsTimOutFile = TRUE;
					break;
				case VERBOSE:
					bVerbose = TRUE;
					break;
				case CONCATENATE:
					bConcatenate = TRUE;
					break;
				case TIMVERIFY:
					bIsTimVerify = TRUE;
					break;
				case RESERVED:
					if (currentOption == argc)
					{
						printf ("\n  Error:   Command Line Parsing Error:");
						printf (" Missing reserved data file name parameter.\n");
						return FALSE;
					}
					strcpy (ReservedFileName,argv[currentOption]);
					bIsReservedDataInFile = TRUE;
					break;
				case HELP:
					PrintUsage();
					break;
				default:
					printf ("  Error: Command line option -%c is invalid!\n",Option);
					return FALSE;
			}
		}
		else
		{
			currentOption++;
		}
	}

	if (CheckYourOptions () == FALSE) return FALSE;

	return TRUE;
}

bool CheckYourOptions ()
{

	// Check for TIM verify functionality for trusted operation.
	if ( ( iOption != 1 ) && ( bIsNonTrusted && bIsTimVerify ) )
	{
		printf ("////////////////////////////////////////////////////////////\n");
		printf ("  Error: Invalid Options Error:\n");
		printf ("Missing switch, non-trusted descriptor file or incorrect option mode.\n");
		printf ("Example: tbb -m 1 -r timdescriptorfile.txt -V\n");
		printf ("////////////////////////////////////////////////////////////\n");
		return FALSE;
	}
	switch (iOption)
	{
		case 1:
			if (!bIsKeyFile)
			{
				printf ("////////////////////////////////////////////////////////////\n");
				printf ("  Error: Invalid Options Error:\n");
				printf ("Missing TIM descriptor file name parameter. (-r filename.txt)\n");
				printf ("Example: tbb -m 1 -r timdescriptorfile.txt\n");
				printf ("////////////////////////////////////////////////////////////\n");
				return FALSE;
			}
			break;
		case 2:
			if (!bIsKeyFile || !bIsHashFile)
			{
				printf ("////////////////////////////////////////////////////////////\n");
				printf ("  Error: Invalid Options Error:\n");
				printf ("Missing TIM descriptor file name parameter. (-r filename.txt) OR\n");
				printf ("Missing output hash key file name parameter. (-h filename.bin)\n");
				printf ("Example: tbb -m 2 -r timdescriptorfile.txt -h timhash.bin\n");
				printf ("////////////////////////////////////////////////////////////\n");
				return FALSE;
			}
			break;
		case 3:
			if (!bIsKeyFile || !bIsHashFile || !bIsDsFile)
			{
				printf ("////////////////////////////////////////////////////////////\n");
				printf ("  Error: Invalid Options Error:\n");
				printf ("Missing private key file name parameter. (-r filename.txt) OR\n");
				printf ("Missing input hash key file name parameter. (-h filename.bin) OR\n");
				printf ("Missing output digital signature file name parameter. (-s filename.bin)\n");
				printf ("////////////////////////////////////////////////////////////\n");
				return FALSE;
			}
			break;
		case 4:
			if (!bIsTimInFile || !bIsTimOutFile || !bIsDsFile)
			{
				printf ("////////////////////////////////////////////////////////////\n");
				printf ("  Error: Invalid Options Error:\n");
				printf ("Missing input TIM file name parameter. (-i filename.bin) OR\n");
				printf ("Missing output TIM file name parameter. (-o filename.bin) OR\n");
				printf ("Missing input digital signature file name parameter. (-s filename.bin)\n");
				printf ("////////////////////////////////////////////////////////////\n");
				return FALSE;
			}
			break;
		default:
			printf ("////////////////////////////////////////////////////////////\n");
			printf ("  Error: Invalid Options Error:\n");
			printf ("You entered an incorrect value for the -m option or\n");
			printf ("Example: tbb -m 1 -r TimDescriptorFile.txt\n");
			printf ("Note: If you leave out the -m option the program will\n");
			printf ("      default this option to 1.\n");
			printf ("////////////////////////////////////////////////////////////\n");
			return FALSE;
	}

	return TRUE;
}

bool CheckImageOverlap(pIMAGE_INFO pImageInfo, unsigned int num_images, unsigned int BootFlashSign)
{
	unsigned int i, j;
	unsigned int s1, s2, e1, e2;//starting and ending addresses of the two images being compared


	for(i = 0; i < num_images; i++)
	{
		//printf("Size: %d\n", pImageInfo[i].ImageSize);
		for(j = i + 1; j < num_images; j++)	//don't compare the image to itself
		{
			if(BootFlashSign == 0x4e414e01)//MDOC flash location format is different than all other flashes
			{
				if((pImageInfo[i].FlashEntryAddr & 0xF) != (pImageInfo[j].FlashEntryAddr & 0xF))
					continue;  //if the images are NOT on the same partition, then they won't overlap... so continue search

				s1 = ((pImageInfo[i].FlashEntryAddr & 0xFFFFFFF0) >> 4) * MSysSectorSize;
				s2 = ((pImageInfo[j].FlashEntryAddr & 0xFFFFFFF0) >> 4) * MSysSectorSize;
			}
			else // standard absolute address compare
			{
				s1 = pImageInfo[i].FlashEntryAddr;
				s2 = pImageInfo[j].FlashEntryAddr;
			}
			e1 = s1 + pImageInfo[i].ImageSize;
			e2 = s2 + pImageInfo[j].ImageSize;
			if (s1 > s2)//if image 2 is at a lower location in flash
			{
				if(e2 > s1)//make sure the end of image 2 is also at a lower location in flash
				{
					printf("Error: Images %#08x and %#08x overlap in flash!\n", pImageInfo[i].ImageID, pImageInfo[j].ImageID);
					//printf("Bad: i1 %d i2 %d s1 %d s2 %d e1 %d e2 %d", i, j, s1, s2, e1, e2);
					return FALSE;
				}
			}
			else		//if image 1 is at a lower location in flash
			{
				if(e1 > s2)//make sure the end of image 1 is also at a lower location in flash
				{
					printf("Error: Images %#08x and %#08x overlap in flash!\n", pImageInfo[i].ImageID, pImageInfo[j].ImageID);
					//printf("Bad: i1 %d i2 %d s1 %d s2 %d e1 %d e2 %d", i, j, s1, s2, e1, e2);
					return FALSE;
				}
			}
		}
	}
	return TRUE;
}

void PrintUsage ()
{
	printf ("\n WTP Trusted Image Module Builder\n\n");
	#if !TRUSTED
	printf ("  ***Note: This is the non-trusted version only***\n\n");
	#endif
	printf ("  Usage:\n");
	printf ("  ======\n");
	printf ("  >> TBB.exe <Option>|<Option file> ...\n\n");
	printf ("  Options:\n");
	printf ("  ========\n");
	printf ("  -v\t\tVerbose mode.  Debugging information will be printed.\n\n");
	printf ("  Build a trusted or non trusted image module:\n\tTBB -m 1 -r sample_tim.txt\n\n");
	#if TRUSTED
	printf ("  Build an image module without digital signature and a hash of that image:\n");
	printf ("  \tTBB -m 2 -r sample_tim.txt -h timhash.bin\n\n");
	printf ("  Build a digital signature from image module hash and private key:\n");
	printf ("  \tTBB -m 3 -r privatekey.txt -h timhash.bin -s ds.bin\n\n");
	printf ("  Build a trusted image module from digital signature and image module:\n");
	printf ("  \tTBB -m 4 -s ds.bin -i TIMtestTrustNoDs.bin -o TIMtestTrustFromDs.bin\n\n");
	printf (" TIM Verify:\n");
	printf (" ===========\n");
	printf (" >>TBB.exe -m 1 -r TIM.txt -V\n\n");
	#endif
	printf (" Reading reserved data from external file:\n");
	printf (" =========================================\n");
	printf (" >>TBB.exe -r (N)TIM.txt -R ReservedDataFile.bin\n");
}

